/*****************************************************************************/
/*                                                                           */
/*   VECTREX HP3000 Software Emulator              Copyright K.Wilkins 1996  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Title      : VECTREX HP3000 Analog CRT X,Y,Z & Compare Emulation Module */
/*                                                                           */
/*   File Name  : ANALOG.C                                                   */
/*                                                                           */
/*   Author     : Keith Wilkins / small changes by C.S.                      */
/*                                                                           */
/*   Version    : 1.20                                                       */
/*                                                                           */
/*   Desciption : The functions is this module are designed to emulate the   */
/*                analog portion of the Vectrex that generates the CRT X,Y   */
/*                and Z signals and the COMPARE signal that is used to       */
/*                perform the AtoD function for the Controller Pads.         */
/*                                                                           */
/*                The main fAnalogTick() function in this block emulates     */
/*                the analog functions for N ticks where a tick is the       */
/*                smallest unit of activity in the alanog block. It is the   */
/*                responsibility of the Sequencer to call this function at   */
/*                the appropriate time with an appropriate value for N       */
/*                                                                           */
/*   Functions  : fAnalogInit() - Initialise the Analog block                */
/*                fAnalogTick() - Simulate N time units                      */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Revision History:                                                       */
/*                                                                           */
/*   Version    Date    Who  Description of changes                          */
/*   -------    ----    ---  ----------------------                          */
/*                                                                           */
/*    0.01    13/03/93  K.W  Creation                                        */
/*    1.00    05/08/96  K.W  Public release of DVE V1.0                      */
/*    1.20    20/10/96  C.S  Public release of DVE V1.2                      */
/*                                                                           */
/*****************************************************************************/

#define ANALOG_C

#include "vectrex.h"
#include "math.h"


SBYTE      lXSampleHold=0; //slong
SBYTE      lYSampleHold=0; //slong
UBYTE      tZSampleHold=0;


SLONG      lSIGDEF_XOUTPUT=0;
SLONG      lSIGDEF_YOUTPUT=0;
UBYTE      tSIGDEF_ZOUTPUT=0;

SLONG      DAC_lookup[256];

/************************************************************************/
/* some variables changed to globales... since fAnalogTick is sometimes */
/* inline now...                                                        */
/************************************************************************/

ULONG lwTickMark=0;
UBYTE tLastRamp=0;
SLONG lX1=0,lY1=0;
UBYTE tOldBlank=0,tBright=0;;

void (* fAnalogTick)(void);
void (* fAnalogJoystickUpdate)(void);


/*****************************************************************************/
/*                                                                           */
/*   Function    : fAnalogInit()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function initialises the variables and workings of   */
/*                 the module that simulates the analog functions of the     */
/*                 Vectrex.                                                  */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : BYTE of initialisation status TRUE/FALSE                  */
/*                                                                           */
/*****************************************************************************/

BOOL fAnalogInit(void)
{
 SWORD   loop;

 /*******************************/
 /* Create DAC log lookup table */
 /*******************************/
 /*
 DAC_lookup[0]=0;
 for(loop=1;loop<128;loop++)
 {
  DAC_lookup[loop]=32*log(abs(loop));
  printf("1DAC_lookup[%i]:=%i\n",loop,DAC_lookup[loop]);
  DAC_lookup[256-loop]=-(32*log(abs(loop)));
 }
 DAC_lookup[128]=-(32*log(abs(128)));
*/

 for(loop=0;loop<256;loop++)
 {
  DAC_lookup[loop]=(loop^0x80)-128;
 }

 return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fAnalogReset()                                            */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function resets the analog block, a hard reset       */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void fAnalogReset(void)
{

 lXSampleHold=0;
 lYSampleHold=0;
 tZSampleHold=0;


 lSIGDEF_XOUTPUT=0;
 lSIGDEF_YOUTPUT=0;
 tSIGDEF_ZOUTPUT=0;
 return;
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fAnalogTick()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function emulates the analog functions of the        */
/*                 vectrex. These functions produce the X,Y,Z deflection     */
/*                 signals for the CRT and the COMPARE signal for the        */
/*                 jostick AtoD system. The registers and values it needs    */
/*                 are taken from the 6522A Port A and B registers. The      */
/*                 value on port A is used as the DAC value an port B        */
/*                 is used for a number of the control lines that drive      */
/*                 the block such as SW,SEL0,SEL1,RAMP,BLANK.                */
/*                                                                           */
/*                -This function is sometimes inline in sequencer (6809...)  */
/*                -This function exists now 8 times, depending on            */
/*                 startup information the required is chosen at runtime     */
/*                                                                           */
/*                                                                           */
/*   Accepts     :                                                           */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fAnalogTick_linear_shift(void)
#define ANALOG_TICK_LINEAR_SHIFT
#include "anatick.cc"
#undef ANALOG_TICK_LINEAR_SHIFT

void fAnalogTick_linear_div(void)
#define ANALOG_TICK_LINEAR_DIV
#include "anatick.cc"
#undef ANALOG_TICK_LINEAR_DIV

void fAnalogTick_bank_shift(void)
#define ANALOG_TICK_BANKED_SHIFT
#include "anatick.cc"
#undef ANALOG_TICK_BANKED_SHIFT

void fAnalogTick_bank_div(void)
#define ANALOG_TICK_BANKED_DIV
#include "anatick.cc"
#undef ANALOG_TICK_BANKED_DIV

#define LIGHTPEN
void fAnalogTick_linear_shift_lightpen(void)
#define ANALOG_TICK_LINEAR_SHIFT
#include "anatick.cc"
#undef ANALOG_TICK_LINEAR_SHIFT

void fAnalogTick_linear_div_lightpen(void)
#define ANALOG_TICK_LINEAR_DIV
#include "anatick.cc"
#undef ANALOG_TICK_LINEAR_DIV

void fAnalogTick_bank_shift_lightpen(void)
#define ANALOG_TICK_BANKED_SHIFT
#include "anatick.cc"
#undef ANALOG_TICK_BANKED_SHIFT

void fAnalogTick_bank_div_lightpen(void)
#define ANALOG_TICK_BANKED_DIV
#include "anatick.cc"
#undef ANALOG_TICK_BANKED_DIV
#undef LIGHTPEN

#define ANALOG_IMAGER
void fAnalogTick_linear_shift_imager(void)
#define ANALOG_IMAGER
#define ANALOG_TICK_LINEAR_SHIFT
#include "anatick.cc"
#undef ANALOG_TICK_LINEAR_SHIFT

void fAnalogTick_linear_div_imager(void)
#define ANALOG_TICK_LINEAR_DIV
#include "anatick.cc"
#undef ANALOG_TICK_LINEAR_DIV

void fAnalogTick_bank_shift_imager(void)
#define ANALOG_TICK_BANKED_SHIFT
#include "anatick.cc"
#undef ANALOG_TICK_BANKED_SHIFT

void fAnalogTick_bank_div_imager(void)
#define ANALOG_TICK_BANKED_DIV
#include "anatick.cc"
#undef ANALOG_TICK_BANKED_DIV

#undef ANALOG_IMAGER

/*****************************************************************************/
/*                                                                           */
/*   Function    : fAnalogJoystickUpdate()                                   */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function updates the compare pin based on all of the */
/*                 current switch settings                                   */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void fAnalogJoystickUpdateDigital(void)
{
 SBYTE   sbDACVal;
 SBYTE   sbJOYVal=0;

 sbDACVal=(SBYTE)tSIGDEF_DACBUS;

 switch (tSIGDEF_AMUXSEL)
 {
  case 0: // P1 controller Y axis
   if (bGblDigitalJoystick)
   {
    if (fKBDisdown(wGblPly1JoystkRt))
     sbJOYVal=127; //Max negative
    if (fKBDisdown(wGblPly1JoystkLf))
     sbJOYVal=-127; //Max positive
   }
   break;
  case 2: // P1 controller X axis
   if(bGblDigitalJoystick)
   {
    if (fKBDisdown(wGblPly1JoystkDn))
     sbJOYVal=-127; //Max negative
    if (fKBDisdown(wGblPly1JoystkUp))
     sbJOYVal=127; //Max positive
   }
   break;
  case 4: // P2 controller Y axis
   if (bGblDigitalJoystick)
   {
    if (fKBDisdown(wGblPly2JoystkRt))
     sbJOYVal=127; //Max negative
    if (fKBDisdown(wGblPly2JoystkLf))
     sbJOYVal=-127; //Max positive
   }
   break;
  case 6: // P2 controller X axis
   if (bGblDigitalJoystick)
   {
    if (fKBDisdown(wGblPly2JoystkDn))
     sbJOYVal=-127; //Max negative
    if (fKBDisdown(wGblPly2JoystkUp))
     sbJOYVal=127; //Max positive
   }
   break;
  default:
   fDebug(VERROR,("fAnalogTick() Illegal analog mux setting"));
   break;
 }

 /**************/
 /* Do compare */
 /**************/
 tSIGDEF_COMPARE=(sbJOYVal>sbDACVal)?0x20:0;
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fAnalogJoystickUpdate()                                   */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function updates the compare pin based on all of the */
/*                 current switch settings                                   */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fAnalogJoystickUpdateAnalog(void)
{
 SBYTE   sbDACVal;
 SBYTE   sbJOYVal=0;

 static  SLONG anajoyx1=0,anajoyy1=0;
 static  SLONG anajoyx2=0,anajoyy2=0;

 sbDACVal=(SBYTE)tSIGDEF_DACBUS;

 /*******************************************************/
 /* If analog joystick is enabled then process movement */
 /*******************************************************/
  if (fKBDisdown(wGblPly1JoystkRt))
  {
   if (anajoyy1<32767-tGblJoystickSensitivity)
    anajoyy1+=tGblJoystickSensitivity;
  }
  else if (fKBDisdown(wGblPly1JoystkLf))
  {
   if (anajoyy1>-32768+tGblJoystickSensitivity)
   anajoyy1-=tGblJoystickSensitivity;
  }
  else if (bGblJoystickAutoCentre)
  {
   if (anajoyy1>0)
    anajoyy1-=tGblJoystickSensitivity;
   else if(anajoyy1<0)
    anajoyy1+=tGblJoystickSensitivity;
  }
  if (fKBDisdown(wGblPly1JoystkUp))
  {
   if (anajoyx1<32767-tGblJoystickSensitivity)
    anajoyx1+=tGblJoystickSensitivity;
  }
  else if (fKBDisdown(wGblPly1JoystkDn))
  {
   if (anajoyx1>-32768+tGblJoystickSensitivity)
    anajoyx1-=tGblJoystickSensitivity;
  }
  else if (bGblJoystickAutoCentre)
  {
   if (anajoyx1>0)
    anajoyx1-=tGblJoystickSensitivity;
   else if(anajoyx1<0)
    anajoyx1+=tGblJoystickSensitivity;
  }
  if (fKBDisdown(wGblPly2JoystkRt))
  {
   if (anajoyy2<32767-tGblJoystickSensitivity)
    anajoyy2+=tGblJoystickSensitivity;
  }
  else if (fKBDisdown(wGblPly2JoystkLf))
  {
   if (anajoyy2>-32768+tGblJoystickSensitivity)
    anajoyy2-=tGblJoystickSensitivity;
  }
  else if(bGblJoystickAutoCentre)
  {
   if (anajoyy2>0)
    anajoyy2-=tGblJoystickSensitivity;
   else if (anajoyy2<0)
    anajoyy2+=tGblJoystickSensitivity;
  }
  if (fKBDisdown(wGblPly2JoystkUp))
  {
   if(anajoyx2<32767-tGblJoystickSensitivity) anajoyx2+=tGblJoystickSensitivity;
  }
  else if (fKBDisdown(wGblPly2JoystkDn))
  {
   if (anajoyx2>-32768+tGblJoystickSensitivity)
    anajoyx2-=tGblJoystickSensitivity;
  }
  else if (bGblJoystickAutoCentre)
  {
   if (anajoyx2>0)
    anajoyx2-=tGblJoystickSensitivity;
   else if (anajoyx2<0)
    anajoyx2+=tGblJoystickSensitivity;
  }
 switch (tSIGDEF_AMUXSEL)
 {
  case 0: // P1 controller Y axis
     sbJOYVal=anajoyy1>>8;
   break;
  case 2: // P1 controller X axis
    sbJOYVal=anajoyx1>>8;
   break;
  case 4: // P2 controller Y axis
    sbJOYVal=anajoyy2>>8;
   break;
  case 6: // P2 controller X axis
    sbJOYVal=anajoyx2>>8;
   break;
  default:
   fDebug(VERROR,("fAnalogTick() Illegal analog mux setting"));
   break;
 }

 /**************/
 /* Do compare */
 /**************/
 tSIGDEF_COMPARE=(sbJOYVal>sbDACVal)?0x20:0;
}
/* End of ANALOG.C */

