/*****************************************************************************/
/*                                                                           */
/*   VECTREX HP3000 Software Emulator              Copyright K.Wilkins 1996  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Title      : Vectrex Cartridge ROM Emulation function module            */
/*                                                                           */
/*   File Name  : CART_ROM.C                                                 */
/*                                                                           */
/*   Author     : Keith Wilkins                                              */
/*                                                                           */
/*   Version    : 1.00                                                       */
/*                                                                           */
/*   Desciption : This file contains the functions that emulate the Vectrex  */
/*                Cartridge ROM. The ROM write is for patching and debugging */
/*                                                                           */
/*   Functions  : fCartridgeROMLoad()   - Init & Load the System RAM         */
/*                fCartridgeROMSave()   - Save the System RAM area to disk   */
/*                fCartridgeROMRead()   - Read a RAM location                */
/*                fCartridgeROMWrite()  - Write to a RAM location            */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Revision History:                                                       */
/*                                                                           */
/*   Version    Date    Who  Description of changes                          */
/*   -------    ----    ---  ----------------------                          */
/*                                                                           */
/*    0.01    14/03/93  K.W  Creation of empty file and descriptions         */
/*    1.00    05/08/96  K.W  Public release of DVE V1.0                      */
/*                                                                           */
/*****************************************************************************/

#define CART_ROM_C

#include "vectrex.h"


/*****************************************************************************/
/*                                                                           */
/*   Function    : fCartridgeROMLoad()                                       */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function will open and load a binary file into the   */
/*                 32Kb Cartridge ROM space. Before loading all addresses    */
/*                 will be written to CART_ROM_FILL_VALUE. The file is       */
/*                 opened before the ROM space is cleared as if the load     */
/*                 fails we havent destroyed the current ROM Image.          */
/*                                                                           */
/*   Accepts     : char * Pointer to a file name, any length.                */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE/FALSE depending on load success.             */
/*                                                                           */
/*****************************************************************************/

BOOL fCartridgeROMLoad(UBYTE *pcROMFileName)
{
    static BOOL tFirstTime=TRUE;

    BOOL   bRetVal=TRUE;
    BOOL   bTmp;
    FILE   *pfCartridgeROM=NULL;
    ULONG  tLoop;
    UBYTE  tData;

    /* Only on the 1st execution do we declare the ROM Space */

    if(tFirstTime==TRUE)
    {
        tFirstTime=FALSE;
        if(f6809DeclareAddress(CART_ROM_BASE,CART_ROM_SIZE,fCartridgeROMRead,fCartridgeROMWrite)==FALSE)
        {
            fDebug(VERROR,("fCartridgeROMLoad() ROM Space delaration failed"));
            return(FALSE);
        }
    }

    /* Enable Cartridge ROM writes, and save current stats */

    bTmp=bGblCartridgeROMWriteEnable;
    bGblCartridgeROMWriteEnable=TRUE;

    /* Now clear out the Cartridge ROM area */

    for(tLoop=CART_ROM_BASE;tLoop<=CART_ROM_TOP;tLoop++)
    {
        f6809AddressWrite((UWORD)tLoop,CART_ROM_FILL_VALUE);
    }
    fDebug(MED,("fCartridgeROMLoad() Cartridge ROM Space Cleared"));

    /* Try and open the ROM image file */

    if(pcROMFileName!=NULL && strcmp(pcROMFileName,"")!=0)
    {
        if((pfCartridgeROM=fopen(pcROMFileName,"rb"))==NULL)
        {
            fDebug(VERROR,("fCartridgeROMLoad() Couldnt open file \"%s\"",pcROMFileName));
            bRetVal=FALSE;
        }
        else
        {
            fDebug(MED,("fCartridgeROMLoad() \"%s\" opened O.K",pcROMFileName));

            tLoop=CART_ROM_BASE;

            while(tLoop<=CART_ROM_TOP && fread(&tData,sizeof(UBYTE),1,pfCartridgeROM)==1)
            {
                f6809AddressWrite((UWORD)tLoop++,tData);
            }
            fDebug(MED,("fCartridgeROMLoad() Loaded %d Bytes",tLoop-CART_ROM_BASE));
            
            fclose(pfCartridgeROM);
        }
    }

    /* Restore the ROM writes flag */

    bGblCartridgeROMWriteEnable=bTmp;

    /* All finished */

    return(bRetVal);

}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fCartridgeROMSave()                                       */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Open a file on disk, write the contents of the cartridge  */
/*                 to disk and close the file                                */
/*                                                                           */
/*   Accepts     : char * Pointer to a file name, any length.                */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE/FALSE depending on load success.             */
/*                                                                           */
/*****************************************************************************/

BOOL fCartridgeROMSave(UBYTE *pcROMFileName)
{
    FILE    *pfCartridgeROM=NULL;
    UWORD   tLoop;
    UBYTE   tData;

    /* Try and open the ROM image file */

    if((pfCartridgeROM=fopen(pcROMFileName,"wb"))==NULL)
    {
        fDebug(VERROR,("fCartridgeROMSave() Couldnt open file %s",pcROMFileName));
        return(FALSE);
    }
    else
    {
        fDebug(MED,("fCartridgeROMSave() %s opened O.K",pcROMFileName));
    }

    /* Now Save CART_ROM_SIZE bytes */

    tLoop=CART_ROM_BASE;

    while(tLoop<CART_ROM_BASE+CART_ROM_SIZE)
    {
	    tData=fCartridgeROMRead(tLoop++);
       if(fwrite(&tData,sizeof(UBYTE),1,pfCartridgeROM)==0)
       {
           fclose(pfCartridgeROM);
           fDebug(VERROR,("fCartridgeROMSave() File write error, aborting save"));
           return(FALSE);
       }
    }
    fDebug(MED,("fCartridgeROMSave() Saved %d Bytes",tLoop-CART_ROM_BASE));

    fclose(pfCartridgeROM);

    /* All finished */

    return(TRUE);

}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fCartridgeROMRead()                                       */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Read a value from the ROM image, apply the address mask   */
/*                 to the address before reading to ensure were not out of   */
/*                 bounds.                                                   */
/*                                                                           */
/*   Accepts     : WORD Address pointer (16bit)                              */
/*                                                                           */
/*   Returns     : BYTE Data value                                           */
/*                                                                           */
/*****************************************************************************/

UBYTE fCartridgeROMRead(UWORD tAddress)
{
    return(atSystemImage[tAddress]);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fCartridgeROMWrite()                                      */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Write a byte to the image array, the address is masked    */
/*                 before the write to ensure we're within the bounds of the */
/*                 array. The write is only performed if the ROM write       */
/*                 enable flag is TRUE.                                      */
/*                                                                           */
/*   Accepts     : WORD Address and a BYTE of Data                           */
/*                                                                           */
/*   Returns     : BYTE of Data                                              */
/*                                                                           */
/*****************************************************************************/

UBYTE fCartridgeROMWrite(UWORD tAddress,UBYTE tData)
{
    if(bGblCartridgeROMWriteEnable==FALSE)
    {
        fDebug(VERROR,("fCartridgeROMWrite() Attempt to Write #$%02x to ROM $%04x",tData,tAddress));
    }
    else
    {
        atSystemImage[tAddress]=tData;
    }
    return(tData);
}






/* End of CART_ROM.C */



