/*****************************************************************************/
/*                                                                           */
/*   VECTREX HP3000 Software Emulator              Copyright K.Wilkins 1996  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Title      : 6809 Emulator code module                                  */
/*                                                                           */
/*   File Name  : EMU6809.C                                                  */
/*                                                                           */
/*   Author     : Keith Wilkins / C.S.                                       */
/*                                                                           */
/*   Version    : 1.20                                                       */
/*                                                                           */
/*   Desciption : This file contains all of the code that makes up the 6809  */
/*                emulator, it interfaces to the emulator through the        */
/*                functions shown below:                                     */
/*                                                                           */
/*   Functions  : f6809Init           - Initialises the 6809 emulator        */
/*                f6809Reset          - Performs a 6809 hardware reset       */
/*                f6809Tick           - Run a 6809 CPU Cycle                 */
/*                f6809DeclareAddress - Define H/W mapping for emulator      */
/*                f6809AddressRead    - Read from 6809 Address space         */
/*                f6809AddressWrite   - Write to 6809 Address space          */
/*                f6809RegistersRead  - Read 6809 CPU registers (struct)     */
/*                f6809RegistersWrite - Write 6809 CPU regiisters (struct)   */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Revision History:                                                       */
/*                                                                           */
/*   Version    Date    Who  Description of changes                          */
/*   -------    ----    ---  ----------------------                          */
/*                                                                           */
/*    0.01    14/03/93  K.W  Creation of empty stub file, description only   */
/*    1.00    05/08/96  K.W  Public release of DVE V1.0                      */
/*    1.20    20/10/96  C.S  MultiTick & Inline Stuff                        */
/*                                                                           */
/*****************************************************************************/

/* Notes : Address read/write functions are not used by the emulator code */
/*         as they write direct, these are for H/W emulation and debug    */
/*         use only. The register read/write passes a pointer to the      */
/*         registers structure defined in 6809emu.h.                      */
/*                                                                        */
/*         Interrupts are handled via the register read/write functions   */


#define EMU6809_C

#include "vectrex.h"

/*****************************/
/* global <inline> variables */
/*****************************/
SWORD  iCount=0;
UWORD  iInstruction=0;
#ifdef DEBUG_MON
BOOL   iBPTriggered=FALSE;
#endif

/* Define the Local copy of the CPU registers used in all CPU operations */

CPUREGISTERS  tsCPURegisters;


// Define the arrays of read/write function ptrs


UBYTE (*apfReadHandlers[CPU_MEMORY_SIZE])(UWORD);
UBYTE (*apfWriteHandlers[CPU_MEMORY_SIZE])(UWORD,UBYTE);


/* Define the unmapped memory handlers */

UBYTE    fUnmappedMemoryRead(UWORD tAddr)
{
 fDebug(VERROR,("f6809MemoryRead() - <PC=%04x> Attempt to read from unmapped memory at $%04x",tsCPURegisters.tPC,tAddr));
 tAddr=0;
 return(CPU_UNMAPPED_FILL);
}

UBYTE    fUnmappedMemoryWrite(UWORD tAddr,UBYTE tData)
{
 fDebug(VERROR,("f6809MemoryWrite() - <PC=%04x> Attempt to write #$%02x to unmapped memory at $%04x",tsCPURegisters.tPC,tData,tAddr));
 tAddr=0;
 tData=0;
 return(CPU_UNMAPPED_FILL);
}

UWORD i6809status=CPU_HALT;
UWORD iLastAddress=0;
UWORD iBreakpoint=0x0;
UWORD iBreakpointEnable=FALSE;
UWORD iSkipCount=0;   

//
// Define a static memory space for "straight" memory applications
//

static UBYTE atSystemImage[CPU_MEMORY_SIZE];

//
// Instruction & cycle count file inclusions
//

#include "cod6809.h"
#include "ins6809.h"
#include "cnt6809.h"



/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809Init()                                               */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function initialises the CPU emulator and all assoc- */
/*                 iated arrays i.e the Memory handlers. And then resets the */
/*                 CPU ready for program execution.                          */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : UBYTE TRUE/FALSE depending on success                      */
/*                                                                           */
/*****************************************************************************/

BOOL f6809Init(void)
{
 ULONG loop;
 UWORD opcode;
 UWORD operand;

 // Initialise the handler array, to unmapped

 for (loop=0;loop<CPU_MEMORY_SIZE;loop++)
 {
  apfReadHandlers[loop]=fUnmappedMemoryRead;
  apfWriteHandlers[loop]=fUnmappedMemoryWrite;
 }

 /* Make sure 6809 is halted */
 i6809status=CPU_HALT;

 // Create and initialise the cycle count lookup table

 // 64K Instruction lookup table - Handles Page 2 & 3 insructions
 // without the need for doing a compare in the 6809tick.

 for(loop=0;loop<0x010000;loop++)
 {
  operand=loop>>8;
  opcode=loop&0xff;

  if (opcode==0x10)
   opcode=0x100+operand;
  if(opcode==0x11)
   opcode=0x200+operand;
  cycle_count[loop]=packed_cycle_count[opcode];
 }
 f6809Reset();
 return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809Reset()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This executes a hardware reset of the 6809 CPU. Set all   */
/*                 registers to reset state and fetch the PC from $FFFE and  */
/*                 $FFFF.                                                    */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void f6809Reset(void)
{
   UWORD    tBootPtr;

   /* On reset DP must be cleared */

   tsCPURegisters.tDP=0;

   /* Clear all other resgisters for the mo */

   tsCPURegisters.tX=0;
   tsCPURegisters.tY=0;
   tsCPURegisters.uABD.tD=0;
   tsCPURegisters.tUSP=0;
   tsCPURegisters.tHSP=0;
   tsCPURegisters.uCC.tCC=0;
   f6809Init8086CC();

//   tsCPURegisters.uCC8086.sCC.bC=1;
//   tsCPURegisters.uCC8086.sCC.bH=1;
//   tsCPURegisters.uCC8086.sCC.bZ=1;
//   tsCPURegisters.uCC8086.sCC.bN=1;
//   tsCPURegisters.uCC8086.sCC.bV=1;

   tsCPURegisters.tPC=0;

   /* Read the PC from $FFFE (high),$FFFF (low) */

   tBootPtr=(UWORD) f6809AddressRead(0xffff)+(f6809AddressRead(0xfffe)<<8);

   tsCPURegisters.tPC=tBootPtr;
   iLastAddress=tBootPtr;

   /* Clear the Hardware IRQs */

   tsCPURegisters.tNMIFlag=FALSE;
   tsCPURegisters.tIRQFlag=FALSE;
   tsCPURegisters.tFIRQFlag=FALSE;

   /* Mask the IRQ and FIRQ in the CC register */

   tsCPURegisters.uCC.sCC.bI=1;
   tsCPURegisters.uCC.sCC.bF=1;

//   tsCPURegisters.uCC.tCC=tsCPURegisters.tCC | CC_I_FLAG | CC_F_FLAG;

   i6809status=CPU_FETCH;
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809Tick()                                               */
/*                                                                           */
/*   Author      : K.W / C.S.                                                */
/*                                                                           */
/*   Description : This function runs one cycle of CPU time emulation.       */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void f6809Tick(void)
{
 // Code moved to sequence.c and then seq.cc and then *cpu.cc
}
 
/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809LastInsAddr()                                        */
/*                                                                           */
/*   Return address of last/current instruction executed                     */
/*                                                                           */
/*****************************************************************************/

UWORD f6809LastInsAddr(void)
{
 return(iLastAddress);
}
 
/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809StatusRead()                                         */
/*                                                                           */
/*   Return current status of the processor                                  */
/*                                                                           */
/*****************************************************************************/

UWORD f6809StatusRead(void)
{
 return(i6809status);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809StatusWrite()                                        */
/*                                                                           */
/*   Set the current status of the processor                                 */
/*                                                                           */
/*****************************************************************************/

void f6809StatusWrite(UWORD tStat)
{
 i6809status=tStat;
 return;
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809SetBreakpoint()                                      */
/*                                                                           */
/*   Set the CPU breakpoint                                                  */
/*                                                                           */
/*****************************************************************************/

void f6809SetBreakpoint(UWORD tBreakpoint,UWORD tSkipCount)
{
 iBreakpoint=tBreakpoint;
 iSkipCount=tSkipCount;
 iBreakpointEnable=TRUE;
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809SClearBreakpoint()                                   */
/*                                                                           */
/*   Clear processor breakpoint                                              */
/*                                                                           */
/*****************************************************************************/

void f6809ClearBreakpoint(void)
{
 iBreakpointEnable=FALSE;
}


/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809ReadBreakpoint()                                     */
/*                                                                           */
/*   Readback the processor breakpoint                                       */
/*                                                                           */
/*****************************************************************************/

void f6809ReadBreakpoint(UWORD *tBreakpoint,UWORD *tSkip,UWORD *tEnable)
{
 if (tBreakpoint!=NULL)
  *tBreakpoint=iBreakpoint;
 if (tSkip!=NULL)
  *tSkip=iSkipCount;
 if (tEnable!=NULL)
  *tEnable=iBreakpointEnable;
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809DelcareAddress()                                     */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function declares a memory block to the CPU, so that */
/*                 whenever a read/write is done to that area the read/write */
/*                 functions specified here are called.                      */
/*                                                                           */
/*   Accepts     : UWORD address base and block size, function pointers r/w   */
/*                                                                           */
/*   Returns     : UBYTE TRUE/FALSE on success fail.                          */
/*                                                                           */
/*****************************************************************************/

BOOL f6809DeclareAddress(UWORD wAddrBase,UWORD bSize,UBYTE (*pfReadFunc)(UWORD),UBYTE (*pfWriteFunc)(UWORD,UBYTE))
{
 ULONG lwLoop;
// fDebug(LOW,("f6809DeclareAddress() - Base %04x, Size %04x",wAddrBase,bSize));

 /* No go through the hanlde map array and setup the handle no */

 for (lwLoop=(ULONG)wAddrBase;lwLoop<(ULONG)wAddrBase+bSize;lwLoop++)
 {
  apfReadHandlers[lwLoop]=pfReadFunc;
  apfWriteHandlers[lwLoop]=pfWriteFunc;
 }

 /* return the handle number, init would have allocated handle=0 */
 /* for the unmapped status so we are guaranteed to be >0==TRUE  */
 return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809AddressRead()                                        */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Read a UBYTE from the given address, using the device      */
/*                 table to revector the read request. The read counters are */
/*                 not incremented by this function.                         */
/*                                                                           */
/*   Accepts     : UWORD address                                              */
/*                                                                           */
/*   Returns     : UBYTE data read from address                               */
/*                                                                           */
/*****************************************************************************/

UBYTE f6809AddressRead(UWORD tAddr)
{
//  fDebug(LOW,("f6809AddrRead() - Read from %04x",tAddr));
// printf("\nf6809AddrRead() - Read from %04x",tAddr);
 return( (*apfReadHandlers[tAddr]) (tAddr) );
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809AddressWrite()                                       */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Write a UBYTE to the given address, using the device table */
/*                 to revector the write request. The write counters are not */
/*                 incremented by this function.                             */
/*                                                                           */
/*   Accepts     : UWORD address and a UBYTE of DATA to be written             */
/*                                                                           */
/*   Returns     : UBYTE data from write address                              */
/*                                                                           */
/*****************************************************************************/

UBYTE f6809AddressWrite(UWORD tAddr,UBYTE tData)
{
//  fDebug(LOW,("f6809AddrWrite() - Write %02x to %04x",tData,tAddr));
 return( (*apfWriteHandlers[tAddr]) (tAddr,tData) );
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809RegistersRead()                                      */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This copies the current 6809 register structure into the  */
/*                 given pointer.                                            */
/*                                                                           */
/*   Accepts     : pointer to register structure                             */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void f6809RegistersRead(CPUREGISTERS *pts6809Registers)
{
 // Before we release the registers we must update the 6809 CC

 tsCPURegisters.uCC.sCC.bC=tsCPURegisters.uCC8086.sCC.bC;
 tsCPURegisters.uCC.sCC.bH=tsCPURegisters.uCC8086.sCC.bH;
 tsCPURegisters.uCC.sCC.bZ=tsCPURegisters.uCC8086.sCC.bZ;
 tsCPURegisters.uCC.sCC.bN=tsCPURegisters.uCC8086.sCC.bN;
 tsCPURegisters.uCC.sCC.bV=tsCPURegisters.uCC8086.sCC.bV;

 *pts6809Registers=tsCPURegisters;
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : f6809RegistersWrite()                                     */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This copies the passed resgister structre into the current*/
/*                 CPU register strucure, update ALL registers. When you     */
/*                 want to change a single register do a register read, then */
/*                 modify your chosen register then to a registe write.      */
/*                                                                           */
/*   Accepts     : pointer to a register structure                           */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void f6809RegistersWrite(CPUREGISTERS *pts6809Registers)
{
 tsCPURegisters=*pts6809Registers;

 // Before we exit we must update the 8086 CC

 tsCPURegisters.uCC8086.sCC.bC=tsCPURegisters.uCC.sCC.bC;
 tsCPURegisters.uCC8086.sCC.bH=tsCPURegisters.uCC.sCC.bH;
 tsCPURegisters.uCC8086.sCC.bZ=tsCPURegisters.uCC.sCC.bZ;
 tsCPURegisters.uCC8086.sCC.bN=tsCPURegisters.uCC.sCC.bN;
 tsCPURegisters.uCC8086.sCC.bV=tsCPURegisters.uCC.sCC.bV;
}

/* End of EMU6809.C */
