/*****************************************************************************/
/*                                                                           */
/*   VECTREX HP3000 Software Emulator              Copyright K.Wilkins 1996  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Title      : Vectrex emulator text/graphics I/O functions               */
/*                                                                           */
/*   File Name  : IO.C                                                       */
/*                                                                           */
/*   Author     : Keith Wilkins / Christopher Salomon                        */
/*                                                                           */
/*   Version    : 1.20                                                       */
/*                                                                           */
/*   Desciption : This file contains the functions that allow the vectrex    */
/*                emulator to output data to the screen and read data from   */
/*                the keyboard,mouse etc..                                   */
/*                                                                           */
/*                These functions are the most PRIMATIVE Level of IO and     */
/*                as such are the fastest. The next level up from here are   */
/*                the window primatives found in WINDOW.C. Hopefully most    */
/*                functions can interface thorugh the upper functions as     */
/*                these provide better control.                              */
/*                                                                           */
/* Changes:       1.2                                                        */
/*                Removed nearly all VESA code of Keith.                     */
/*                It wasn't all that compatible in my opinion                */
/*                Granularity was defined as 64K per default, which is       */
/*                not allways true, bank switching was done via interrupt    */
/*                (very time consuming), now via realmode function call,     */
/*                which is still not that good, but I didn't want to fiddle  */
/*                with the 32bit pmode routine provided with VESA 2.0.       */
/*                Furthermore he used to set both, read AND write windows    */
/*                which allways took TWO interrupts to complete!!!           */
/*                                                                           */
/*                I use the function provided by 1.? BIOS extentions.        */
/*                I used the SVGALIB V6.0 from scitech.                      */
/*                The only function call which differs from the version 1.0  */
/*                is:  void fIOsetpalette(void)                              */
/*                and palette.h is now an own palette.c file                 */
/*                                                                           */
/*                  Note: LineDraw() function does NO CLIPPING anymore!      */
/*                        Since the emulator uses 'inline' linedrawing which */
/*                        uses clipping, I changed the function not to use   */
/*                        clipping, for drawing windows and the like with    */
/*                        it.                                                */
/*                                                                           */
/*                                                                           */
/*   Functions  : fIOInit()        - Initialise the I/O System               */
/*                fIORestore()     - Restore I/O to its virgin state         */
/*                fIOOutChar()     - Output a single char to the screen      */
/*                fIOOutText()     - Output a text string to the screen      */
/*                fIOPrintf()      - Print a text string to the screen       */
/*                fIOSetClipping() - Set the clipping window                 */
/*                fIOGetScreenRes()- Return the screen size                  */
/*                fIOPlotPixel()   - Plot a pixel                            */
/*                fIOGetPixel()    - get a pixel                             */
/*                fIODrawLine()    - Draw a line                             */
/*                fIODrawBox()     - Draw a box on screen filled/outline     */
/*                fIOAreaSave()    - Save a screen area to memory            */
/*                fIOAreaLoad()    - Load a screen area from memory          */
/*                fIOAreaMove()    - Move a screen area                      */
/*                fIOAreaScroll()  - Scroll a given screen area              */
/*                fIOAreaClear()   - Erase a screen area                     */
/*                fIOload_pcx_overlay() - Load a PCX overlay file            */
/*                fIOInitMouse()                                             */
/*****************************************************************************/
/*                                                                           */
/*   Revision History:                                                       */
/*                                                                           */
/*   Version    Date    Who  Description of changes                          */
/*   -------    ----    ---  ----------------------                          */
/*                                                                           */
/*    0.01    20/03/93  K.W  Creation                                        */
/*    1.00    05/08/96  K.W  Public release of DVE V1.0                      */
/*    1.20    03/10/96  C.S  Basic functions rewritten                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* These functions MUST NOT use any fDebug calls as they may be begin used   */
/* to setup the debug window and as such will generate recusive function     */
/* calls until the stack if they call fDebug whilst it is being setup        */
/*****************************************************************************/



#define IO_C

#include "vectrex.h"
#include "font.h"
#include "palette.h"


/*********************************************************/
/* function prototypes for banked and linear framebuffer */
/*********************************************************/
void fIOPlotPixellin(SWORD iXpos,SWORD iYpos,SWORD iColour);
void fIOPlotPixelbank(SWORD iXpos,SWORD iYpos,SWORD iColour);
UBYTE fIOGetPixellin(SWORD iXpos,SWORD iYpos);
UBYTE fIOGetPixelbank(SWORD iXpos,SWORD iYpos);
void fIODrawLinelin(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2,SWORD iColour);
void fIODrawLinebank(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2,SWORD iColour);
void fIODrawMaskLinelin(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2,SWORD iColour);
void fIODrawMaskLinebank(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2,SWORD iColour);
void fIOcopyvesascreenlin(UBYTE *source,SLONG swidth,SLONG sheight,SLONG bpp);
void fIOcopyvesascreenbank(UBYTE *source,SLONG swidth,SLONG sheight,SLONG bpp);
BOOL fIOAreaLoadlin(SWORD iXdest,SWORD iYdest,BITBLOCK *tData);
BOOL fIOAreaLoadbank(SWORD iXdest,SWORD iYdest,BITBLOCK *tData);

BITBLOCK* fIOAreaSavelin(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2);
BITBLOCK* fIOAreaSavebank(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2);

UWORD fIOInitMouselin(void);
UWORD fIOInitMousebank(void);

UBYTE (* fIOGetPixel) (SWORD, SWORD);

/***********************************************************/
/* old style functions get pointer to the above at runtime */
/***********************************************************/
void      (* fIOPlotPixel)(SWORD,SWORD,SWORD);
void      (* fIODrawMaskLine)(SWORD,SWORD,SWORD,SWORD,SWORD);
void      (* fIODrawLine)(SWORD,SWORD,SWORD,SWORD,SWORD);
void      (* fIOcopyvesascreen)(UBYTE *,SLONG,SLONG,SLONG);
BOOL      (* fIOAreaLoad)(SWORD,SWORD,BITBLOCK *);
BITBLOCK* (* fIOAreaSave)(SWORD,SWORD,SWORD,SWORD);
UWORD     (* fIOInitMouse)(void);


/********************/
/* global variables */
/********************/
SV_devCtx *svdc=NULL;
SWORD iSWTmp;
int offset;
UWORD  rectengle_clipping_enable=TRUE;
UBYTE* clip_memory=NULL;
ULONG  linestartpos[MAX_SCREEN_LINES];
UBYTE* overlay_memory=NULL;
int framebuffer=FB_NONE;
UWORD vesa_offset_compare;
UWORD vesa_segment_of_window[2]={0,0};
UWORD vesa_address_of_window[2]={0,0};
UWORD vesa_granularity;
UWORD vesa_granularity_potenz;
UWORD vesa_and_x;
ULONG vesa_write_seg;
ULONG vesa_read_seg;
UWORD vesa_read_window;
UWORD vesa_write_window;
UWORD WinFuncSegment,WinFuncOffset;

UWORD offset_x;
UWORD new_win_address;

volatile ULONG mouse_code=0;
volatile ULONG mouse_x=0;
volatile ULONG mouse_y=0;
volatile ULONG vmouse_x=0;
volatile ULONG vmouse_y=0;
volatile ULONG mouse_button=0;
volatile int mchanged=FALSE;



/***************************/
/* module global variables */
/***************************/
static SLONG mlastoffset=-1;
static SLONG mlastoffsetx=-1;
static SLONG mlastwin=-1;
static UBYTE mlastscreendata=0;
static UBYTE mlastclipdata=0;
static int minstalled=0;


static SWORD iScreenXres;
static SWORD iScreenYres;
static SWORD iScreenZres;

static SWORD iClippingX1;
static SWORD iClippingY1;

static SWORD iClippingX2;
static SWORD iClippingY2;

static BOOL initialised=FALSE;

static FILE *pcx_akt_handle=NULL;
static PCX_HEADER pcx_header;
static UBYTE *pcx_load_buffer=NULL;
static UWORD pcx_muss_gelesen_werden=FALSE;
static int pcx_buffer_counter=0;

/******************************************************************/
/* function prototypes for banked framebuffer bankswitch routines */
/******************************************************************/
void (*vesa_set_write_window)(void)=NULL;
void (*vesa_set_read_window)(void)=NULL;

/**********************************************/
/* function to clear vesa mode variable stuff */
/**********************************************/
void vesa_clear_modus_infos(void)
{
 vesa_read_window=-1;              // 0 or 1
 vesa_write_window=-1;             // 0 or 1
 vesa_segment_of_window[0]=0;      // probably 0xa000
 vesa_segment_of_window[1]=0;      // probably 0xa000
 vesa_address_of_window[0]=0;      // window position on VESA-Card (read or write)
 vesa_address_of_window[1]=0;      // window position on VESA-Card (read or write)
 vesa_read_seg=0;                  // probably 0xa000
 vesa_write_seg=0;                 // probably 0xa000
 vesa_granularity=-1;              // usually 64, but sometime 4
 vesa_granularity_potenz=0;        // well...
 vesa_and_x=1023;
 vesa_offset_compare=0;
 WinFuncOffset=0;                  // realmode bankswitch function offset
 WinFuncSegment=0;                 // realmode bankswitch function segment
}

/***********************************************************/
/* banked framebuffer bank set function CALL, WRITE WINDOW */
/***********************************************************/
void set1(void)
{
 RMREGS wregs;
 RMSREGS wsregs;
 wregs.x.bx=vesa_write_window;
 wregs.x.dx=vesa_address_of_window[vesa_write_window];
 PM_callRealMode(WinFuncSegment,WinFuncOffset,&wregs,&wsregs);
}

/**********************************************************/
/* banked framebuffer bank set function INT, WRITE WINDOW */
/**********************************************************/
void set2(void)
{
 union REGS regs;
 regs.w.ax = 0x4f05;
 regs.w.bx = vesa_write_window;
 regs.w.cx = 0;
 regs.w.dx = vesa_address_of_window[vesa_write_window];
 int386(0x10,&regs,&regs);
}

/**********************************************************/
/* banked framebuffer bank set function CALL, READ WINDOW */
/**********************************************************/
void set3(void)
{
 RMREGS rregs;
 RMSREGS rsregs;
 rregs.x.bx=vesa_read_window;
 rregs.x.dx=vesa_address_of_window[vesa_read_window];
 PM_callRealMode(WinFuncSegment,WinFuncOffset,&rregs,&rsregs);
}

/*********************************************************/
/* banked framebuffer bank set function INT, READ WINDOW */
/*********************************************************/
void set4(void)
{
 union REGS regs;
 regs.w.ax = 0x4f05;
 regs.w.bx = vesa_read_window;
 regs.w.cx = 0;
 regs.w.dx = vesa_address_of_window[vesa_read_window];
 int386(0x10,&regs,&regs);
}

/***************************************/
/* initialize banked framebuffer modus */
/***************************************/
void vesa_init_modus_infos(UWORD mode)
{
 ModeInfoBlock far *modeInfo=NULL;
 REALPTR modeInfo_rp;
 union REGS regs;
 struct SREGS sregs;
 DOS4GW_RMREGS rmregs;

 /*****************************/
 /* get the mode informations */
 /*****************************/
 modeInfo = (ModeInfoBlock far *) fDOSmalloc(&modeInfo_rp,sizeof(ModeInfoBlock));
 if (modeInfo==NULL)
 {
  fFatal("fIOinit() - Couldn't get DOS memory!!!");
 }

 memset(&rmregs,0,sizeof(rmregs));
 memset(&regs,0,sizeof(regs));
 memset(&sregs,0,sizeof(sregs));
 segread(&sregs);

 rmregs.eax = 0x4f01;
 rmregs.ecx = (int) mode;
 rmregs.es = modeInfo_rp.Segment;
 rmregs.ds = modeInfo_rp.Segment;
 regs.x.eax = 0x300;
 regs.x.ebx = 0x10;
 regs.x.ecx = 0;
 sregs.es = FP_SEG(&rmregs);
 regs.x.edi = FP_OFF(&rmregs);
 int386x(0x31,&regs,&regs,&sregs);
 if((rmregs.eax&0xffff) != 0x4f)
 {
  fFatal("fIOinit() - Couldn't get VESA Info on Mode!");
 }

 /**********************************************************/
 /* ok VESA modus got, now initialize the needed variables */
 /**********************************************************/
 vesa_clear_modus_infos();

 /*******************************************/
 /* First locate the READ and WRITE windows */
 /*******************************************/
 if (modeInfo->WinAAttributes&1)
 {
  if (modeInfo->WinAAttributes&2)
  {
   /*************************************/
   /* First Window is a readable window */
   /*************************************/
   vesa_read_window=0;
   vesa_segment_of_window[vesa_read_window]=modeInfo->WinASegment;
  }
  if (modeInfo->WinAAttributes&4)
  {
   /**************************************/
   /* First Window is a writeable window */
   /**************************************/
   vesa_write_window=0;
   vesa_segment_of_window[vesa_write_window]=modeInfo->WinASegment;
  }
 }
 if (modeInfo->WinBAttributes&1)
 {
  if (modeInfo->WinBAttributes&2)
  {
   /***************************************/
   /* Second Window is a readeable window */
   /***************************************/
   vesa_read_window=1;
   vesa_segment_of_window[vesa_read_window]=modeInfo->WinBSegment;
  }
  if (modeInfo->WinBAttributes&4)
  {
   /***************************************/
   /* Second Window is a writeable window */
   /***************************************/
   vesa_write_window=1;
   vesa_segment_of_window[vesa_write_window]=modeInfo->WinBSegment;
  }
 }

 /********************************************************/
 /* Get granularity and calculate some stuff, so we only */
 /* use shifts and 'AND's to calculate pages             */
 /********************************************************/
 vesa_granularity=modeInfo->WinGranularity;
 while (vesa_granularity>1)
 {
  vesa_granularity=vesa_granularity>>1;
  vesa_granularity_potenz++;
  vesa_and_x=(vesa_and_x<<1)+1;
 }
 vesa_granularity_potenz+=10; /* 10 equals * 1024 (== 2^10) */
 vesa_granularity=modeInfo->WinGranularity;

 /*****************************************************/
 /* look if there is a realmode bank setting routine  */
 /* I'm to lazy to look for a protected mode routine, */
 /* which is only available under VESA 2.0            */
 /*****************************************************/
 WinFuncSegment=modeInfo->WinFuncSegment;
 WinFuncOffset=modeInfo->WinFuncOffset;
 if (WinFuncOffset+WinFuncSegment==NULL)
 {
  /********************************************************/
  /* no function found, have to bankswitch via INT (yuk!) */
  /********************************************************/
  vesa_set_write_window=set2;
  vesa_set_read_window=set4;
 }
 else
 {
  /******************************************/
  /* yep, set bank via far call (real mode) */
  /******************************************/
  vesa_set_write_window=set1;
  vesa_set_read_window=set3;
 }
 /****************************************************************/
 /* just set the rest of the variables and initialize vesa-banks */
 /****************************************************************/
 vesa_offset_compare=vesa_granularity << 10;
 vesa_write_seg=vesa_segment_of_window[vesa_write_window]*16;
 vesa_read_seg=vesa_segment_of_window[vesa_read_window]*16;
 vesa_address_of_window[vesa_write_window]=0;
 vesa_set_write_window();
 vesa_address_of_window[vesa_read_window]=0;
 vesa_set_read_window();

 vesa_address_of_window[vesa_write_window]=-1;
 vesa_address_of_window[vesa_read_window]=-1;

 fDOSfree(&modeInfo_rp);
}

/*********************************/
/* get vesa modus on modus datas */
/*********************************/
/*****************************************************************/
/* For VESA 2.0 compliance, the VESA modes are not fixed anymore */
/* we have to look for the required mode ourself                 */
/*****************************************************************/
UWORD get_mode_on_specs(short x_res, short y_res, short color_bits, short special_flags)
{
 short i=0;
 SV_modeInfo modeInfo;
 while (svdc->modeList[i]!=(ushort)-1)      /* Pointer to supported modes       */
 {
  if (SV_getModeInfo(svdc->modeList[i],&modeInfo)==FALSE)
  {
   /***************************/
   /* only dummy mode go here */
   /***************************/
   if (bGblDisplayVerbose==TRUE)
    printf("\nMode:%04hx not available!", svdc->modeList[i]);
  }
  else
  {
   /*********************************************/
   /* just check the things which were required */
   /* pretty straight forward                   */
   /*********************************************/
   if ((modeInfo.XResolution==x_res)
     &&(modeInfo.YResolution==y_res)
     &&(modeInfo.BitsPerPixel==color_bits)
     &&((modeInfo.Attributes&special_flags)==special_flags))
   {
    if (bGblDisplayVerbose==TRUE)
     printf("\nMode:%04hx, X:%4hi, Y:%4hi, BITS:%2hi, Attrib:%04X",
            svdc->modeList[i],
            modeInfo.XResolution,
            modeInfo.YResolution,
            modeInfo.BitsPerPixel,
            modeInfo.Attributes);
    iScreenXres = modeInfo.XResolution;
    iScreenYres = modeInfo.YResolution;
    iScreenZres = 1<<modeInfo.BitsPerPixel;
    return svdc->modeList[i];
   }
  }
  i++;
 }
 return 0;
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOInit()                                                 */
/*                                                                           */
/*   Author      : original K.W now C.S.                                     */
/*                                                                           */
/*   Description : This function sets the screen mode and any other IO module*/
/*                 variables such as cursor position. It also sets the       */
/*                 screen pallette.                                          */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
BOOL fIOInit(void)
{
 int   loop;
 UWORD mode=0;
 ULONG screenmem;

/************************/
/* Sanity check on mode */
/************************/
 if(wGblDisplayMode>=RESOLUTION_COUNT)
 {
  fFatal("fIOinit() - Invalid display mode chosen");
 }

/****************/
/* INIT SVGALIB */
/****************/
 svdc=SV_init(TRUE);
 if (svdc==NULL)
 {
  printf("\nNo Vesa found (SVGALIB)!");
  fFatal("fIOinit() - SV_INIT initialization failed!!!");
 }
 else
 {
  if (bGblDisplayVerbose==TRUE)
   printf("\nVBE Version %i.%i found!",svdc->VBEVersion/256,svdc->VBEVersion-(svdc->VBEVersion/256)*256);
 }

 /*******************************/
 /* Look for suitable Vesa Mode */
 /*******************************/
 /***************************************************/
 /* first check for Linear Framebuffer (VBE >= 2.0) */
 /***************************************************/
 if ((svdc->VBEVersion/256>=2)&&(bGblForceVesa12==FALSE))
 {
  switch (wGblDisplayMode)
  {
   case SVGA640X480:
   {
    framebuffer=FB_LINEAR;
    mode=get_mode_on_specs(640,480,8,svHaveLinearBuffer);
    break;
   }
   case SVGA800X600:
   {
    framebuffer=FB_LINEAR;
    mode=get_mode_on_specs(800,600,8,svHaveLinearBuffer);
    break;
   }
   case SVGA1024X768:
   {
    framebuffer=FB_LINEAR;
    mode=get_mode_on_specs(1024,768,8,svHaveLinearBuffer);
    break;
   }
   case SVGA1280X1024:
   {
    framebuffer=FB_LINEAR;
    mode=get_mode_on_specs(1280,1024,8,svHaveLinearBuffer);
    break;
   }
   case SVGA1600X1200:
   {
    framebuffer=FB_LINEAR;
    mode=get_mode_on_specs(1600,1200,8,svHaveLinearBuffer);
    break;
   }
  }
 }
 if (mode==0)
 {
  if (bGblForceVesa12==FALSE)
  {
   if (bGblDisplayVerbose==TRUE)
    printf("\nNo Linear Framebuffer found, trying banked Framebuffer!");
  }
 }
 else
 {
  if (bGblDisplayVerbose==TRUE)
   printf("\nLinear Framebuffer Video Mode found: %04hx!\n",mode);
  if (SV_setMode(mode|svLinearBuffer,FALSE,FALSE,0)==FALSE)
  {
   if (bGblDisplayVerbose==TRUE)
   {
    printf("\nCouldn't set Linear Framebuffer Video Mode (%04hx), trying banked now!\n",mode);
    framebuffer=FB_NONE;
    mode=0;
   }
  }
 }
 /*********************************************/
 /* check for Banked Framebuffer (VBE >= 1.0) */
 /*********************************************/
 if (mode==0) /* only if no mode found yet */
 {
  switch (wGblDisplayMode)
  {
   case SVGA640X480:
   {
    mode=get_mode_on_specs(640,480,8,0);
    framebuffer=FB_BANKED;
    break;
   }
   case SVGA800X600:
   {
    mode=get_mode_on_specs(800,600,8,0);
    framebuffer=FB_BANKED;
    break;
   }
   case SVGA1024X768:
   {
    mode=get_mode_on_specs(1024,768,8,0);
    framebuffer=FB_BANKED;
    break;
   }
   case SVGA1280X1024:
   {
    framebuffer=FB_BANKED;
    mode=get_mode_on_specs(1280,1024,8,0);
    break;
   }
   case SVGA1600X1200:
   {
    framebuffer=FB_BANKED;
    mode=get_mode_on_specs(1600,1200,8,0);
    break;
   }
  }
 }
 if (mode==0)
 {
  fFatal("fIOinit() - corresponding VESA mode not found!!!");
 }
 else if (framebuffer==FB_BANKED)
 {
  if (bGblDisplayVerbose==TRUE)
  {
   printf("\nBanked Framebuffer Video Mode found: %04hx!\n",mode);
  }
  if (SV_setMode(mode,FALSE,FALSE,0)==FALSE)
  {
   fFatal("fIOinit() - Failed to set VESA Banked Framebuffer Video mode %03x",mode);
  }
 }
 if (framebuffer==FB_BANKED)
 {
  vesa_init_modus_infos(mode);
  fIOPlotPixel=fIOPlotPixelbank;
  fIODrawLine=fIODrawLinebank;
  fIODrawMaskLine=fIODrawMaskLinebank;
  fIOcopyvesascreen=fIOcopyvesascreenbank;
  fIOAreaLoad=fIOAreaLoadbank;
  fIOAreaSave=fIOAreaSavebank;
  fIOInitMouse=fIOInitMousebank;
  fIOGetPixel=fIOGetPixelbank;
 }
 else if (framebuffer==FB_LINEAR)
 {
  fIOPlotPixel=fIOPlotPixellin;
  fIODrawLine=fIODrawLinelin;
  fIODrawMaskLine=fIODrawMaskLinelin;
  fIOcopyvesascreen=fIOcopyvesascreenlin;
  fIOAreaLoad=fIOAreaLoadlin;
  fIOAreaSave=fIOAreaSavelin;
  fIOInitMouse=fIOInitMouselin;
  fIOGetPixel=fIOGetPixellin;
 }
 else
 {
  fFatal("fIOinit() - Unkown VESA initialization ERROR!");
 }

/********************************************/
/* Set default Clipping regions (old style) */
/********************************************/
 iClippingX1=0;
 iClippingY1=0;
 iClippingX2=iScreenXres-1;
 iClippingY2=iScreenYres-1;

/*******************************************/
/* Setup the array of line start positions */
/* 65536, since a negative line positions  */
/* possible (slight speedup)               */
/*******************************************/
 for (loop=0;loop<MAX_SCREEN_LINES;loop++)
 {
  if (loop < iScreenYres)
   linestartpos[(UWORD)loop]=loop*iScreenXres;
  else if (loop < iScreenYres+10000)
  {
   linestartpos[(UWORD)loop]=(iScreenYres-1)*iScreenXres;
  }
  else
   linestartpos[(UWORD)loop]=0;
 }

/***************************************************/
/* Setup the colour map that we will use (default) */
/***************************************************/
 fIOsetpalette();

/*******************************/
/* Create overlay memory block */
/*******************************/
 screenmem=(ULONG)iScreenXres*(ULONG)iScreenYres;
 if ((overlay_memory=malloc(screenmem))==NULL)
  fFatal("fIOInit() - Display overlay memory allocation failure (%ld bytes)",screenmem);
 memset(overlay_memory,wGblDisplayBaseColour<<4,screenmem);

/**********************************************/
/* Create Clipping memory (0 no clip, 1 clip) */
/* slight speed increase over checking        */
/* coordinates and needed for PCX Overlays    */
/**********************************************/
 if ((clip_memory=malloc(screenmem))==NULL)
  fFatal("fIOInit() - Display clip allocation failure (%ld bytes)",screenmem);
 memset(clip_memory,1,screenmem); /* clipped all */

/*****************/
/* Init complete */
/*****************/
 initialised=TRUE;
 return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIORestore()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Restore the IO system back to the condition it started in */
/*                 by putting the screen back and releasing any borrowed irq */
/*                 functions.                                                */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE=OK FALSE=ERROR                               */
/*                                                                           */
/*****************************************************************************/

BOOL fIORestore(void)
{
 struct SREGS sregs;
 union REGS inregs, outregs;
 segread(&sregs);

 if (bGblLightpenEnable==TRUE)
 {
  /**************************************************/
  /* Mouse reset, doesn't matter it's there or not! */
  /**************************************************/
  inregs.w.ax = 0;
  int386( 0x33, &inregs, &outregs );
 }

 if (overlay_memory)
  free(overlay_memory);
 if (clip_memory)
  free(clip_memory);

 if(initialised)
 {
  SV_restoreMode();
 }
 initialised=FALSE;
 return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOOutChar()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Output a single ASCII character to the graphics screen.   */
/*                 All control codes are ignored.                            */
/*                                                                           */
/*   Accepts     : Position of char,colour,size and char itself              */
/*                                                                           */
/*   Returns     : PIXEL position                                            */
/*                                                                           */
/*****************************************************************************/
/* Chris: This sends shivers down my spine... but speedup is not neccessary. */
/*****************************************************************************/
PIXEL fIOOutChar(SWORD iXpos,SWORD iYpos,SWORD iColour,SWORD iBackground,SWORD iFontSize,UBYTE cData)
{
 PIXEL tsPosition;
 SWORD iXloop,iYloop,iXmag,iYmag;
 SWORD iXHome,iYHome,iPixCol;
 UBYTE tBitMap;

 if(iFontSize<FONT_MIN_SIZE || iFontSize>FONT_MAX_SIZE)
 {
  iFontSize=FONT_NORMAL;
 }
 /* Draw the given char fron the bitmap in FONT.H */
 iXHome=iXpos;
 iYHome=iYpos;
 for(iYloop=0;iYloop<BASE_FONT_HEIGHT;iYloop++)
 {
  for(iYmag=0;iYmag<iFontSize;iYmag++)
  {
   /* Fetch a line of the char */
   tBitMap=atFontDef[cData&0x7f][iYloop];
   iXpos=iXHome;

   for (iXloop=BASE_FONT_WIDTH-1;iXloop>=0;iXloop--)
   {
    for(iXmag=0;iXmag<iFontSize;iXmag++)
    {
     iPixCol=((tBitMap>>iXloop)&0x01)==0 ? iBackground : iColour;
     fIOPlotPixel(iXpos++,iYpos,iPixCol);
    }
   }
   iYpos++;
  }
 }
 /* Fetch code and return the new position */
 tsPosition.iXpos=iXpos;
 tsPosition.iYpos=iYHome;
 tsPosition.iColour=iColour;
 return(tsPosition);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOOutText()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Print an unformatted text string to the screen.           */
/*                                                                           */
/*   Accepts     : PIXEL giving position and colour of text                  */
/*               : char *pointer to terminated string                        */
/*                                                                           */
/*   Returns     : PIXEL giving current cursor position after printing       */
/*                                                                           */
/*****************************************************************************/

PIXEL fIOOutText(SWORD iXpos,SWORD iYpos,SWORD iColour,SWORD iBackground,SWORD iFontSize,UBYTE *pcTextBuf)
{
 PIXEL           tsPosition;

 if(iFontSize<FONT_MIN_SIZE || iFontSize>FONT_MAX_SIZE)
 {
  iFontSize=FONT_NORMAL;
 }

 /*************************************/
 /* Set up the pixel for the 1st call */
 /*************************************/
 tsPosition.iXpos=iXpos;
 tsPosition.iYpos=iYpos;
 tsPosition.iColour=iColour;

 /*******************************************************/
 /* Blat the text out a pixel at a time in a while loop */
 /*******************************************************/
 while(*pcTextBuf!='\0')
 {
  switch(*pcTextBuf)
  {
   case '\n':
    tsPosition.iYpos+=BASE_FONT_HEIGHT*iFontSize;
    tsPosition.iXpos=iXpos;
    break;
   case '\b':
    tsPosition.iXpos-=BASE_FONT_WIDTH*iFontSize;
    break;
   case '\t':
    tsPosition=fIOOutText(tsPosition.iXpos,tsPosition.iYpos,tsPosition.iColour,iBackground,iFontSize,"    ");
    break;
   default:
    tsPosition=fIOOutChar(tsPosition.iXpos,tsPosition.iYpos,tsPosition.iColour,iBackground,iFontSize,*pcTextBuf);
    break;
  }
  pcTextBuf++;
 }
 return(tsPosition);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOPrintf                                                 */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Print a text string with formatting to the screen.        */
/*                                                                           */
/*   Accepts     : PIXEL giving position and colour of text                  */
/*               : char *pointer to buffer, variable argument list           */
/*                                                                           */
/*   Returns     : PIXEL giving current cursor position after printing       */
/*                                                                           */
/*****************************************************************************/

PIXEL fIOPrintf(SWORD iXpos,SWORD iYpos,SWORD iColour,SWORD iBackground,SWORD iFontSize,UBYTE *pcTextFormat,...)
{
 PIXEL       tPosition;
 UBYTE       acTextBuffer[IO_PRINT_BUF];
 va_list     tGlop;

 /*********************************************************/
 /* Expand the varable argument list into the text buffer */
 /*********************************************************/
 va_start(tGlop,pcTextFormat);                          /* Initialise the variable array */
 if(vsprintf(acTextBuffer,pcTextFormat,tGlop)==-1)       /* Expand the args to a string */
 {
  fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_LARGE,"fIOPrintf: Text Buffer Overflow");
 }
 va_end(tGlop);
 tPosition=fIOOutText(iXpos,iYpos,iColour,iBackground,iFontSize,acTextBuffer);
 return(tPosition);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOGetLine()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function reads a line from the keyboard terminated   */
/*                 by a \n char. The only editing keys that are handled are  */
/*                 backspace and delete which has the same function. On exit */
/*                 the string is terminated correctly. When the max line     */
/*                 length is exceeded the string is automatically terminated */
/*                                                                           */
/*   Accepts     : PIXEL giving position and colour of text                  */
/*               : char * pointer to buffer, BYTE max buffer length          */
/*                                                                           */
/*   Returns     : PIXEL giving current cursor position after data entry     */
/*                                                                           */
/*****************************************************************************/

PIXEL fIOGetLine(SWORD iXpos,SWORD iYpos,SWORD iColour,SWORD iBackground,SWORD iFontSize,UBYTE *pcTextBuf,UBYTE tMaxLength)
{
 static BOOL  bInsertMode=TRUE;
 static UBYTE acOldBuffer[256]="";

 UWORD  tInchar=0;
 UBYTE  tLength=0;
 SWORD  iRecallPointer=0;
 UBYTE  *pcBufferBase;
 BOOL   bCopyFromOldBuffer=FALSE;

 PIXEL  tsCursorPos;

 /*********************************/
 /* Initialise the PIXEL position */
 /*********************************/

 tsCursorPos.iXpos=iXpos;
 tsCursorPos.iYpos=iYpos;
 tsCursorPos.iColour=iColour;

 /**********************************************************************/
 /* Termintate the string before we start so editing can work properly */
 /**********************************************************************/

 pcBufferBase=pcTextBuf;
 *pcTextBuf=0;

 /********************************/
 /* Do a while loop to get chars */
 /********************************/

 while (tLength<tMaxLength)
 {
  /***************************************************************/
  /* If we are copying from the old buffer then dont read  a key */
  /***************************************************************/
  if(bCopyFromOldBuffer==TRUE)
  {
   tInchar=KEY_RIGHT;
  }
  else
  {
   /***************************************/
   /* Terminate the loop on the enter key */
   /***************************************/
   if ((tInchar=fKBDgetch())==KEY_ENTER)
    break;

   /*******************************/
   /* Process Extended characters */
   /*******************************/
   if (tInchar==0)
   {
    tInchar=0x80+fKBDgetch();
   }
  }

  /***************************************/
  /* Process the input for control chars */
  /***************************************/
  switch(tInchar)
  {
   case KEY_F3:
   case KEY_UP:
   {
    bCopyFromOldBuffer=TRUE;
    break;
   }
   case KEY_INS:
   {
    if(bInsertMode==FALSE) bInsertMode=TRUE; else bInsertMode=FALSE;
    break;
   }
   case KEY_BSPACE:
   case KEY_DEL:
   case KEY_LEFT:
   {
    if(tLength>0)
    {
     tLength--;
     pcTextBuf--;
     *pcTextBuf=0;
     if(iRecallPointer>0) iRecallPointer--;
     tsCursorPos=fIOOutText(iXpos,iYpos,iColour,iBackground,iFontSize,"\b \b");
    }
    break;
   }
   case KEY_RIGHT:
   {
    /************************************************************************/
    /* Fudge the new key from the old buffer run on into the default output */
    /************************************************************************/
    if (acOldBuffer[iRecallPointer]!=0)
    {
     tInchar=(UWORD)acOldBuffer[iRecallPointer++];
    }
    else
    {
     bCopyFromOldBuffer=FALSE;
     break;
    }
   }
   default:
   {
    *(pcTextBuf++)=(UBYTE)tInchar;
    *pcTextBuf=0;
    tsCursorPos=fIOOutChar(iXpos,iYpos,iColour,iBackground,iFontSize,(UBYTE)tInchar);
    tLength++;
    break;
   }
  }
  /*********************************************/
  /* Load the new cursor position from the OLD */
  /*********************************************/
  iXpos=tsCursorPos.iXpos;
  iYpos=tsCursorPos.iYpos;
  iColour=tsCursorPos.iColour;
 }

 /*******************************************/
 /* Save the edited line in the copy buffer */
 /*******************************************/
 if (strlen(pcBufferBase)<256)
  strcpy(acOldBuffer,pcBufferBase);

 /************/
 /* All done */
 /************/

 return(tsCursorPos);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOSetClipping()                                          */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Set the clipping rectangle for all printing,plotting and  */
/*                 drawing operations.                                       */
/*                                                                           */
/*   Accepts     : int X and Y Corner window positions                       */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fIOUnSetClipBox(int iXpos,int iYpos,int iXsize,int iYsize)
{
 int i;
  for (i=iYpos;i<=iYsize;i++)
   memset(clip_memory+linestartpos[(UWORD)i]+iXpos,0,iXsize-iXpos);
}
void fIOSetClipBox(int iXpos,int iYpos,int iXsize,int iYsize)
{
 int i;
  for (i=iYpos;i<=iYsize;i++)
   memset(clip_memory+linestartpos[(UWORD)i]+iXpos,1,iXsize-iXpos);
}
void fIOSetClipping(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2)
{
 int i;

 /******************************************/
 /* Do a sanity check and make sure X1,Y1  */
 /* is upper left and X2,Y2 is lower right */
 /******************************************/
 if(iX1>iX2)
  fSwap(&iX1,&iX2);
 if(iY1>iY2)
  fSwap(&iY1,&iY2);

//printf("Clipping X1=%i, Y1=%i, X2=%i,Y2=%i\n",iX1,iY1,iX2,iY2);

 /*********************************************************/
 /* If any parameters are -1 then set MAX clipping window */
 /*********************************************************/
 if (rectengle_clipping_enable==TRUE)
 {
  /******************/
  /* First Clip all */
  /******************/
  for (i=0;i<iScreenYres;i++)
   memset(clip_memory+linestartpos[(UWORD)i],1,iScreenXres);
  if ((iY1==-1)||(iY2==-1)||(iX1==-1)||(iX2==-1))
  {
   iX1=0;
   iX2=iScreenXres-1;
   iY1=0;
   iY2=iScreenYres-1;
  }

  /****************************************/
  /* Then set region which is 'unclipped' */
  /****************************************/
  for (i=iY1;i<iY2+1;i++)
   memset(clip_memory+linestartpos[(UWORD)i]+iX1,0,iX2-iX1);
 }

 if (iX1<0)               iX1=0;
 if (iX1>iScreenXres-1)   iX1=iScreenXres-1;

 if (iY1<0)               iY1=0;
 if (iY1>iScreenYres-1)   iY1=iScreenYres-1;

 if (iX2<0)               iX2=iScreenXres-1;
 if (iX2>iScreenXres-1)   iX2=iScreenXres-1;

 if (iY2<0)               iY2=iScreenYres-1;
 if (iY2>iScreenYres-1)   iY2=iScreenYres-1;

 iClippingX1=iX1;
 iClippingY1=iY1;
 iClippingX2=iX2;
 iClippingY2=iY2;
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOSetCursorPos()                                         */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Move the graphics cursor to the given position.           */
/*                                                                           */
/*   Accepts     : int iX iY                                                 */
/*                                                                           */
/*   Returns     : PIXEL struct with position                                */
/*                                                                           */
/*****************************************************************************/

PIXEL fIOSetCursorPos(SWORD iXpos,SWORD iYpos,SWORD iColour)
{
 PIXEL tsPosition;

 tsPosition.iXpos=iXpos;
 tsPosition.iYpos=iYpos;
 tsPosition.iColour=iColour;

 return(tsPosition);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOGetScreenRes()                                         */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Returns the current screen X,Y and Z resolution in a PIXEL*/
/*                 structure.                                                */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : PIXEL with X,Y and colour resolutions                     */
/*                                                                           */
/*****************************************************************************/
PIXEL fIOGetScreenRes(void)
{
 PIXEL   tsResolution;

 tsResolution.iXpos=iScreenXres;
 tsResolution.iYpos=iScreenYres;
 tsResolution.iColour=iScreenZres;

 return(tsResolution);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOGetOverlayBuffer()                                     */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Returns address of overlay buffer                         */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : ptr to overlay buffer                                     */
/*                                                                           */
/*****************************************************************************/

UBYTE* fIOGetOverlayBuffer(void)
{
 return(overlay_memory);
}
/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOGetClipBuffer()                                        */
/*                                                                           */
/*   Author      : C.S                                                       */
/*                                                                           */
/*   Description : Returns address of overlay buffer                         */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : ptr to overlay buffer                                     */
/*                                                                           */
/*****************************************************************************/
UBYTE* fIOGetClipBuffer(void)
{
 return(clip_memory);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIODrawBox()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Draws either a filled or a box outline using the          */
/*                                                                           */
/*   Accepts     : PIXELs given box corners and colour                       */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void fIODrawBox(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2,SWORD iColour,BOOL tSolid)
{
 UWORD yloop;

 /******************************************/
 /* Do a sanity check and make sure X1,Y1  */
 /* is upper left and X2,Y2 is lower right */
 /******************************************/
 if(iX1>iX2)
  fSwap(&iX1,&iX2);
 if(iY1>iY2)
  fSwap(&iY1,&iY2);

 /**********************/
 /* Set the BOX colour */
 /**********************/
 /*******************************/
 /* Draw filled or unfilled box */
 /*******************************/
 if (tSolid)
 {
  for(yloop=iY1;yloop<=iY2;yloop++)
   fIODrawLine(iX1,yloop,iX2,yloop,iColour);
 }
 else
 {
  fIODrawLine(iX1,iY1,iX2,iY1,iColour);
  fIODrawLine(iX1,iY2,iX2,iY2,iColour);
  fIODrawLine(iX1,iY1,iX1,iY2,iColour);
  fIODrawLine(iX2,iY1,iX2,iY2,iColour);
 }
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOAreaMove()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Reads a block from one position and moves it to another   */
/*                                                                           */
/*   Accepts     : PIXEL pointing to source and dest positions               */
/*                                                                           */
/*   Returns     : BOOLEAN to say if area was moved sucessfully              */
/*                                                                           */
/*****************************************************************************/

BOOL fIOAreaMove(SWORD iXsrc1,SWORD iYsrc1,SWORD iXsrc2,SWORD iYsrc2,SWORD iXdest,SWORD iYdest)
{
 BITBLOCKPTR tSaved;
 BOOL bResult;

 /********************************************/
 /* Do a sanity check and make sure X1,Y1 is */
 /* upper left and X2,Y2 is lower right      */
 /********************************************/
 if(iXsrc1>iXsrc2)
  fSwap(&iXsrc1,&iXsrc2);
 if(iYsrc1>iYsrc2)
  fSwap(&iYsrc1,&iYsrc2);

 /****************************************/
 /* Step1 Get the image area to be moved */
 /****************************************/

 if ((tSaved=fIOAreaSave(iXsrc1,iYsrc1,iXsrc2,iYsrc2))==NULL)
  return(FALSE);

 /***********************************************************/
 /* Step 2 put it to the destination and return sucess/fail */
 /***********************************************************/
 bResult=fIOAreaLoad(iXdest,iYdest,tSaved);

 /******************************************/
 /* Step 3 Release the bitblock we created */
 /******************************************/
 fIOBitBlockRelease(tSaved);

 return(bResult);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOAreaScroll()                                           */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Move then clear an area using the predefined primatives   */
/*                                                                           */
/*   Accepts     : PIXEL source boundaries, PIXEL dest                       */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE=OK FALSE = ERROR OCCURED                     */
/*                                                                           */
/*****************************************************************************/
BOOL fIOAreaScroll(SWORD iXsrc1,SWORD iYsrc1,SWORD iXsrc2,SWORD iYsrc2,SWORD iXdest,SWORD iYdest,SWORD iBackgroundColour)
{
 BITBLOCKPTR tArea;
 BOOL     bRetVal=TRUE;

 SWORD         iEraseX1,iEraseY1;
 SWORD         iEraseX2,iEraseY2;

 if (iXsrc1==iXdest && iYsrc1==iYdest)
  return(TRUE);

 /*****************************/
 /* Step 1 Save move the area */
 /*****************************/
 if ((tArea=fIOAreaSave(iXsrc1,iYsrc1,iXsrc2,iYsrc2))==NULL)
  return(FALSE);

 if (iXsrc1!=iXdest && iYsrc1!=iYdest)
 {
  /****************************************************************************/
  /* Multiaxis move so erase then re-draw to save any more programming effort */
  /****************************************************************************/
  fIOAreaClear(iXsrc1,iYsrc1,iXsrc2,iYsrc2,iBackgroundColour);

  /************************************************/
  /* Put the save block down into its destination */
  /************************************************/
  bRetVal&=fIOAreaLoad(iXdest,iYdest,tArea);

  /***********************************/
  /* Release the bitblock we created */
  /***********************************/
  fIOBitBlockRelease(tArea);
 }
 else
 {
  /************************************************/
  /* Put the save block down into its destination */
  /************************************************/
  bRetVal&=fIOAreaLoad(iXdest,iYdest,tArea);

  /***********************************/
  /* Release the bitblock we created */
  /***********************************/
  fIOBitBlockRelease(tArea);

  if (iYsrc1==iYdest)
  {
   /**********************/
   /* X Axis Only scroll */
   /**********************/
   if(iXdest>iXsrc1)
   {
    iEraseX1=iXsrc1;
    iEraseX2=iXdest;
   }
   else
   {
    iEraseX1=iXdest+(iXsrc2-iXsrc1);
    iEraseX2=iXsrc2;
   }
   fIOAreaClear(iEraseX1,iYsrc1,iEraseX2,iYsrc2,iBackgroundColour);
  }
  else if(iXsrc1==iXdest)
  {
   /**********************/
   /* Y Axis Only scroll */
   /**********************/
   if (iYdest>iYsrc1)
   {
    iEraseY1=iYsrc1;
    iEraseY2=iYdest;
   }
   else
   {
    iEraseY1=iYdest+(iYsrc2-iYsrc1);
    iEraseY2=iYsrc2;
   }
   fIOAreaClear(iXsrc1,iEraseY1,iXsrc2,iEraseY2,iBackgroundColour);
  }
 }
 /********************************************/
 /* Step 5, ther is no step 5 we've finished */
 /********************************************/
 return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOAreaClear()                                            */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Clear an area back to the backgrounf colour.              */
/*                                                                           */
/*   Accepts     : PIXELs given box corners and colour                       */
/*                                                                           */
/*   Returns     : BOOLEAN to say if area was cleared or not                 */
/*                                                                           */
/*****************************************************************************/
void fIOAreaClear(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2,SWORD iBackgroundColour)
{
 if(iX1==-1) iX1=0;
 if(iX2==-1) iX2=iScreenXres-1;
 if(iY1==-1) iY1=0;
 if(iY2==-1) iY2=iScreenYres-1;

 fIODrawBox(iX1,iY1,iX2,iY2,iBackgroundColour,BOX_FILLED);
 return;
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOBitBlockRelease()                                      */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Free the memory allocated to a bit block.                 */
/*                                                                           */
/*   Accepts     : Pointer to a bitblock                                     */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void fIOBitBlockRelease(BITBLOCKPTR tBlock)
{
 /******************************/
 /* Check its a real block 1st */
 /******************************/

 if(tBlock->tBitblockMagicNumber1 != tBlock->tBitblockMagicNumber2)
 {
  fFatal("I/O ERROR: fIOBitBlockRelease() Bad Block Handle, aborting load");
 }
 else
 {
  free(tBlock->tBlockData);
  free(tBlock);
 }
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fGFXsetInk()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Change pen colour                                         */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : PIXEL struct                                              */
/*                                                                           */
/*****************************************************************************/
void fIOsetink(UWORD iInk,UBYTE ubRed,UBYTE ubGreen,UBYTE ubBlue)
{
 palette[iInk][PRED]=ubRed;
 palette[iInk][PGREEN]=ubGreen;
 palette[iInk][PBLUE]=ubBlue;
 return;
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fGFXsetpalette()                                          */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Wham down a new pallete during the vertical re-trace      */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : PIXEL struct                                              */
/*                                                                           */
/*****************************************************************************/
void fIOsetpalette(void)
{
 SV_setPalette(0,256,(SV_palette *)&palette[0][0],-1);
 return;
}

/*************************************************************************/




/************************************************************************/
/* with all following functions I do NOT CHECK if iX1<0 this shouldn't  */
/* matter as long as iY1 is not 0                                       */
/*                                                                      */
/* the same goes for iX2>iScreenResX and iY2=iScreenResY                */
/*                                                                      */
/* The point of not checking is speed.                                  */
/* If the Scaling factors are set correctly nothing bad should come out */
/* of it, but if one tends to fill the entire screen with vectrex       */
/* or I dunno what... If the above should occure a PAGE Fault           */
/* WILL come down on you!!!                                             */
/************************************************************************/




/*****************************/
/* from here linear funtions */
/*****************************/

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOcopyvesascreen()                                       */
/*                                                                           */
/*   Author      : K.W from original code by P.Jones (fil@muon.demon.co.uk)  */
/*                                                                           */
/*   Description : Copys a screen buffer to the active screen area           */
/*                                                                           */
/*   Accepts     : Pointer to source buffer size, width & depth              */
/*                                                                           */
/*   Returns     : nowt                                                      */
/*                                                                           */
/*****************************************************************************/

void fIOcopyvesascreenlin(UBYTE *source,SLONG swidth,SLONG sheight,SLONG bpp)
{
 memcpy(svdc->videoMem,(void *)source,iScreenXres*iScreenYres);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOPlotPixel()                                            */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Plot a Pixel to the screen, pixels are checked for screen */
/*                 boundaries before plotting.                               */
/*                                                                           */
/*   Accepts     : PIXEL giving position and colour                          */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fIOPlotPixellin(SWORD iXpos,SWORD iYpos,SWORD iColour)
{
// if (*(clip_memory+linestartpos[(UWORD)iYpos]+(ULONG)iXpos))
//  return;
 /*******************************************/
 /* Calulate byte position + Write the byte */
 /*******************************************/
 *(((UBYTE*)svdc->videoMem)+linestartpos[(UWORD)iYpos]+(ULONG)iXpos)=(UBYTE)iColour;
}
UBYTE fIOGetPixellin(SWORD iXpos,SWORD iYpos)
{
 /*******************************************/
 /* Calulate byte position + Read the byte */
 /*******************************************/
 return (*(((UBYTE*)svdc->videoMem)+linestartpos[(UWORD)iYpos]+(ULONG)iXpos));
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIODrawLine()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Draws a line on the screen clipped to the screen boundary */
/*                 and in the given colour.                                  */
/*                                                                           */
/*   Accepts     : PIXELs given start and end position of line and colour    */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fIODrawLinelin(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2,SWORD iColour)
{
 UBYTE *dest=(UBYTE*)svdc->videoMem;
 SWORD dx=0,dy=0,err=0,inc=0;

/**************************/
/* Do straight line cases */
/**************************/
 if(iX1==iX2)
 {
  if(iY1>iY2)
  {
   fSwap(&iY1,&iY2);
  }
  do
  {
   /******************************************************************************************/
   /* not to be put out of loop, offset can get to high... (high iY1 are pointing to maxline)*/
   /******************************************************************************************/
   offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
//   if (!(*(clip_memory+offset)))
   {
    *(dest+offset)=iColour;
   }
  }
  while(iY1++!=iY2);
 }
 else if(iY1==iY2)
 {
  if(iX1>iX2)
  {
   fSwap(&iX1,&iX2);
  }
  offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
  do
  {
//   if (!(*(clip_memory+offset)))
   {
    *(dest+offset)=iColour;
   }
   offset++;
  }
  while(iX1++!=iX2);
 }
 else
 {
  /*************/
  /* X Variant */
  /*************/
  dx=iX2-iX1;
  dy=iY2-iY1;
  if(abs(dx)>=abs(dy))
  {
   if(iX1>iX2)
   {
    fSwap(&iX1,&iX2);
    fSwap(&iY1,&iY2);
   }
   dx=iX2-iX1;
   dy=iY2-iY1;
   err=-(dx>>1);
   if(dy>0)
   {
    inc=1;
   }
   else
   {
    inc=-1;
    dy=-dy;
   }
   do
   {
    /**************************/
    /* Calulate byte position */
    /**************************/
    offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
//    if (!(*(clip_memory+offset)))
    {
     *(dest+offset)=iColour;
    }
    /**********************/
    /* Check error status */
    /**********************/
    err+=dy;
    if(err>0)
    {
     err-=dx;
     iY1+=inc;
    }
   }
   while(iX1++!=iX2);
  }
  else
  {
   if(iY1>iY2)
   {
    fSwap(&iX1,&iX2);
    fSwap(&iY1,&iY2);
   }
   dx=iX2-iX1;
   dy=iY2-iY1;
   err=-(dy>>1);
   if(dx>0)
   {
    inc=1;
   }
   else
   {
    inc=-1;
    dx=-dx;
   }
   do
   {
    /**************************/
    /* Calulate byte position */
    /**************************/
    offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
//    if (!(*(clip_memory+offset)))
    {
     *(dest+offset)=iColour;
    }
    /**********************/
    /* Check error status */
    /**********************/
    err+=dx;
    if(err>=0)
    {
     err-=dy;
     iX1+=inc;
    }
   }
   while(iY1++!=iY2);
  }
 }
}
/*****************************************************************************/
/*                                                                           */
/*   Function    : fIODrawMaskLine()                                         */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Draws a line on the screen clipped to the screen boundary */
/*                 and in the given colour BUT ALL screen writes are OR'd    */
/*                 with the overlay memory BEFORE being written.             */
/*                                                                           */
/*   Accepts     : PIXELs given start and end position of line and colour    */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void fIODrawMaskLinelin(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2,SWORD iColour)
{
 SWORD   dx=0,dy=0,err=0,inc=0;
 UBYTE   *dest=(UBYTE*)svdc->videoMem;
 UBYTE   data;
 /**************************/
 /* Do straight line cases */
 /**************************/
 if(iX1==iX2)
 {
  if(iY1>iY2)
  {
   fSwap(&iY1,&iY2);
  }
  do
  {
   /******************************************************************************************/
   /* not to be put out of loop, offset can get to high... (high iY1 are pointing to maxline)*/
   /******************************************************************************************/
   offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
   if (!(*(clip_memory+offset)))
   {
    data=(UBYTE)iColour | *(overlay_memory+offset);
    *(dest+offset)=data;
   }
  }
  while(iY1++!=iY2);
 }
 else if (iY1==iY2)
 {
  if (iX1>iX2)
  {
   fSwap(&iX1,&iX2);
  }
  offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
  do
  {
   if (!(*(clip_memory+offset)))
   {
    data=(UBYTE)iColour | *(overlay_memory+offset);
    *(dest+offset)=data;
   }
   offset++;
  }
  while(iX1++!=iX2);
 }
 else
 {
  /*************/
  /* X Variant */
  /*************/
  dx=iX2-iX1;
  dy=iY2-iY1;

  if (abs(dx)>=abs(dy))
  {
   if (iX1>iX2)
   {
    fSwap(&iX1,&iX2);
    fSwap(&iY1,&iY2);
   }
   dx=iX2-iX1;
   dy=iY2-iY1;

   err=-(dx>>1);
   if (dy>0)
   {
    inc=1;
   }
   else
   {
    inc=-1;
    dy=-dy;
   }
   do
   {
    /**************************/
    /* Calulate byte position */
    /**************************/
    offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
    if (!(*(clip_memory+offset)))
    {
     data=(UBYTE)iColour | *(overlay_memory+offset);
     *(dest+offset)=data;
    }
    /**********************/
    /* Check error status */
    /**********************/
    err+=dy;
    if(err>0)
    {
     err-=dx;
     iY1+=inc;
    }
   }
   while(iX1++!=iX2);
  }
  else
  {
   if(iY1>iY2)
   {
    fSwap(&iX1,&iX2);
    fSwap(&iY1,&iY2);
   }
   dx=iX2-iX1;
   dy=iY2-iY1;

   err=-(dy>>1);
   if (dx>0)
   {
    inc=1;
   }
   else
   {
    inc=-1;
    dx=-dx;
   }
   do
   {
    /**************************/
    /* Calulate byte position */
    /**************************/
    offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
    if (!(*(clip_memory+offset)))
    {
     data=(UBYTE)iColour | *(overlay_memory+offset);
     *(dest+offset)=data;
    }
    /**********************/
    /* Check error status */
    /**********************/
    err+=dx;

    if (err>=0)
    {
     err-=dy;
     iX1+=inc;
    }
   }
   while(iY1++!=iY2);
  }
 }
}
/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOAreaSave()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Save a screen area and return a pointer to a bitbloc      */
/*                 structure containing a pointer to the data. The user must */
/*                 use fIOBitBlockRelease() to free an unwanted block.       */
/*                                                                           */
/*                                                                           */
/*   Accepts     : PIXELs giving Area boundaries                             */
/*                                                                           */
/*   Returns     : Pointer to save bitblock area                             */
/*                                                                           */
/*****************************************************************************/

BITBLOCK* fIOAreaSavelin(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2)
{
 BITBLOCKPTR   tBlock=NULL;
 ULONG         lImageSize;
 ULONG         src;
 UBYTE         *base=(UBYTE*)svdc->videoMem;
 UBYTE         *dest;
 UWORD         xloop,yloop;

 /********************************************************************************/
 /* Do a sanity check and make sure X1,Y1 is upper left and X2,Y2 is lower right */
 /********************************************************************************/
 if (iX1>iX2)
  fSwap(&iX1,&iX2);
 if (iY1>iY2)
  fSwap(&iY1,&iY2);

 /************************************************/
 /* Malloc some space for the bitblock structure */
 /************************************************/
 if ((tBlock=malloc(sizeof(BITBLOCK)))==NULL)
 {
  fFatal("I/O ERROR: fIOAreaSave() malloc Failed for BITBLOCK structure");
  return(NULL);
 }

 /*************************************/
 /* Calc the image size and malloc it */
 /*************************************/
 lImageSize=((iX2-iX1)+1)*((iY2-iY1)+1);
 if ((dest=tBlock->tBlockData=(UBYTE*)malloc(lImageSize*sizeof(UBYTE)))==NULL)
 {
  fFatal("I/O ERROR: fIOAreaSave() malloc failed for image");
  free(tBlock);
  return(NULL);
 }

 /*************************************/
 /* Now get the image into the buffer */
 /*************************************/
 tBlock->width=(iX2-iX1)+1;
 tBlock->height=(iY2-iY1)+1;
 src=(iY1*iScreenXres)+iX1;
 for(yloop=iY1;yloop<=iY2;yloop++)
 {
  for(xloop=iX1;xloop<=iX2;xloop++)
  {
   *(dest++)=*(base+(src++));
  }
  src+=iScreenXres-tBlock->width;
 }

 /*************************/
 /* Set the Magic numbers */
 /*************************/
 tBlock->tBitblockMagicNumber1=0xa55aa55al;
 tBlock->tBitblockMagicNumber2=0xa55aa55al;

 return(tBlock);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOAreaLoad()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Loads a saved image into an area, erasing the old area.   */
/*                                                                           */
/*   Accepts     : PIXEL pointing to load position and block pointer         */
/*                                                                           */
/*   Returns     : BOOLEAN to say if area was loaded sucessfully             */
/*                                                                           */
/*****************************************************************************/

BOOL fIOAreaLoadlin(SWORD iXdest,SWORD iYdest,BITBLOCK *tData)
{
 UBYTE         *base=(UBYTE*)svdc->videoMem;
 UBYTE         *src;
 UWORD         xloop,yloop;
 ULONG         dest;

 if(tData->tBitblockMagicNumber1 != tData->tBitblockMagicNumber2)
 {
  fFatal("I/O ERROR: fIOAreaLoad() Bad Block Handle, aborting load");
 }
 else
 {
  /*************************************/
  /* Now get the image into the buffer */
  /*************************************/
  dest=(iYdest*iScreenXres)+iXdest;
  src=tData->tBlockData;
  for(yloop=iYdest;yloop<iYdest+tData->height;yloop++)
  {
   for(xloop=iXdest;xloop<iXdest+tData->width;xloop++)
   {
    /*****************************************/
    /* Check clipping and reject if required */
    /*****************************************/
    if (xloop>=iClippingX1 && xloop<=iClippingX2 && yloop>=iClippingY1 && yloop<=iClippingY2)
    {
     *(base+dest)=*src;
    }
    dest++;
    src++;
   }
   dest+=iScreenXres-(tData->width);
  }
 }
 return(TRUE);
}





/*******************************************************************/
/* Note all banked functions would get a 'great' speed increase if */
/* a check was done for if a new bank is reached within the        */
/* actual bounderies..., than go for it straight on without any    */
/* further checking... or perhaps even draw in parts...            */
/*                                                                 */
/* I didn't do it 'cause I'm lazy and I                            */
/* have a linear framebuffer grafics card...                       */
/*******************************************************************/

/*****************************/
/* from here banked funtions */
/*****************************/
/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOcopyvesascreen()                                       */
/*                                                                           */
/*   Author      : K.W from original code by P.Jones (fil@muon.demon.co.uk)  */
/*                                                                           */
/*   Description : Copys a screen buffer to the active screen area           */
/*                                                                           */
/*   Accepts     : Pointer to source buffer size, width & depth              */
/*                                                                           */
/*   Returns     : nowt                                                      */
/*                                                                           */
/*****************************************************************************/

void fIOcopyvesascreenbank(UBYTE *source,SLONG swidth,SLONG sheight,SLONG bpp)
{
 int i;
 vesa_address_of_window[vesa_write_window]=0;
 vesa_set_write_window();
 for (i=0;i<=((iScreenXres*iScreenYres+1023)/1024)/vesa_granularity;i++)
 {
  memcpy((void *)vesa_write_seg,(void *)source,vesa_granularity*1024);
  source+=vesa_granularity*1024;
  vesa_address_of_window[vesa_write_window]++;
  vesa_set_write_window();
 }
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOPlotPixel()                                            */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Plot a Pixel to the screen, pixels are checked for screen */
/*                 boundaries before plotting.                               */
/*                                                                           */
/*   Accepts     : PIXEL giving position and colour                          */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fIOPlotPixelbank(SWORD iXpos,SWORD iYpos,SWORD iColour)
{
 /*******************************************/
 /* Calulate byte position + Write the byte */
 /*******************************************/
 offset=iXpos+linestartpos[(UWORD)iYpos];
 offset_x=(UWORD)(offset&vesa_and_x);
 if (vesa_address_of_window[vesa_write_window]!=(new_win_address=(UWORD)(offset>>vesa_granularity_potenz)))
 {
  vesa_address_of_window[vesa_write_window]=new_win_address;
  vesa_set_write_window();
 }
 *((UBYTE *)(vesa_write_seg+offset_x))=(UBYTE)iColour;
}

UBYTE fIOGetPixelbank(SWORD iXpos,SWORD iYpos)
{
 /*******************************************/
 /* Calulate byte position + Read the byte */
 /*******************************************/
 offset=iXpos+linestartpos[(UWORD)iYpos];
 offset_x=(UWORD)(offset&vesa_and_x);
 if (vesa_address_of_window[vesa_read_window]!=(new_win_address=(UWORD)(offset>>vesa_granularity_potenz)))
 {
  vesa_address_of_window[vesa_read_window]=new_win_address;
  vesa_set_read_window();
 }
 return (*((UBYTE *)(vesa_read_seg+offset_x)));
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIODrawLine()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Draws a line on the screen clipped to the screen boundary */
/*                 and in the given colour.                                  */
/*                                                                           */
/*   Accepts     : PIXELs given start and end position of line and colour    */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fIODrawLinebank(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2,SWORD iColour)
{
 SWORD   dx=0,dy=0,err=0,inc=0;
 /**************************/
 /* Do straight line cases */
 /**************************/
 if(iX1==iX2)
 {
  if(iY1>iY2)
  {
   fSwap(&iY1,&iY2);
  }
  do
  {
   /******************************************************************************************/
   /* not to be put out of loop, offset can get to high... (high iY1 are pointung to maxline)*/
   /******************************************************************************************/
   offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
//   if (!(*(clip_memory+offset)))
   {
    offset_x=(UWORD)(offset&vesa_and_x);
    if (vesa_address_of_window[vesa_write_window]!=(new_win_address=(UWORD)(offset>>vesa_granularity_potenz)))
    {
     vesa_address_of_window[vesa_write_window]=new_win_address;
     vesa_set_write_window();
    }
    *((UBYTE *)(vesa_write_seg+offset_x))=(UBYTE)iColour;
   }
  }
  while(iY1++!=iY2);
 }
 else if(iY1==iY2)
 {
  if(iX1>iX2)
  {
   fSwap(&iX1,&iX2);
  }
  offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
  do
  {
//   if (!(*(clip_memory+offset)))
   {
    offset_x=(UWORD)(offset&vesa_and_x);
    if (vesa_address_of_window[vesa_write_window]!=(new_win_address=(UWORD)(offset>>vesa_granularity_potenz)))
    {
     vesa_address_of_window[vesa_write_window]=new_win_address;
     vesa_set_write_window();
    }
    *((UBYTE *)(vesa_write_seg+offset_x))=(UBYTE)iColour;
   }
   offset++;
  }
  while(iX1++!=iX2);
 }
 else
 {
  /*************/
  /* X Variant */
  /*************/
  dx=iX2-iX1;
  dy=iY2-iY1;
  if(abs(dx)>=abs(dy))
  {
   if(iX1>iX2)
   {
    fSwap(&iX1,&iX2);
    fSwap(&iY1,&iY2);
   }
   dx=iX2-iX1;
   dy=iY2-iY1;
   err=-(dx>>1);
   if(dy>0)
   {
    inc=1;
   }
   else
   {
    inc=-1;
    dy=-dy;
   }
   do
   {
    /**************************/
    /* Calulate byte position */
    /**************************/
    offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
//    if (!(*(clip_memory+offset)))
    {
     offset_x=(UWORD)(offset&vesa_and_x);
     if (vesa_address_of_window[vesa_write_window]!=(new_win_address=(UWORD)(offset>>vesa_granularity_potenz)))
     {
      vesa_address_of_window[vesa_write_window]=new_win_address;
      vesa_set_write_window();
     }
     *((UBYTE *)(vesa_write_seg+offset_x))=(UBYTE)iColour;
    }
    /**********************/
    /* Check error status */
    /**********************/
    err+=dy;
    if(err>0)
    {
     err-=dx;
     iY1+=inc;
    }
   }
   while(iX1++!=iX2);
  }
  else
  {
   if(iY1>iY2)
   {
    fSwap(&iX1,&iX2);
    fSwap(&iY1,&iY2);
   }
   dx=iX2-iX1;
   dy=iY2-iY1;
   err=-(dy>>1);
   if(dx>0)
   {
    inc=1;
   }
   else
   {
    inc=-1;
    dx=-dx;
   }
   do
   {
    /**************************/
    /* Calulate byte position */
    /**************************/
    offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
//    if (!(*(clip_memory+offset)))
    {
     offset_x=(UWORD)(offset&vesa_and_x);
     if (vesa_address_of_window[vesa_write_window]!=(new_win_address=(UWORD)(offset>>vesa_granularity_potenz)))
     {
      vesa_address_of_window[vesa_write_window]=new_win_address;
      vesa_set_write_window();
     }
     *((UBYTE *)(vesa_write_seg+offset_x))=(UBYTE)iColour;
    }
    /**********************/
    /* Check error status */
    /**********************/
    err+=dx;
    if(err>=0)
    {
     err-=dy;
     iX1+=inc;
    }
   }
   while(iY1++!=iY2);
  }
 }
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIODrawMaskLine()                                         */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Draws a line on the screen clipped to the screen boundary */
/*                 and in the given colour BUT ALL screen writes are OR'd    */
/*                 with the overlay memory BEFORE being written.             */
/*                                                                           */
/*   Accepts     : PIXELs given start and end position of line and colour    */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fIODrawMaskLinebank(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2,SWORD iColour)
{
 SWORD   dx=0,dy=0,err=0,inc=0;
 UBYTE   data;
 /**************************/
 /* Do straight line cases */
 /**************************/
 if(iX1==iX2)
 {
  if(iY1>iY2)
  {
   fSwap(&iY1,&iY2);
  }
  do
  {
   /******************************************************************************************/
   /* not to be put out of loop, offset can get to high... (high iY1 are pointing to maxline)*/
   /******************************************************************************************/
   offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
   if (!(*(clip_memory+offset)))
   {
    data=(UBYTE)iColour | *(overlay_memory+offset);
    offset_x=(UWORD)(offset&vesa_and_x);
    if (vesa_address_of_window[vesa_write_window]!=(new_win_address=(UWORD)(offset>>vesa_granularity_potenz)))
    {
     vesa_address_of_window[vesa_write_window]=new_win_address;
     vesa_set_write_window();
    }
    *((UBYTE *)(vesa_write_seg+offset_x))=(UBYTE)data;
   }
  }
  while(iY1++!=iY2);
 }
 else if(iY1==iY2)
 {
  if(iX1>iX2)
  {
   fSwap(&iX1,&iX2);
  }
  offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
  do
  {
   if (!(*(clip_memory+offset)))
   {
    data=(UBYTE)iColour | *(overlay_memory+offset);
    offset_x=(UWORD)(offset&vesa_and_x);
    if (vesa_address_of_window[vesa_write_window]!=(new_win_address=(UWORD)(offset>>vesa_granularity_potenz)))
    {
     vesa_address_of_window[vesa_write_window]=new_win_address;
     vesa_set_write_window();
    }
    *((UBYTE *)(vesa_write_seg+offset_x))=(UBYTE)data;
   }
   offset++;
  }
  while(iX1++!=iX2);
 }
 else
 {
  /* X Variant */
  dx=iX2-iX1;
  dy=iY2-iY1;

  if(abs(dx)>=abs(dy))
  {
   if(iX1>iX2)
   {
    fSwap(&iX1,&iX2);
    fSwap(&iY1,&iY2);
   }

   dx=iX2-iX1;
   dy=iY2-iY1;

   err=-(dx>>1);

   if(dy>0)
   {
    inc=1;
   }
   else
   {
    inc=-1;
    dy=-dy;
   }
   do
   {
    /**************************/
    /* Calulate byte position */
    /**************************/
    offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
    if (!(*(clip_memory+offset)))
    {
     data=(UBYTE)iColour | *(overlay_memory+offset);
     offset_x=(UWORD)(offset&vesa_and_x);
     if (vesa_address_of_window[vesa_write_window]!=(new_win_address=(UWORD)(offset>>vesa_granularity_potenz)))
     {
      vesa_address_of_window[vesa_write_window]=new_win_address;
      vesa_set_write_window();
     }
     *((UBYTE *)(vesa_write_seg+offset_x))=(UBYTE)data;
    }
    /**********************/
    /* Check error status */
    /**********************/
    err+=dy;
    if (err>0)
    {
     err-=dx;
     iY1+=inc;
    }
   }
   while(iX1++!=iX2);
  }
  else
  {
   if (iY1>iY2)
   {
    fSwap(&iX1,&iX2);
    fSwap(&iY1,&iY2);
   }
   dx=iX2-iX1;
   dy=iY2-iY1;

   err=-(dy>>1);
   if(dx>0)
   {
    inc=1;
   }
   else
   {
    inc=-1;
    dx=-dx;
   }
   do
   {
    /**************************/
    /* Calulate byte position */
    /**************************/
    offset=linestartpos[(UWORD)iY1]+(ULONG)iX1;
    if (!(*(clip_memory+offset)))
    {
     data=(UBYTE)iColour | *(overlay_memory+offset);
     offset_x=(UWORD)(offset&vesa_and_x);
     if (vesa_address_of_window[vesa_write_window]!=(new_win_address=(UWORD)(offset>>vesa_granularity_potenz)))
     {
      vesa_address_of_window[vesa_write_window]=new_win_address;
      vesa_set_write_window();
     }
     *((UBYTE *)(vesa_write_seg+offset_x))=(UBYTE)data;
    }
    /**********************/
    /* Check error status */
    /**********************/
    err+=dx;

    if (err>=0)
    {
     err-=dy;
     iX1+=inc;
    }
   }
   while(iY1++!=iY2);
  }
 }
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOAreaSave()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Save a screen area and return a pointer to a bitbloc      */
/*                 structure containing a pointer to the data. The user must */
/*                 use fIOBitBlockRelease() to free an unwanted block.       */
/*                                                                           */
/*                                                                           */
/*   Accepts     : PIXELs giving Area boundaries                             */
/*                                                                           */
/*   Returns     : Pointer to save bitblock area                             */
/*                                                                           */
/*****************************************************************************/

BITBLOCK* fIOAreaSavebank(SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2)
{
 BITBLOCKPTR   tBlock=NULL;
 ULONG         lImageSize;
 ULONG         src;
 UBYTE         *dest;
 UWORD         xloop,yloop;

 /********************************************************************************/
 /* Do a sanity check and make sure X1,Y1 is upper left and X2,Y2 is lower right */
 /********************************************************************************/
 if (iX1>iX2)
  fSwap(&iX1,&iX2);
 if (iY1>iY2)
  fSwap(&iY1,&iY2);

 /************************************************/
 /* malloc some space for the bitblock structure */
 /************************************************/
 if ((tBlock=malloc(1*sizeof(BITBLOCK)))==NULL)
 {
  fFatal("I/O ERROR: fIOAreaSave() malloc Failed for BITBLOCK structure");
  return(NULL);
 }

 /*************************************/
 /* Calc the image size and malloc it */
 /*************************************/
 lImageSize=((iX2-iX1)+1)*((iY2-iY1)+1);
 if ((dest=tBlock->tBlockData=(UBYTE*)malloc(lImageSize*sizeof(UBYTE)))==NULL)
 {
  fFatal("I/O ERROR: fIOAreaSave() malloc failed for image");
  free(tBlock);
  return(NULL);
 }

 /*************************************/
 /* Now get the image into the buffer */
 /*************************************/
 tBlock->width=(iX2-iX1)+1;
 tBlock->height=(iY2-iY1)+1;

 src=(iY1*iScreenXres)+iX1;
 for(yloop=iY1;yloop<=iY2;yloop++)
 {
  for(xloop=iX1;xloop<=iX2;xloop++)
  {
   offset_x=(UWORD)(src&vesa_and_x);
   if (vesa_address_of_window[vesa_read_window]!=(new_win_address=(UWORD)(src>>vesa_granularity_potenz)))
   {
    vesa_address_of_window[vesa_read_window]=new_win_address;
    vesa_set_read_window();
   }
   *(dest++)=*((UBYTE *)(vesa_read_seg+offset_x));
   src++;
  }
  src+=iScreenXres-tBlock->width;
 }

 /*************************/
 /* Set the Magic numbers */
 /*************************/
 tBlock->tBitblockMagicNumber1=0xa55aa55al;
 tBlock->tBitblockMagicNumber2=0xa55aa55al;
 return(tBlock);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fIOAreaLoad()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Loads a saved image into an area, erasing the old area.   */
/*                                                                           */
/*   Accepts     : PIXEL pointing to load position and block pointer         */
/*                                                                           */
/*   Returns     : BOOLEAN to say if area was loaded sucessfully             */
/*                                                                           */
/*****************************************************************************/

BOOL fIOAreaLoadbank(SWORD iXdest,SWORD iYdest,BITBLOCK *tData)
{
 UBYTE         *src;
 UWORD         xloop,yloop;
 ULONG         dest;

 if(tData->tBitblockMagicNumber1 != tData->tBitblockMagicNumber2)
 {
  fFatal("I/O ERROR: fIOAreaLoad() Bad Block Handle, aborting load");
 }
 else
 {
  /*************************************/
  /* Now get the image into the buffer */
  /*************************************/
  dest=(iYdest*iScreenXres)+iXdest;
  src=tData->tBlockData;
  for(yloop=iYdest;yloop<iYdest+tData->height;yloop++)
  {
   for(xloop=iXdest;xloop<iXdest+tData->width;xloop++)
   {
    /*****************************************/
    /* Check clipping and reject if required */
    /*****************************************/
    if (xloop>=iClippingX1 && xloop<=iClippingX2 && yloop>=iClippingY1 && yloop<=iClippingY2)
    {
     offset_x=(UWORD)(dest&vesa_and_x);
     if (vesa_address_of_window[vesa_write_window]!=(new_win_address=(UWORD)(dest>>vesa_granularity_potenz)))
     {
      vesa_address_of_window[vesa_write_window]=new_win_address;
      vesa_set_write_window();
     }
     *((UBYTE *)(vesa_write_seg+offset_x))=*(src);
    }
    dest++;
    src++;
   }
   dest+=iScreenXres-(tData->width);
  }
 }
 return(TRUE);
}

/***************************************************************/

/* Interrupts */
#define MAUS_INTERRUPT 0x33

/* Maus Funktionen */
#define MAUS_RESET 0x0
#define MAUS_BEWEGE_ZEIGER 0x04
#define MAUS_HORIZONTALE_BEWEGUNG 0x07
#define MAUS_VERTIKALE_BEWEGUNG 0x08
#define MAUS_EVENT_HANDLER_SETZEN 0x0c
#define MAUS_ALTERNATIVEN_EVENT_HANDLER_SETZEN 0x18
#define MAUS_VERHAELTNIS_MICKEY_PUNKTE 0x0f

/* sonstige defines */
#define MAUS_EVENT_HANDLER_AKTIV_BITS (255)
#define MAUS_ALTERNATIVE_EVENT_HANDLER_AKTIV_BITS (1+2+4+8+16+32+64+128)
#define MAUS_TREIBER_TEST 0xffff

/* Initialisierung globaler Daten */
UWORD vesa_mouse_initialisiert=FALSE;

/**********************************************************************/

void vesa_mouse_update_bank(void)
{
 static UWORD read_win,write_win;
 /*************************/
 /* restore last position */
 /*************************/
 if (mlastoffset!=-1)
 {
  if (vesa_address_of_window[vesa_write_window]!=mlastwin)
  {
   vesa_address_of_window[vesa_write_window]=mlastwin;
   vesa_set_write_window();
  }
  if (mlastclipdata)
   *((UBYTE *)(vesa_write_seg+mlastoffsetx))=(UBYTE)mlastscreendata;
  else
   *((UBYTE *)(vesa_write_seg+mlastoffsetx))=(UBYTE)0;
  *(((UBYTE *)clip_memory)+mlastoffset)=(UBYTE)mlastclipdata;
 }

 /**********************/
 /* save this position */
 /**********************/
 mlastoffset=mouse_x+linestartpos[(UWORD)mouse_y];
 mlastoffsetx=(UWORD)(mlastoffset&vesa_and_x);
 mlastwin=(UWORD)(mlastoffset>>vesa_granularity_potenz);
 mlastclipdata=*(((UBYTE *)clip_memory)+mlastoffset);
 if (vesa_address_of_window[vesa_read_window]!=mlastwin)
 {
  vesa_address_of_window[vesa_read_window]=mlastwin;
  vesa_set_read_window();
 }
 mlastscreendata=*((UBYTE *)(vesa_read_seg+mlastoffsetx));

 /**********************/
 /* draw mouse pointer */
 /**********************/
 if (vesa_address_of_window[vesa_write_window]!=mlastwin)
 {
  vesa_address_of_window[vesa_write_window]=mlastwin;
  vesa_set_write_window();
 }
 if (mouse_button)
  *((UBYTE *)(vesa_write_seg+mlastoffsetx))=254; // white dot for now
 else
  *((UBYTE *)(vesa_write_seg+mlastoffsetx))=255; // white dot for now
 *(((UBYTE *)clip_memory)+mlastoffset)=1; // white dot for now

 if (vesa_address_of_window[vesa_write_window]!=write_win)
 {
  vesa_address_of_window[vesa_write_window]=write_win;
  vesa_set_write_window();
 }
 if (vesa_address_of_window[vesa_read_window]!=read_win)
 {
  vesa_address_of_window[vesa_read_window]=read_win;
  vesa_set_read_window();
 }
}

/**********************************************************************/

/********************************************************************/
/* pragma only needed when check stack active, but what the hell... */
/********************************************************************/
/***************************************************************/
/* for some reason or another I have not been able to call the */
/* setbank routine from mouse interrupt in protected mode code */
/* now I have to do it via a poll..., anybody got a clue why?  */
/***************************************************************/
#pragma off (check_stack)
void _loadds far vesa_mouse_get_mouse(ULONG max, ULONG mcx, ULONG mdx, ULONG mbx)
{
#pragma aux vesa_mouse_get_mouse parm [EAX] [ECX] [EDX] [EBX]
 _disable();
 mouse_code=max;
 mouse_x=mcx;
 mouse_y=mdx;
 vmouse_x=(mcx-mausxmin-((mausxmax-mausxmin)/2))*dividerx;
 vmouse_y=-(mdx-mausymin-((mausymax-mausymin)/2))*dividery;
 mouse_button=mbx;
 mchanged=TRUE;
 _enable();
}
#pragma on (check_stack)

/**********************************************************************/

UWORD fIOInitMousebank(void)
{
 struct SREGS sregs;
 union REGS inregs, outregs;
 segread(&sregs);

// fFatal("fIOInitMousebank() - Mouse for banked VESA not supported yet!");

 /**************************/
 /* check for mouse driver */
 /**************************/
 inregs.w.ax = 0;
 int386(MAUS_INTERRUPT, &inregs, &outregs);
 if (!(minstalled=(outregs.w.ax == 0xffff)))
 {
  fFatal("fIOInitMouse() - Mouse Driver not found!");
 }
 /*****************************/
 /* install vesa_mouse_update */
 /*****************************/
 inregs.w.ax = 0xC;
 inregs.w.cx = 1+2+4+8+16; /* move, right/left press and release */
 inregs.x.edx = FP_OFF(vesa_mouse_get_mouse);
 sregs.es     = FP_SEG(vesa_mouse_get_mouse);
 int386x(MAUS_INTERRUPT, &inregs, &outregs, &sregs);

 inregs.x.eax=MAUS_VERHAELTNIS_MICKEY_PUNKTE;
 inregs.x.ecx=8;
 inregs.x.edx=8;
 int386x(MAUS_INTERRUPT,&inregs,&outregs,&sregs);

 inregs.x.eax=MAUS_BEWEGE_ZEIGER;
 inregs.x.edx=0;
 inregs.x.ecx=0;
 int386x(MAUS_INTERRUPT,&inregs,&outregs,&sregs);

 inregs.x.eax=MAUS_HORIZONTALE_BEWEGUNG;
 inregs.x.ecx=mausxmin;
 inregs.x.edx=mausxmax;
 int386x(MAUS_INTERRUPT,&inregs,&outregs,&sregs);

 inregs.x.eax=MAUS_VERTIKALE_BEWEGUNG;
 inregs.x.ecx=mausymin;
 inregs.x.edx=mausymax;
 int386x(MAUS_INTERRUPT,&inregs,&outregs,&sregs);

 if (bGblDisplayVerbose==TRUE)
  printf("\nbanked mouse installed!");

 return 1;
}

/**********************************************************************/

/********************************************************************/
/* pragma only needed when check stack active, but what the hell... */
/********************************************************************/
#pragma off (check_stack)
void _loadds far vesa_mouse_update_lin(ULONG max, ULONG mcx, ULONG mdx, ULONG mbx)
{
#pragma aux vesa_mouse_update_lin parm [EAX] [ECX] [EDX] [EBX]
 mouse_code=max;
 mouse_x=mcx;
 mouse_y=mdx;
 vmouse_x=(mcx-mausxmin-((mausxmax-mausxmin)/2))*dividerx;
 vmouse_y=-(mdx-mausymin-((mausymax-mausymin)/2))*dividery;
 mouse_button=mbx;
 _disable();

 /*************************/
 /* restore last position */
 /*************************/
 if (mlastoffset!=-1)
 {
  if (mlastclipdata)
   *(((UBYTE *)svdc->videoMem)+mlastoffset)=(UBYTE)mlastscreendata;
  else
   *(((UBYTE *)svdc->videoMem)+mlastoffset)=(UBYTE)0;
  *(((UBYTE *)clip_memory)+mlastoffset)=(UBYTE)mlastclipdata;
 }
 /**********************/
 /* save last position */
 /**********************/
 mlastoffset=mouse_x+linestartpos[mouse_y];
 mlastscreendata=*(((UBYTE *)svdc->videoMem)+mlastoffset);
 mlastclipdata=*(((UBYTE *)clip_memory)+mlastoffset);
 if (mouse_button)
  *(((UBYTE *)svdc->videoMem)+mlastoffset)=254; //* white dot for now
 else
  *(((UBYTE *)svdc->videoMem)+mlastoffset)=255; //* white dot for now
 *(((UBYTE *)clip_memory)+mlastoffset)=1; // white dot for now
 _enable();
}
#pragma on (check_stack)

/**********************************************************************/

UWORD fIOInitMouselin(void)
{
 struct SREGS sregs;
 union REGS inregs, outregs;
 segread(&sregs);

 /**************************/
 /* check for mouse driver */
 /**************************/
 inregs.w.ax = 0;
 int386(MAUS_INTERRUPT, &inregs, &outregs);
 if (!(minstalled=(outregs.w.ax == 0xffff)))
 {
  fFatal("fIOInitMouse() - Mouse Driver not found!");
 }
 /*****************************/
 /* install vesa_mouse_update */
 /*****************************/
 inregs.w.ax = 0xC;
 inregs.w.cx = 1+2+4+8+16; /* move, right/left press and release */
 inregs.x.edx = FP_OFF(vesa_mouse_update_lin);
 sregs.es     = FP_SEG(vesa_mouse_update_lin);
 int386x(MAUS_INTERRUPT, &inregs, &outregs, &sregs);

 inregs.x.eax=MAUS_VERHAELTNIS_MICKEY_PUNKTE;
 inregs.x.ecx=8;
 inregs.x.edx=8;
 int386x(MAUS_INTERRUPT,&inregs,&outregs,&sregs);

 mlastoffset=0;
 mlastscreendata=*(((UBYTE *)svdc->videoMem)+mlastoffset);
 mlastclipdata=*(((UBYTE *)clip_memory)+mlastoffset);

 inregs.x.eax=MAUS_BEWEGE_ZEIGER;
 inregs.x.edx=0;
 inregs.x.ecx=0;
 int386x(MAUS_INTERRUPT,&inregs,&outregs,&sregs);

 inregs.x.eax=MAUS_HORIZONTALE_BEWEGUNG;
 inregs.x.ecx=mausxmin;
 inregs.x.edx=mausxmax;
 int386x(MAUS_INTERRUPT,&inregs,&outregs,&sregs);

 inregs.x.eax=MAUS_VERTIKALE_BEWEGUNG;
 inregs.x.ecx=mausymin;
 inregs.x.edx=mausymax;
 int386x(MAUS_INTERRUPT,&inregs,&outregs,&sregs);

 if (bGblDisplayVerbose==TRUE)
  printf("\nlinear mouse installed!");

 return 1;
}












/***********************************************/
/* file must be open, handle in pcx_akt_handle */
/* loads some of a pcx file into a buffer      */
/* returns number of bytes read                */
/* sets pcx_muss_gelesen_werden to FALSE       */
/* if EOF                                      */
/***********************************************/
UWORD pcx_load_bild_buffer(void)
{
 UWORD fehler=FALSE;
 UWORD bytes_read=0;
 if (pcx_muss_gelesen_werden==TRUE)
 {
  bytes_read=fread(pcx_load_buffer, 1, LOAD_BUFFER_SIZE,pcx_akt_handle);
  if (bytes_read!=LOAD_BUFFER_SIZE)
   pcx_muss_gelesen_werden=FALSE;
  pcx_buffer_counter=0;
  fehler=TRUE;
 }
 return fehler;
}

/*************************************************************************/

/********************************************************************/
/* loads color information of a pcx file into the 'vectrex-palette' */
/* file must be open and handle in 'pcx_akt_handle'                 */
/* colors 15,31,...239 are 'line-colors', they are graded down to 0 */
/* within their 16 color block                                      */
/* colors 240-255 are 'border' colors, which are taken directly     */
/* from the file to the palette                                     */
/********************************************************************/
void pcx_get_color256(void)
{
 int i;
 FARB_TABELLE_256 pcx_256_farb_tabelle;
 fseek(pcx_akt_handle,-768,FROM_FILE_END);
 fread(&pcx_256_farb_tabelle,3,256,pcx_akt_handle);

 /* try for background == 0 differnt from black 1 */
 for (i=0;i<256;i++)
 {
  {
   palette[i][PRED]=pcx_256_farb_tabelle.rgb_farben[i].red;
   palette[i][PGREEN]=pcx_256_farb_tabelle.rgb_farben[i].green;
   palette[i][PBLUE]=pcx_256_farb_tabelle.rgb_farben[i].blue;
  }
 }

 if (bGblLoadColorsExact!=TRUE)
 {
  for (i=15+16;i<240;i+=16)
  {
   /****************************/
   /* only process colors != 0 */
   /****************************/
   if (pcx_256_farb_tabelle.rgb_farben[i].red+
       pcx_256_farb_tabelle.rgb_farben[i].green+
       pcx_256_farb_tabelle.rgb_farben[i].blue!=0)

   {
    if (bGblNoShades==TRUE)
    {
     int ii;
     for (ii=0;ii<16;ii++)
     {
      palette[i-ii][PRED]=pcx_256_farb_tabelle.rgb_farben[i].red;
      palette[i-ii][PGREEN]=pcx_256_farb_tabelle.rgb_farben[i].green;
      palette[i-ii][PBLUE]=pcx_256_farb_tabelle.rgb_farben[i].blue;
     }
    }
    else
    {
     /******************************************/
     /* scale the color from highest to lowest */
     /******************************************/
     int ii;
     float r,g,b;
     r=abs((int) (pcx_256_farb_tabelle.rgb_farben[i].red-pcx_256_farb_tabelle.rgb_farben[i>>4].red))/15;
     g=abs((int) (pcx_256_farb_tabelle.rgb_farben[i].green-pcx_256_farb_tabelle.rgb_farben[i>>4].green))/15;
     b=abs((int) (pcx_256_farb_tabelle.rgb_farben[i].blue-pcx_256_farb_tabelle.rgb_farben[i>>4].blue))/15;
     for (ii=0;ii<16;ii++)
     {
      palette[i-ii][PRED]=pcx_256_farb_tabelle.rgb_farben[i].red-r*ii;
      palette[i-ii][PGREEN]=pcx_256_farb_tabelle.rgb_farben[i].green-g*ii;
      palette[i-ii][PBLUE]=pcx_256_farb_tabelle.rgb_farben[i].blue-b*ii;
     }
    }
   }
  }
 }
 for (i=240;i<256;i++)
 {
  /********************************************/
  /* just takes these colors into the palette */
  /********************************************/
  if (pcx_256_farb_tabelle.rgb_farben[i].red+
      pcx_256_farb_tabelle.rgb_farben[i].green+
      pcx_256_farb_tabelle.rgb_farben[i].blue!=0)
  {
   palette[i][PRED]=pcx_256_farb_tabelle.rgb_farben[i].red;
   palette[i][PGREEN]=pcx_256_farb_tabelle.rgb_farben[i].green;
   palette[i][PBLUE]=pcx_256_farb_tabelle.rgb_farben[i].blue;
  }
 }
}

/*************************************************************************/

/**********************************************************************/
/* reads and pcx file to the screen                                   */
/* pcx file MUST HAVE THE SAME SIZE AS SCREEN RESOLUTION!!!           */
/* colors 0-239 are taken as overlay colors, screen set to background */
/*              clipping memory is set to 0                           */
/*              overlay is set to the correspronding color            */
/*                                                                    */
/* colors 240-255 are printed to screen as they are                   */
/*                clipping memory is set to 1                         */
/*                overlay is ignored                                  */
/**********************************************************************/
UWORD fIOload_pcx_overlay(char *filename)
{
 int zeilen_laenge;
 int zeilen_counter;
 int i;
 int zeilen=0;
 UWORD pcx_bild_art=BILD_UNBEKANNT;
 UBYTE crunched_wert;
 UBYTE anzahl_der_bytes;
 UBYTE *overlay_pointer=(UBYTE*)overlay_memory;
 UBYTE *clip_pointer=(UBYTE*)clip_memory;

 /*****************/
 /* open PCX file */
 /*****************/
 if ((pcx_akt_handle=fopen(filename,"rb"))==NULL)
  fFatal("load_pcx_overlay() - Picture not found!");

 /***************************/
 /* read header information */
 /***************************/
 fread(&pcx_header, sizeof(PCX_HEADER), 1, pcx_akt_handle);
 zeilen_laenge=pcx_header.bytes_pro_zeile;

 /*************************************/
 /* is the number of colors correct ? */
 /*************************************/
 if ((pcx_header.bits_pro_pixel==8)&&(pcx_header.anzahl_farbebenen==1))
 {
  /***************************/
  /* has is the right size ? */
  /***************************/
  if (   (iScreenXres!=pcx_header.xmax-pcx_header.xmin+1)
      || (iScreenYres!=pcx_header.ymax-pcx_header.ymin+1))
  {
   fclose(pcx_akt_handle);
   fFatal("load_pcx_overlay() - Wrong overlay size!");
  }
  pcx_bild_art=BILD_IST_256_FARBE_BILD;
 }
 if (pcx_bild_art==BILD_UNBEKANNT)
 {
  fclose(pcx_akt_handle);
  fFatal("load_pcx_overlay() - Wrong picture format!");
 }

 /************************************************/
 /* PCX file is ok, set file to start of picture */
 /************************************************/
 fseek(pcx_akt_handle,128,FROM_FILE_START);

 /******************************/
 /* get memory for load buffer */
 /******************************/
 if ((pcx_load_buffer=(UBYTE  *)malloc(LOAD_BUFFER_SIZE+1))==NULL )
 {
  fclose(pcx_akt_handle);
  fFatal("load_pcx_overlay() - No memory buffer left!");
 }
 pcx_muss_gelesen_werden=TRUE;
 pcx_load_bild_buffer();

 /*****************************************************/
 /* well and decode it... I won't go into it here ... */
 /*****************************************************/
 while (zeilen<iScreenYres)
 {
  zeilen_counter=0;
  while (zeilen_counter!=zeilen_laenge)
  {
   crunched_wert=pcx_load_buffer[pcx_buffer_counter++];
   if (pcx_buffer_counter==LOAD_BUFFER_SIZE)
    pcx_load_bild_buffer();
   if (crunched_wert<192)
   {
    fIOPlotPixel(zeilen_counter,zeilen,crunched_wert>>4);
    overlay_pointer[zeilen_counter]=crunched_wert&(128+64+32+16);
    clip_pointer[zeilen_counter]=0;
    zeilen_counter++;
   }
   else
   {
    anzahl_der_bytes=crunched_wert&63;
    crunched_wert=pcx_load_buffer[pcx_buffer_counter++];
    if (pcx_buffer_counter==LOAD_BUFFER_SIZE)
     pcx_load_bild_buffer();
    for (i=0;i<anzahl_der_bytes;i++)
    {
     if (crunched_wert<240)
     {
      fIOPlotPixel(zeilen_counter,zeilen,crunched_wert>>4);
      overlay_pointer[zeilen_counter]=crunched_wert&(128+64+32+16);
      clip_pointer[zeilen_counter]=0;
      zeilen_counter++;
     }
     else
     {
      fIOPlotPixel(zeilen_counter,zeilen,crunched_wert);
      clip_pointer[zeilen_counter]=1;
      zeilen_counter++;
     }
    }
   }
  }
  clip_pointer+=zeilen_laenge;
  overlay_pointer+=zeilen_laenge;
  zeilen++;
 }
 if (pcx_load_buffer)
 {
  free(pcx_load_buffer);
  pcx_load_buffer=NULL;
 }
 pcx_get_color256();
 fclose(pcx_akt_handle);
 return TRUE;
}

/*************************************************************************/

UWORD fIOpcx_write_256pic(void)
{
 FILE *this_file=NULL;
 PCX_HEADER this_pcx_header;
 UBYTE zeile[2000];

#ifndef CLAY_ONLY
 char *filename="screen00.pcx";
#endif


 char fname[8+3+1+1];

 FARB_TABELLE_256 this_256_tabelle;
 int x,y;
 UBYTE this_byte,zaehler;
 this_pcx_header.hersteller=10;
 this_pcx_header.version=5;
 this_pcx_header.kodierungsart=1;
 this_pcx_header.bits_pro_pixel=8;
 this_pcx_header.xmin=0;
 this_pcx_header.ymin=0;
 this_pcx_header.xmax=iScreenXres-1;
 this_pcx_header.ymax=iScreenYres-1;
 this_pcx_header.aufloesung_x=iScreenXres;
 this_pcx_header.aufloesung_y=iScreenYres;
 this_pcx_header.anzahl_farbebenen=1;
 this_pcx_header.bytes_pro_zeile=iScreenXres;
 this_pcx_header.farb_art=1;


 #ifdef CLAY_ONLY
 sprintf(fname,"%08i.pcx",cgrabsdone);


 #else
  strcpy(fname,filename);
  while ((this_file = fopen(fname, "rb"))!= NULL)
 {
  fclose(this_file);
  fname[7]++;
  if (fname[7]=='9'+1)
  {
   fname[7]='0';
   fname[6]++;
   if (fname[6]=='9'+1)
   {
    fFatal("fIOpcx_write_256pic() - Filenames all used!");
   }
  }
 }
#endif
 this_file = fopen(fname, "w+b");
 if (this_file==NULL)
 {
  return FALSE;
 }
 fwrite(&this_pcx_header,sizeof(PCX_HEADER),1,this_file);
 for (y=0;y<iScreenYres;y++)
 {
  for (x=0;x<iScreenXres;x++)
  {
   zeile[x]=fIOGetPixel(x,y);
  }
  x=0;
  while (x<iScreenXres)
  {
   this_byte=zeile[x++];
   zaehler=1+64+128;
   if (this_byte>63)
   {
    while ((zeile[x]==this_byte)&&(zaehler!=255)&&(x<iScreenXres))
    {
     x++;
     zaehler++;
    }
    fwrite(&zaehler,1,1,this_file);
    fwrite(&this_byte,1,1,this_file);
   }
   else
   {
    if (zeile[x]!=this_byte)
     fwrite(&this_byte,1,1,this_file);
    else
    {
     while ((zeile[x])==this_byte&&(zaehler!=255)&&(x<iScreenXres))
     {
      x++;
      zaehler++;
     }
     fwrite(&zaehler,1,1,this_file);
     fwrite(&this_byte,1,1,this_file);
    }
   }
  }
 }
 zaehler=12; /* einleitung zur 256 Farbpalette */
 fwrite(&zaehler,1,1,this_file);

 for (y=0;y<256;y++)
 {
  {
   this_256_tabelle.rgb_farben[y].red=palette[y][PRED];
   this_256_tabelle.rgb_farben[y].green=palette[y][PGREEN];
   this_256_tabelle.rgb_farben[y].blue=palette[y][PBLUE];
  }
 }

 fwrite(&this_256_tabelle,768,1,this_file);
 fclose(this_file);
 return TRUE;
}

/*************************************************************************/
/* End of IO.C */

