/*****************************************************************************/
/*                                                                           */
/*   LIBRARY Keyboard functions                    Copyright K.Wilkins 1996  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Title      : Keyboard handler functions                                 */
/*                                                                           */
/*   File Name  : KEYBOARD.C                                                 */
/*                                                                           */
/*   Author     : Keith Wilkins                                              */
/*                                                                           */
/*   Version    : 1.00                                                       */
/*                                                                           */
/*   Desciption : This file contains the functions handle the keyabord       */
/*                input and control.                                         */
/*                                                                           */
/*                These functions are the most PRIMATIVE Level of IO and     */
/*                as such are the fastest.                                   */
/*                                                                           */
/*   Functions  : fKBDinit()      - Initialise the I/O System                */
/*                fKBDrestore()   - Restore I/O to its virgin state          */
/*                fKBDgetch()     - Wait and get a pressed key               */
/*                fKBDungetch()   - Wait and get a pressed key               */
/*                fKBDhit()       - Is a key being pressed                   */
/*                fKBDisdown()    - Check a particular key for press         */
/*                fKBDfuncreg()   - Map a function pointer to a keypress     */
/*                fKBDdokey()     - Execute a function if a key is pressed   */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Revision History:                                                       */
/*                                                                           */
/*   Version    Date    Who  Description of changes                          */
/*   -------    ----    ---  ----------------------                          */
/*                                                                           */
/*    0.01    20/03/93  K.W  Creation                                        */
/*                                                                           */
/*    0.02    04/06/94  K.W  Mouse routines added, initial code lifted from  */
/*                           DEU 5.2 Courstesy of Raphael Quinet             */
/*                                                                           */
/*    0.03    01/11/94  K.W  Mods to use new bailout fFatal, non-returning   */
/*                                                                           */
/*    0.04    12/11/94  K.W  Removed keyboard code from GFX.C and interfaced */
/*                           to LMW's keyboard handler.                      */
/*    1.00    05/08/96  K.W  Public release of DVE V1.0                      */
/*                                                                           */
/*****************************************************************************/
/* These functions MUST NOT use any fDebug calls as they may be begin used   */
/* to setup the debug window and as such will generate recusive function     */
/* calls until the stack if they call fDebug whilst it is being setup        */
/*****************************************************************************/


#define KEYBOARD_C

#include "vectrex.h"

/***************/
/* Scan buffer */
/***************/
#define  NO_OF_SCANMAPS 0x80
volatile UBYTE abScanKeyMap[NO_OF_SCANMAPS];

/*************************************/
/* Scan to Key code conversion array */
/*************************************/
#include "keyconv.h"


/*******************/
/* Keyboard buffer */
/*******************/
#define  KEYBUFSIZE     512
static   SWORD          aiKeyBuff[KEYBUFSIZE];
static   volatile SWORD iKeyBuffPtr=-1;


/**************************************************************************/
/* The keyboard function pointer array holds a function pointer for every */
/* key on the keyboard, both normal and extended chars are handled        */
/*                                                                        */
/*               000-127 Mapped to ASCII codes 00-7f                      */
/*               128-384 Mapped to Extended Keyboard codes 0000-00ff      */
/**************************************************************************/
#define NO_OF_KEYMAPS   385
static void(*apfKeyboard_Map[NO_OF_KEYMAPS])(UWORD);
BOOL bKeyBufferEnable=TRUE;

void fKBDnotused(UWORD tKeyVal)
{
 tKeyVal=0;  /* Stop compiler warnings */
//  putch(7);
}

/******************************/
/* DOS Keyboard vector handle */
/******************************/
static void __interrupt *fOld_key_handler;
static void __interrupt fKBDhandler(void);

extern BOOL bKBDKeyPressed;
static BOOL initialised=FALSE;


/*****************************************************************************/
/*                                                                           */
/*   Function    : fKBDinit()                                                */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function intercepts the keyabord interrupt and       */
/*                 installs a new one.                                       */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

BOOL fKBDinit(void)
{
   UWORD  iLoop;
/**********************************************************************/
/* Set all entries of the KeyMap to fNotUsed, makes a beep if pressed */
/**********************************************************************/
 for(iLoop=0;iLoop<NO_OF_KEYMAPS;iLoop++)
 {
  apfKeyboard_Map[iLoop]=fKBDnotused;
 }

/************************************/
// Intercept the keyboard interrupt */
/************************************/
 fOld_key_handler=_dos_getvect(0x09);
 _dos_setvect(0x09,fKBDhandler);

 /**************************/
 /* Flush the scan key map */
 /**************************/
 memset((void *)abScanKeyMap,0,NO_OF_SCANMAPS);
 initialised=TRUE;
 return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fKBDrestore()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Restore the old keyboard handler.                         */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : BOOL TRUE=OK FALSE=VERROR                                  */
/*                                                                           */
/*****************************************************************************/

BOOL fKBDrestore(void)
{
 if (initialised)
  _dos_setvect(0x09,fOld_key_handler);
 initialised=FALSE;
 return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fKBDbuffering()                                           */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Switch keyboard buffering ON/OFF.                         */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : BOOL TRUE=OK FALSE=VERROR                                  */
/*                                                                           */
/*****************************************************************************/

void fKBDbuffering(BOOL bState)
{
 bKeyBufferEnable=bState;
 while(bKBDKeyPressed)
  fKBDgetch();
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fKBDhandler()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function intercepts the keyboard interrupt and       */
/*                 processes the interrupt.                                  */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

static void __interrupt fKBDhandler(void)
{
 /*****************************************************/
 /* -=- get the scan code of the key just pressed -=- */
 /*****************************************************/
 UBYTE        cScan_code = inp(0x60);
 static SWORD iKeyCode=0;
 static SWORD iOffset=0;
   
 /************************************************/
 /* -=- if it is an 0xe0 or 0 then ignore it -=- */
 /************************************************/

 if ((cScan_code != 0xe0) && (cScan_code != 0))
 {
  /********************************************************/
  /* Step one setup the SCAN map table, ONLY BASE VALUES  */
  /* -=- if the top bit is set then its a key release -=- */
  /********************************************************/
  if (cScan_code&0x80)
  {
   /************************************/
   /* -=- mark the key as released -=- */
   /************************************/
   abScanKeyMap[cScan_code&0x7f] = 0;
  }
  else
  {
   /***********************************/
   /* -=- make the key as pressed -=- */
   /***********************************/
   abScanKeyMap[cScan_code&0x7f] = 0x01;
   abScanKeyMap[0] = cScan_code;
  }
  /**********************************************************/
  /* Step two process the input scancode based on meta-keys */
  /* only register keys on a press DOWN stroke              */
  /**********************************************************/
  if (!(cScan_code&0x80) && (bKeyBufferEnable || (!bKeyBufferEnable&&!bKBDKeyPressed)))
  {
   switch(cScan_code)
   {
    case SCAN_LSHIFT: /**********************************/
    case SCAN_RSHIFT: /* Don't buffer any of these keys */
    case SCAN_CTRL:   /**********************************/
    case SCAN_ALT:
     break;
    default:
     /*********************************************************/
     /* First calcuate the offset caused by any active meta's */
     /*********************************************************/
     if (abScanKeyMap[SCAN_LSHIFT] || abScanKeyMap[SCAN_RSHIFT])
     {
      iOffset=SCAN_SHIFT_OFFSET;
     }
     else if(abScanKeyMap[SCAN_CTRL])
     {
      iOffset=SCAN_CTRL_OFFSET;
     }
     else if(abScanKeyMap[SCAN_ALT])
     {
      iOffset=SCAN_ALT_OFFSET;
     }
     else
     {
      iOffset=SCAN_BASE_OFFSET;
     }

     /***************************************************************/
     /* If this is an extended char then stuff a 0x00 in the buffer */
     /***************************************************************/
     if (iKeyBuffPtr<=KEYBUFSIZE && asScanToKeyConvert[(int)cScan_code+iOffset].iExtended)
     {
      aiKeyBuff[++iKeyBuffPtr]=0x00;
     }

     /*********************************/
     /* Now convert this to a keycode */
     /*********************************/
     iKeyCode=asScanToKeyConvert[(int)cScan_code+iOffset].iKeyCode;

     /***********************************/
     /* Place the key in the buffer     */
     /* Do the equivalent of an ungetch */
     /***********************************/
     if (iKeyBuffPtr<=KEYBUFSIZE)
     {
      aiKeyBuff[++iKeyBuffPtr]=iKeyCode;
      bKBDKeyPressed=TRUE;
     }
     break;
   }
  }
 }
 /*****************************/
 /* -=- ack the interrupt -=- */
 /*****************************/
 outp(0x20,0x20);
}
 
/*****************************************************************************/
/*                                                                           */
/*   Function    : fKBDgetch()                                               */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Wait for a keypress and return the value                  */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : WORD key pressed                                          */
/*                                                                           */
/*****************************************************************************/

UWORD fKBDgetch(void)
{
 UWORD iKey;
 /***********************************************/
 /* Wait until there is something in the buffer */
 /***********************************************/
 while(iKeyBuffPtr==-1)
  ;

 /*********************************/
 /* Fetch the key from the buffer */
 /*********************************/
 _disable();
 iKey=aiKeyBuff[iKeyBuffPtr--];
 if(iKeyBuffPtr==-1)
  bKBDKeyPressed=FALSE;
 _enable();

 return(iKey);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fKBDungetch()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Return a key press to the input queue                     */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : WORD key pressed                                          */
/*                                                                           */
/*****************************************************************************/

void fKBDungetch(UWORD tKeypress)
{
 _disable();
 if (iKeyBuffPtr<=KEYBUFSIZE)
 {
  aiKeyBuff[++iKeyBuffPtr]=tKeypress;
//    bKBDKeyPressed=TRUE;
 }
 _enable();
 return;
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fKBDhit()                                                 */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Checks if a key is being pressed. and returns the value   */
/*                 of the key being pressed or zero if nothing pressed.      */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : BYTE flag TRUE/FALSE if key is being pressed              */
/*                                                                           */
/*****************************************************************************/

BOOL fKBDhit(void)
{
 return (bKBDKeyPressed);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fKBDisdown()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Checks if a particular key is being held down at that     */
/*                 moment                                                    */
/*                                                                           */
/*   Accepts     : scan code to check                                        */
/*                                                                           */
/*   Returns     : BYTE flag TRUE/FALSE if key is being pressed              */
/*                                                                           */
/*****************************************************************************/

BOOL fKBDisdown(UWORD iScancode)
{
 if (abScanKeyMap[iScancode])
  return(TRUE);
 else
  return(FALSE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fKBDfuncreg()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Associates a function with a key and that function is     */
/*                 called whenever that key is pressed.                      */
/*                                                                           */
/*   Accepts     : WORD key number, function pointer with WORD argument      */
/*                                                                           */
/*   Returns     : BYTE TRUE/FALSE is OK                                     */
/*                                                                           */
/*****************************************************************************/

BOOL fKBDfuncreg(UWORD tKeyNum,void(*pfKeyFunc)(UWORD))
{
 if(pfKeyFunc==NULL)
 {
  fFatal("Null pointer passed to fKBDfuncreg() in %s line %d",__FILE__,__LINE__);
 }
 if(tKeyNum>=NO_OF_KEYMAPS)
 {
  fFatal("Bad key index passed to fKBDfuncreg() in %s line %d",__FILE__,__LINE__);
 }
 apfKeyboard_Map[tKeyNum]=pfKeyFunc;
 return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fKBDdokey()                                               */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function checks if a key has been pressed and runs   */
/*                 the associated function of that key.                      */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void fKBDdokey(void)
{
 UWORD tUserInput;
 UWORD tFunctionIndex;

 if(fKBDhit()==TRUE)
 {
  /*****************************************************************/
  /* Calculate array index of function pointer from keyboard input */
  /*****************************************************************/
  tUserInput=fKBDgetch();
  if (tUserInput==0)
  {
   tFunctionIndex=128;
   if (!fKBDhit())
   {
    /*******************************************************/
    /* Only half an extended char found in buffer          */
    /* so well quit coz I don't know what else to do ????? */
    /*******************************************************/
    return;
   }
   else
   {
    tUserInput=fKBDgetch();
   }
  }
  else
  {
   tFunctionIndex=0;
  }
  tFunctionIndex+=tUserInput;

  /**************************************************/
  /* Now call the function associated with that key */
  /* If the index is valid                          */
  /**************************************************/
  if (tFunctionIndex && tFunctionIndex<NO_OF_KEYMAPS)
  {
   (apfKeyboard_Map[tFunctionIndex])(tFunctionIndex);
  }
  else
  {
   fFatal("Bad key index generated in fKBDdokey() in %s line %d",__FILE__,__LINE__);
  }
 }
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fKBDbufferenable()                                        */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function enables/disables the keyboard buffering     */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void fKBDbufferenable(BOOL bState)
{
 bKeyBufferEnable=bState;
 while(fKBDhit()) fKBDgetch();
}


/* End of KEYBOARD.C */
