/*****************************************************************************/
/*                                                                           */
/*   VECTREX HP3000 Software Emulator              Copyright K.Wilkins 1996  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Title      : This file contains all of the Debug Monitor command funcs  */
/*                                                                           */
/*   File Name  : MONITOR.C                                                  */
/*                                                                           */
/*   Author     : Keith Wilkins                                              */
/*                                                                           */
/*   Version    : 1.00                                                       */
/*                                                                           */
/*   Desciption : The Vectrex debug monitor commands are all function driven */
/*                and this file conatins all of the command functions that   */
/*                are called by fDebugMonitor() in debug.c.                  */
/*                                                                           */
/*                fDebugMonitor()     - The debug monitor kernel caller      */
/*                fDebugBreakRun()    - Run with breakpoints enabled         */
/*                fDebugDissemble()   - Dissasemble 6809 code                */
/*                fDebugDump()        - Hex to to monitor window             */
/*                fDebugExit()        - Exit from the debug monitor          */
/*                fDebugFind()        - Find a hex string                    */
/*                fDebugHelp()        - Print Monitor help listing           */
/*                fDebugRun()         - Run with breakpoints disabled        */
/*                fDebugSequence()    - Run N steps of the system sequencer  */
/*                fDebugStep()        - Run N single steps of the CPU        */
/*                fDebugTicks()       - Display system tick counts           */
/*                fDebugQuit()        - Quit from the emulator               */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Revision History:                                                       */
/*                                                                           */
/*   Version    Date    Who  Description of changes                          */
/*   -------    ----    ---  ----------------------                          */
/*                                                                           */
/*    0.01    20/03/93  K.W  Creation                                        */
/*    1.00    05/08/96  K.W  Public release of DVE V1.0                      */
/*                                                                           */
/*****************************************************************************/

#define MONITOR_C

#include "vectrex.h"
#include "mon_cmds.h"


#ifdef DEBUG_MON

extern DEBUGWININFO asMonWin[];

/* All functions in the commandlist array accept and integer argument count */
/* and a pointer to and array of pointers to char which contain the args.   */
/* All functions return a BYTE flag TRUE/FALSE as an indication of if the   */
/* debug monitor should exit or not.                                        */

/* Define the command line array, all code for these commands can be found  */
/* in the file called mon_cmds.c                                            */

tsCOMMAND   asCommandList[]=
{                  
   fMonCmdBreakSet     ,"BSET"     ,"BSET xxxx,yyyy           - Set processor breakpoint for addr <xxxx> after <yyyy> (optional) passes",
   fMonCmdBreakClear   ,"BCLR"     ,"BCLR                     - Clear processor breakpoint",
   fMonCmdBreakShow    ,"BSHOW"    ,"BSHOW                    - Show processor breakpoint",
#ifdef DEBUG_MSG
   fMonCmdDebug        ,"DEBUG"    ,"DEBUG                    - Toggle debug window",
   fMonCmdDebug        ,"DBG"      ,"",
#endif
   fMonCmdDump         ,"D"        ,"",
   fMonCmdDump         ,"DUMP"     ,"D(UMP) xxxx,yyyy         - Memory Dump from <xxxx> to (<yyyy>) optional",
   fMonCmdRegDump      ,"DR"       ,"DR                       - Dump 6809 registers",
   fMonCmdFind         ,"F"	    ,"",
   fMonCmdFind	        ,"FIND"		,"F(IND) xxxx,yyyy <key>   - Find <key> hex or string between xxxx and yyyy",
   fMonCmdHelp         ,"H"        ,"",
   fMonCmdHelp         ,"HELP"     ,"H(ELP)                   - Print this text",
   fMonCmdRun          ,"R"        ,"",                     
   fMonCmdRun          ,"RUN"      ,"R(UN) xxxx               - Run the emulator from xxxx (optional) and disable the debug monitor",
   fMonCmdReset        ,"RESET"    ,"RESET <ALL|CPU|VIDEO>    - Reset vectrex hardware",
   fMonCmdStep         ,"S"        ,"",
   fMonCmdStep	        ,"STEP"		,"S(TEP) xxxx              - Single step the CPU <xxxx> cycles",
   fMonCmdStepRun      ,"SRUN"     ,"SRUN xxxx                - Slow run & debug trace from <xxxx> (optional)",
   fMonCmdSequence     ,"SEQ"      ,"SEQ s,yy                 - Exec <yy> steps of <s> (T=System,C=CPU,D=Display,S=Sound)",
   fMonCmdTicks	    ,"T"	    ,"",
   fMonCmdTicks        ,"TICKS"    ,"T(ICKS)                  - Display system tick counts",
   fMonCmdQuit         ,"Q"        ,"",
   fMonCmdQuit         ,"QUIT"     ,"QUIT                     - Quit from the Vectrex emulator back to DOS",
   fMonCmdDissemble    ,"U"        ,"U xxxx,yy                - U xxxx yy  Un-assemble 6809 code from xxxx, yy instructions",
   fMonCmdExit         ,"X"        ,"",
   fMonCmdExit	        ,"EXIT"		,"EXIT|X                   - Exit the debug monitor",
   NULL                ,""         ,""            
};


GWINDOWPTR   pwMonitorWin=NULL;


/*****************************************************************************/
/*                                                                           */
/*   Function    : fDebugMonitor()                                           */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function is the runtime debugger and is called from  */
/*                 the sequencer if the debugger is enabled once on every    */
/*                 loop of the sequencer. It has a number of local flags     */
/*                 that are used for breakpointing the emulator. The func    */
/*                 then either returns or gets user input from a command     */
/*                 line interpreter. See the command list for details of     */
/*                 what commands are available.                              */
/*                 All i/o is done via the functions contained in io.c       */
/*                                                                           */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void fDebugMonitor(void)
{
   static BOOL  tFirstTime=TRUE;
   static BOOL  bStepRun=FALSE;

// BOOL         tReturnFlag=FALSE;

   SWORD        iArgC,iLineLength;
   UBYTE        *apcArgV[MAX_COMMAND_LINE_ARGS];
   UBYTE        acCommandLine[MAX_COMMAND_LINE_LENGTH+1];
   UWORD        iKeyPress;
   UWORD        iBPoint,iBPSkip,iBPEnable;

   SWORD        iTmp;
   CPUREGISTERS tsTmp;    
    
   /* 1st Check all the breakpoints to see if we should run command line */

   if(lwGblSystemTickBreakpoint!=0)
   {
    	if(lwGblSystemTickTotal>=lwGblSystemTickBreakpoint)
   	{
       	fDebug(HIGH,("fDebugMonitor(): SystemTick breakpoint detected at %ld",lwGblSystemTickBreakpoint));
       	lwGblSystemTickBreakpoint=0;
   	}
   	else
	   {
   		return;
   	}
   }

   // Entrance to the monitor is only valid for:
   //
   // 1. When we've just finished a CPU cycle ** Every cycle is a CPU cycle
   // 1. Before a FETCH has completed so BP is triggered i.e state==FETCH
   // 2. After a CPU exception i.e state==WAIT
   //

   if(f6809StatusRead()!=CPU_FETCH && f6809StatusRead()!=CPU_WAIT && f6809StatusRead()!=CPU_CWAI && f6809StatusRead()!=CPU_SYNC)
   {
//    fDebug(MID,("fDebugMonitor() rejected monitor entrance 0x%02x",f6809StatusRead()));
      return;
   }
   

   /* Check if our window is open, if not then open it, otherwise unhide it */

   if(pwMonitorWin==NULL)
   {
       if((pwMonitorWin=fWinOpen(asMonWin[wGblDisplayMode].iWinXpos,asMonWin[wGblDisplayMode].iWinYpos,
                                 asMonWin[wGblDisplayMode].iWinXsize,asMonWin[wGblDisplayMode].iWinYsize,
                                 asMonWin[wGblDisplayMode].iWinFGcolour,asMonWin[wGblDisplayMode].iWinBGcolour,
                                 WIN_DEFAULT_BORDER_WIDTH,asMonWin[wGblDisplayMode].iWinBDcolour))==NULL)
       {
           fDebug(VERROR,("fDebugMonitor() Couldnt open its window"));
           return;
       }
       fWinTitle(pwMonitorWin,"Debug Monitor",FONT_SMALL,(BASE_FONT_HEIGHT*FONT_SMALL)+2,
                 asMonWin[wGblDisplayMode].iTitleFGcolour,asMonWin[wGblDisplayMode].iTitleBGcolour);
   }
   else
   {
       fWinRestore(pwMonitorWin);
   }
   
   /* We must need to run the command line if we've arrived here, phew */

   if(tFirstTime==TRUE)
   {
       tFirstTime=FALSE;
       lwGblSystemTickBreakpoint=0;
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"Vectrex HP3000 Emulator Debug Monitor\n");
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"-------------------------------------\n\n");
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"Type h(elp) for the command list\n\n");
   }


   // Decide what print depending on what has happened - cases are:
   //
   //  1.  CPU exception
   //  2.  Breakpoint
   //  3.  Return to monitor from single step
   //  4.  Return thru monitor in step run mode
   //  5.  Monitor open
   //


   //
   // CASE 1 - CPU EXCEPTION. processor is in WAIT mode
   //
   //      banner, tick count, last ins, registers
   //

   f6809RegistersRead(&tsTmp);
   f6809ReadBreakpoint(&iBPoint,&iBPSkip,&iBPEnable);

   if(f6809StatusRead()==CPU_WAIT)
   {
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n**** 6809 Exception generated at address $%04x ****\n",f6809LastInsAddr());
       fMonCmdTicks(0,NULL);
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n");
       sprintf(acCommandLine,"0x%04x",f6809LastInsAddr());
       apcArgV[0]=acCommandLine;
       apcArgV[1]=acCommandLine;
       apcArgV[2]=acCommandLine;
       fMonCmdDissemble(3,apcArgV); 
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n");
       fMonCmdRegDump(0,NULL);
       bStepRun=FALSE;
   }

   //
   // CASE 2 - Breakpoint entry
   //
   //      tick count, registers, next ins
   //


   else if(iBPoint==tsTmp.tPC && iBPSkip==0 && iBPEnable==TRUE)
   {
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n**** 6809 Breakpoint occured at address $%04x ****\n",tsTmp.tPC);
       fMonCmdTicks(0,NULL);
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n");
       fMonCmdRegDump(0,NULL);
       sprintf(acCommandLine,"0x%04x",tsTmp.tPC);
       apcArgV[0]=acCommandLine;
       apcArgV[1]=acCommandLine;
       apcArgV[2]=acCommandLine;
       fMonCmdDissemble(3,apcArgV);
       bStepRun=FALSE;
   }

   //
   // CASE 3 - Monitor entry after single step
   //
   //      tick count, last ins, registers, next ins
   //


   else if(f6809LastInsAddr()!=tsTmp.tPC && bStepRun==FALSE)
   {
       fMonCmdTicks(0,NULL);
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n");
       sprintf(acCommandLine,"0x%04x",f6809LastInsAddr());
       apcArgV[0]=acCommandLine;
       apcArgV[1]=acCommandLine;
       apcArgV[2]=acCommandLine;
       fMonCmdDissemble(3,apcArgV); 
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n");
       fMonCmdRegDump(0,NULL);
       sprintf(acCommandLine,"0x%04x",tsTmp.tPC);
       apcArgV[0]=acCommandLine;
       apcArgV[1]=acCommandLine;
       apcArgV[2]=acCommandLine;
       fMonCmdDissemble(3,apcArgV); 
   }

   //
   // CASE 4 - Run thru monitor during step run
   //
   //      tick count, last ins, registers
   //

   else if(f6809LastInsAddr()!=tsTmp.tPC && bStepRun==TRUE)
   {
       fMonCmdTicks(0,NULL);
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n");
       sprintf(acCommandLine,"0x%04x",f6809LastInsAddr());
       apcArgV[0]=acCommandLine;
       apcArgV[1]=acCommandLine;
       apcArgV[2]=acCommandLine;
       fMonCmdDissemble(3,apcArgV); 
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n");
       fMonCmdRegDump(0,NULL);
   }

   //
   // CASE 5 - Just monitor open
   //
   //      tick count, registers
   //

   else
   {
      fMonCmdTicks(0,NULL);
      fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n");
      fMonCmdRegDump(0,NULL);
   }


   //
   // In step-run mode we must check for a keypress and abort if required
   //

   if(bStepRun==TRUE)
   {
      if(fKBDhit())
      {
         iKeyPress=fKBDgetch();
         if(iKeyPress==' ')
         {
            if(fKBDgetch()==' ') return;
            
         }
         bStepRun=FALSE;
      }
      else if(f6809StatusRead()==CPU_WAIT)
      {
         bStepRun=FALSE;
      }
      else
      {
         return;
      }
   }

   
   /* Enter the command loop */

   do
   {
       /* Print the prompt */

       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n>");

       /* Fetch the input line */

       fWinInput(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,acCommandLine,MAX_COMMAND_LINE_LENGTH);
       fDebug(MID,("fDebugMonitor(): Got input line \"%s\"",acCommandLine));

       /* Scan the command line and convert to arg list */

       iArgC=0;
       iTmp=0;
       iLineLength=strlen(acCommandLine);

       while(iTmp<iLineLength && iArgC<MAX_COMMAND_LINE_ARGS)
       {
           if(fIsWhiteSpace(acCommandLine[iTmp])==TRUE)
           {
               iTmp++;
           }
           else
           {

               /* Now we found a string start, register the address */

               apcArgV[iArgC++]=&acCommandLine[iTmp];

               /* Scan to the end of the field */

               for(;;)
               {
                   if(fIsWhiteSpace(acCommandLine[iTmp])==FALSE && acCommandLine[iTmp]!=0)
                   {
                       acCommandLine[iTmp]=toupper(acCommandLine[iTmp]);
                       iTmp++;
                   }
                   else
                   {
                       acCommandLine[iTmp]=0;
                       fDebug(MID,("fDebugMonitor(): argv[%d] = \"%s\"",iArgC-1,apcArgV[iArgC-1]));
                       iTmp++;
                       break;
                   }
               }
           }
       }

       /* Do a string match for agrv0 for command name and execute the */
       /* associated command in the command list array                 */

       if(iArgC>0)
       {
       iTmp=0;
           for(;;)
           {
               if(strcmp(apcArgV[0],asCommandList[iTmp].pcCommandName)==0)
               {

                   /* Now we've got a match, execute the command */
       
                   fDebug(HIGH,("fDebugMonitor(): executing command %d (%s)",iTmp,asCommandList[iTmp].pcCommandName));

				/* If a command returns TRUE then exit debug, this is used for */
				/* single step and run commands that dont want the window closed */
				                                                                        
                   switch((*asCommandList[iTmp].pfCommandFunc)(iArgC,apcArgV))
                   {
                   	case DBG_OK:
                   		fDebug(HIGH,("fDebugMonitor(): returned OK"));
                   		break;
                   	case DBG_ERROR:
                   		fDebug(HIGH,("fDebugMonitor(): returned ERROR"));
                   		break;
                   	case DBG_QUIT:
                   		fDebug(HIGH,("fDebugMonitor(): returned QUIT"));
      					   /* Disable the debug Monitor */
   	  					   bGblDebugMonitorEnable=FALSE;
	      				   break;
                   	case DBG_SEQUENCE:
                   		fDebug(HIGH,("fDebugMonitor(): returned SEQUENCE"));
                   		return;
                   		break;
                     case DBG_STEPRUN:
                        fDebug(HIGH,("fDebugMonitor(): returned STEPRUN"));
                        bStepRun=TRUE;
                        return;
                        break;
                   	default:
                   		fDebug(HIGH,("fDebugMonitor(): returned unknown value"));
                   		break;
                   }
                   break;
               }

               /* Check for the end of command array and print commant not */
               /* found error and exit from this command scan loop         */

               iTmp++;
               if(asCommandList[iTmp].pfCommandFunc==NULL)
               {
                   fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\nCommand not found.");
                   break;
               }
           }
       }
   }
   while(bGblDebugMonitorEnable==TRUE);

   /* Either hide the monitor window or delete it if a quit have been triggered */

   if(bGblQuitFlag==FALSE)
   {
       fWinPrintf(pwMonitorWin,asMonWin[wGblDisplayMode].iWinFontsize,"\n");
       fWinHide(pwMonitorWin);
   }
   else
   {
       fWinClose(pwMonitorWin);
       pwMonitorWin=NULL;
   }

}


#endif

/* End of MONITOR.C */
