/*****************************************************************************/
/*                                                                           */
/*   VECTREX HP3000 Software Emulator              Copyright K.Wilkins 1996  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Title      : Vectrex emulator sequencing module functions               */
/*                                                                           */
/*   File Name  : SEQUENCE.C                                                 */
/*                                                                           */
/*   Author     : Keith Wilkins / C.S.                                       */
/*                                                                           */
/*   Version    : 1.20                                                       */
/*                                                                           */
/*   Desciption : This file contains the functions for the Vectrex sequencer */
/*                module. This module executes all of the emulator sub-func- */
/*                tions at the correct times depending on the various tick   */
/*                counters that are used to regulate relative timing.        */
/*                                                                           */
/*   Functions  : fVectrexEmulate() - Run the emulator for a while           */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Revision History:                                                       */
/*                                                                           */
/*   Version    Date    Who  Description of changes                          */
/*   -------    ----    ---  ----------------------                          */
/*                                                                           */
/*    0.01    14/03/93  K.W  Creation of empty file and descriptions         */
/*                                                                           */
/*    0.02    17/03/93  K.W  1st Stab at the sequencer                       */
/*                                                                           */
/*    1.00    05/08/96  K.W  Public release of DVE V1.0                      */
/*                                                                           */
/*    1.20    20/10/96  C.S  MultiTick and inline stuff                      */
/*                                                                           */
/*****************************************************************************/

#define SEQUENCE_C

#include "vectrex.h"

#include "emu6522.h" /* for save function */
#include "analog.h" /* for save function */
#include "soundemu.h" /* for save function */

BOOL (* fVectrexEmulate)(void)=NULL;

/*****************************/
/* DOS Timer vector handlers */
/*****************************/
static void __far __interrupt *dos_timer_handler;
static void __far __interrupt fTimerTick(void);

static volatile ULONG timer_tick=0;

static SLONG tick_timeout=1;
static ULONG last_timer_tick=0;
static ULONG new_ticks=0;
static ULONG counter_since_change=0;
int state=1;
static int wheel_color_rotator=0;

#ifdef CLAY_ONLY
static int cscreenshots=FALSE;
int cgrabsdone=0;
#endif


/************************************/
/* Define local function prototypes */
/************************************/

void fSequencerPause(UWORD tKeyPress);
void fSequencerAbort(UWORD tKeyPress);        // Trigger sequencer abort
void fSequencerReset(UWORD tKeyPress);        // Trigger sequencer reset
void fSequencerReboot(UWORD tKeyPress);       // System reset
void fCarouselSelect(UWORD tKeyPress);        // Select from a number of games
void fScreenshot(UWORD tKeyPress);
void fSaveState(UWORD tKeyPress);
void fLoadState(UWORD tKeyPress);

#ifdef CLAY_ONLY
void fCScreenshot(UWORD tKeyPress);
#endif
/*****************************************************************************/
/*                                                                           */
/*   Function    : fVectrexEmulate()                                         */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function is the HEART of the emluator. It schedules  */
/*                 and executes all of the functions that make up the        */
/*                 Vectrex emulator. The system is run on an infinite loop   */
/*                 which is only broken if the tQuitFlag==TRUE, this is a    */
/*                 global variable and can be set by any module of the user  */
/*                 wants to quit or the error is serious enough. For each    */
/*                 loop the gGblSystemTick variable is incremented. Also     */
/*                 for each loop the tGbl????SystemTickCounters are decreme- */
/*                 nted and when one reaches zero tGbl????SystemTickTotal    */
/*                 is incremented and the f????Tick function is executed.    */
/*                                                                           */
/*                 The debug flags can disable the execution of any of the   */
/*                 tick functions in an individual basis. When the DEBUG     */
/*                 mode is activated the fDoDebug() is called on every       */
/*                 time tGblSystemTick in incrmented.                        */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : TRUE or FALSE depending on if an error occured            */
/*                                                                           */
/*****************************************************************************/

/* different sequencers for different needs... */

BOOL fVectrexEmulate_linear_no_lightpen(void)
#define SEQUENCER_LINEAR
#include "seq.cc"
#undef SEQUENCER_LINEAR

BOOL fVectrexEmulate_linear_imager(void)
#define SEQUENCER_LINEAR
#define SEQUENCER_IMAGER
#include "seq.cc"
#undef SEQUENCER_IMAGER
#undef SEQUENCER_LINEAR


BOOL fVectrexEmulate_banked_no_lightpen(void)
#ifdef TRYING
{return FALSE;}
#endif
#ifndef TRYING
#define SEQUENCER_BANKED
#include "seq.cc"
#undef SEQUENCER_BANKED
#endif

BOOL fVectrexEmulate_banked_imager(void)
#ifdef TRYING
{return FALSE;}
#endif
#ifndef TRYING
#define SEQUENCER_IMAGER
#define SEQUENCER_BANKED
#include "seq.cc"
#undef SEQUENCER_IMAGER
#undef SEQUENCER_BANKED
#endif


#define SEQUENCER_LIGHTPEN
BOOL fVectrexEmulate_linear_lightpen(void)
#ifdef TRYING
{return FALSE;}
#endif
#ifndef TRYING
#define SEQUENCER_LINEAR
#include "seq.cc"
#undef SEQUENCER_LINEAR
#endif

BOOL fVectrexEmulate_banked_lightpen(void)
#ifdef TRYING
{return FALSE;}
#endif
#ifndef TRYING
#define SEQUENCER_BANKED
#include "seq.cc"
#undef SEQUENCER_BANKED
#undef SEQUENCER_LIGHTPEN
#endif

/*****************************************************************************/
/*                                                                           */
/*   Function    : fTimerInit()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Steal the dos timer interrupt 0x1C                        */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : true/false on sucess/fail                                 */
/*                                                                           */
/*****************************************************************************/
BOOL fTimerInit(void)
{
   dos_timer_handler=_dos_getvect(DOS_TIMER_IRQ);
   _dos_setvect(DOS_TIMER_IRQ,fTimerTick);
   return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fTimerRestore()                                           */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Restore the dos timer interrupt                           */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : true/false on success/fail                                */
/*                                                                           */
/*****************************************************************************/
BOOL fTimerRestore(void)
{
   _dos_setvect(DOS_TIMER_IRQ,dos_timer_handler);
   return(TRUE);
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fTimerTick()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Timer tick interrupt  18.2 Hz                             */
/*                                                                           */
/*   Accepts     : nothing                                                   */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/

void __far __interrupt fTimerTick(void)
{
 timer_tick++;
//   (dos_timer_handler)();      // Chain the DOS timer tick
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fPauseSequencer()                                         */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Halt the sequencer temporarily                            */
/*                                                                           */
/*   Accepts     : WORD key index that caused the call                       */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fSequencerPause(UWORD tKeyPress)
{
    tKeyPress=0;
    bGblStatsValid=FALSE;
    fSoundMute();
    fKBDbuffering(TRUE);
    fKBDgetch();
    fKBDbuffering(FALSE);
    fSoundResume();
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fAbortSequencer()                                         */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Set the sequencer abort flag to force the system to exit. */
/*                                                                           */
/*   Accepts     : WORD key index that caused the call                       */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fSequencerAbort(UWORD tKeyPress)
{
 tKeyPress=0;
 bGblQuitFlag=TRUE;
}

void fScreenshot(UWORD tKeyPress)
{
 tKeyPress=0;
 fIOpcx_write_256pic();
}

#ifdef CLAY_ONLY
void fCScreenshot(UWORD tKeyPress)
{
 tKeyPress=0;
 if (cscreenshots==TRUE)
 {
  cscreenshots=FALSE;
  fIOPrintf(0,0,0,0,1,"      ");
  fIOPrintf(0,20,0,0,1,"        ");
 }
 else
  cscreenshots=TRUE;
}
#endif
/*****************************************************************************/
/*                                                                           */
/*   Function    : fSequencerReset()                                         */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Reset the vectrex, same as pressing the front button      */
/*                                                                           */
/*   Accepts     : WORD key index that caused the call                       */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fSequencerReset(UWORD tKeyPress)
{
 tKeyPress=0;
 f6809Reset();
 fDisplayReset();
}

/*****************************************************************************/
/*                                                                           */
/*   Function    : fSequencerReboot()                                        */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Perform complete machine initialisation                   */
/*                                                                           */
/*   Accepts     : Nothing                                                   */
/*                                                                           */
/*   Returns     : BOOL    TRUE=Init OK, FALSE=Init Fail                     */
/*                                                                           */
/*****************************************************************************/
void fSequencerReboot(UWORD tKeyPress)
{
   tKeyPress=0;
   fDebug(HIGH,("fSystemReset() - VECTREX HP3000 Emulator system reset"));

   // Reload RAM if required

   fDisplayReset();
   fSystemRAMLoad(acGblRAMImageFileName);
   fSystemROMLoad(acGblROMImageFileName);
   fCartridgeROMLoad(acGblDefaultCartFileName);
   f6809Reset();
   f6522Reset();
   fSoundReset();
   fAnalogReset();
}
/*****************************************************************************/
/*                                                                           */
/*   Function    : fCarouselSelect()                                         */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Select and load the relevant carousel game                */
/*                                                                           */
/*   Accepts     : WORD key index that caused the call                       */
/*                                                                           */
/*   Returns     : nothing                                                   */
/*                                                                           */
/*****************************************************************************/
void fCarouselSelect(UWORD tKeyPress)
{
   switch(tKeyPress)
   {
       case KEY_ALT_F1:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[0]);
           break;
       case KEY_ALT_F2:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[1]);
           break;
       case KEY_ALT_F3:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[2]);
           break;
       case KEY_ALT_F4:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[3]);
           break;
       case KEY_ALT_F5:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[4]);
           break;
       case KEY_ALT_F6:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[5]);
           break;
       case KEY_ALT_F7:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[6]);
           break;
       case KEY_ALT_F8:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[7]);
           break;
       case KEY_ALT_F9:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[8]);
           break;
       case KEY_ALT_F10:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[9]);
           break;
       case KEY_ALT_F11:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[10]);
           break;
       case KEY_ALT_F12:
           strcpy(acGblDefaultCartFileName,acGblCartCarousel[11]);
           break;
       default:
           break;
   }
   fSequencerReboot(0);
}




void fSaveState(UWORD tKeyPress)
{
 FILE *saveram=NULL;
 unsigned int i;
 char filename[80];

 if(strcmp(acGblDefaultCartFileName,""))
  strcpy(filename,acGblDefaultCartFileName);
 else
  strcpy(filename,acGblROMImageFileName);
 if (strrchr(filename,'.')!=NULL)
  *(strrchr(filename,'.'))='\0';

 if ((tKeyPress>=KEY_ALT_1)&&(tKeyPress<=KEY_ALT_9))
 {
  strcat(filename,".sv ");
  filename[strlen(filename)-1]='1'+tKeyPress-KEY_ALT_1;
 }
 else
 {
  strcat(filename,".sav");
 }

 saveram = fopen(filename, "w+b"); /* file wird ueberschrieben */
 if (saveram==NULL)
 {
  return;
 }
 /***/
 for (i=12*4096+8*256;i<13*4096;i++)
  fwrite((char *)(atSystemImage+i),1,1,saveram);

 if (bGblExtendedRam==TRUE)          /* Animaction Ram 2K */
 {
  for (i=2*4096;i<2*4096+8*256;i++)
   fwrite((char *)(atSystemImage+i),1,1,saveram);
 }
 /***/

 /***/
 fwrite(&lXSampleHold,1,sizeof(lXSampleHold),saveram);
 fwrite(&lYSampleHold,1,sizeof(lYSampleHold),saveram);
 fwrite(&tZSampleHold,1,sizeof(tZSampleHold),saveram);

 fwrite(&lSIGDEF_XOUTPUT,1,sizeof(lSIGDEF_XOUTPUT),saveram);
 fwrite(&lSIGDEF_YOUTPUT,1,sizeof(lSIGDEF_YOUTPUT),saveram);
 fwrite(&tSIGDEF_ZOUTPUT,1,sizeof(tSIGDEF_ZOUTPUT),saveram);
 /***/

 /***/
 fwrite(&t6522ORB_0,1,sizeof(t6522ORB_0),saveram);
 fwrite(&t6522IRB_0,1,sizeof(t6522IRB_0),saveram);

 fwrite(&t6522IRB_0L,
 1,
 sizeof( t6522IRB_0L),
 saveram);

 fwrite(&t6522ORB_1,1,sizeof(t6522ORB_1),saveram);
 fwrite(&t6522IRB_1,1,sizeof(t6522IRB_1),saveram);
 fwrite(&t6522IRB_1L,1,sizeof(t6522IRB_1L),saveram);

 fwrite(&t6522ORB_2,1,sizeof(t6522ORB_2),saveram);
 fwrite(&t6522IRB_2,1,sizeof(t6522IRB_2),saveram);
 fwrite(&t6522IRB_2L,1,sizeof(t6522IRB_2L),saveram);

 fwrite(&t6522ORB_3,1,sizeof(t6522ORB_3),saveram);
 fwrite(&t6522IRB_3,1,sizeof(t6522IRB_3),saveram);
 fwrite(&t6522IRB_3L,1,sizeof(t6522IRB_3L),saveram);

 fwrite(&t6522ORB_4,1,sizeof(t6522ORB_4),saveram);
 fwrite(&t6522IRB_4,1,sizeof(t6522IRB_4),saveram);
 fwrite(&t6522IRB_4L,1,sizeof(t6522IRB_4L),saveram);

 fwrite(&t6522ORB_5,1,sizeof(t6522ORB_5),saveram);
 fwrite(&t6522IRB_5,1,sizeof(t6522IRB_5),saveram);
 fwrite(&t6522IRB_5L,1,sizeof(t6522IRB_5L),saveram);

 fwrite(&t6522ORB_6,1,sizeof(t6522ORB_6),saveram);
 fwrite(&t6522IRB_6,1,sizeof(t6522IRB_6),saveram);
 fwrite(&t6522IRB_6L,1,sizeof(t6522IRB_6L),saveram);

 fwrite(&t6522ORB_7REG,1,sizeof(t6522ORB_7REG),saveram);
 fwrite(&t6522ORB_7,1,sizeof(t6522ORB_7),saveram);
 fwrite(&t6522IRB_7,1,sizeof(t6522IRB_7),saveram);
 fwrite(&t6522IRB_7L,1,sizeof(t6522IRB_7L),saveram);

 fwrite(&t6522ORA,1,sizeof(t6522ORA),saveram);
 fwrite(&t6522IRA,1,sizeof(t6522IRA),saveram);
 fwrite(&t6522IRAL,1,sizeof(t6522IRAL),saveram);

 fwrite(&t6522DDRB,1,sizeof(t6522DDRB),saveram);
 fwrite(&t6522DDRA,1,sizeof(t6522DDRA),saveram);

 fwrite(&t6522SHFTREG,1,sizeof(t6522SHFTREG),saveram);
 fwrite(&t6522SHFTREGACT,1,sizeof(t6522SHFTREGACT),saveram);

 fwrite(&u6522Timer1,1,sizeof(TIMER),saveram);
 fwrite(&u6522Timer2,1,sizeof(TIMER),saveram);
 fwrite(&bTimer2OneShot,1,sizeof(BOOL),saveram);

 fwrite(&t6522AUX_T1CTRL,1,sizeof(t6522AUX_T1CTRL),saveram);
 fwrite(&t6522AUX_T2CTRL,1,sizeof(t6522AUX_T2CTRL),saveram);
 fwrite(&t6522AUX_SHFTCTRL,1,sizeof(t6522AUX_SHFTCTRL),saveram);
 fwrite(&t6522AUX_PA_LE,1,sizeof(t6522AUX_PA_LE),saveram);
 fwrite(&t6522AUX_PB_LE,1,sizeof(t6522AUX_PB_LE),saveram);

 fwrite(&t6522PCR_CB2CTRL,1,sizeof(t6522PCR_CB2CTRL),saveram);
 fwrite(&t6522PCR_CB1CTRL,1,sizeof(t6522PCR_CB1CTRL),saveram);
 fwrite(&t6522PCR_CA2CTRL,1,sizeof(t6522PCR_CA2CTRL),saveram);
 fwrite(&t6522PCR_CA1CTRL,1,sizeof(t6522PCR_CA1CTRL),saveram);

 fwrite(&t6522IER,1,sizeof(t6522IER),saveram);

 fwrite(&t6522IFLG,1,sizeof(t6522IFLG),saveram);

 fwrite(&t6522CA1,1,sizeof(t6522CA1),saveram);
 fwrite(&t6522CA2,1,sizeof(t6522CA2),saveram);
 fwrite(&t6522CB1,1,sizeof(t6522CB1),saveram);
 fwrite(&t6522CB2,1,sizeof(t6522CB2),saveram);

 fwrite(&tsCPURegisters,1,sizeof(tsCPURegisters),saveram);
 fwrite(&iToggle,1,sizeof(iToggle),saveram);
 fwrite(&tcount,1,sizeof(tcount),saveram);
 fwrite(&t1int,1,sizeof(t1int),saveram);
 /***/

 /***/
 fwrite(&iInstruction,1,sizeof(iInstruction),saveram);

 fwrite(&i6809status,1,sizeof(i6809status),saveram);
 fwrite(&iLastAddress,1,sizeof(iLastAddress),saveram);
 /***/

 /***/
// fwrite(&lwGblSystemTickTotal,1,sizeof(lwGblSystemTickTotal),saveram);
// fwrite(&lwGblDisplayTickCounter,1,sizeof(lwGblDisplayTickCounter),saveram);
// fwrite(&tick_timeout,1,sizeof(tick_timeout),saveram);
// fwrite(&last_timer_tick,1,sizeof(last_timer_tick),saveram);
 /***/

 /***/
 fwrite(&tOLD_BC1,1,sizeof(tOLD_BC1),saveram);
 fwrite(&tOLD_BDIR,1,sizeof(tOLD_BDIR),saveram);
 fwrite(&tAddr,1,sizeof(tAddr),saveram);
 /***/

 /***/
 /*
 fwrite(&iLineSelected,1,sizeof(iLineSelected),saveram);
 fwrite(&iLineSelector,1,sizeof(iLineSelector),saveram);
 fwrite(&lwRefreshMarker,1,sizeof(lwRefreshMarker),saveram);

 for (i=0;i<LINESTORE_DEPTH;i++)
  fwrite(&(line_store[0])+i*(sizeof(SDLINE)),1,sizeof(SDLINE),saveram);

 fwrite(&line_freelist,1,sizeof(line_freelist),saveram);
 fwrite(&line_rootnode,1,sizeof(line_rootnode),saveram);

 fwrite(hash_rootnodes,HASH_TABLE_SIZE,sizeof(SWORD),saveram);
 fwrite(&hash_freelist,1,sizeof(SWORD),saveram);

 for (i=0;i<LINESTORE_DEPTH;i++)
  fwrite(&(hash_store[0])+i*(sizeof(SHASH)),1,sizeof(SHASH),saveram);
 */
 /***/
 fclose(saveram);
 tKeyPress=0;
}




void fLoadState(UWORD tKeyPress)
{
// not debug
 FILE *loadram=NULL;
 unsigned int i;
 SWORD loop;
 PIXEL screen;

 char filename[80];

 if(strcmp(acGblDefaultCartFileName,""))
  strcpy(filename,acGblDefaultCartFileName);
 else
  strcpy(filename,acGblROMImageFileName);
 if(strrchr(filename,'.')!=NULL)
  *(strrchr(filename,'.'))='\0';

 if ((tKeyPress>=KEY_1)&&(tKeyPress<=KEY_9))
 {
  strcat(filename,".sv ");
  filename[strlen(filename)-1]='1'+tKeyPress-KEY_1;
 }
 else
 {
  strcat(filename,".sav");
 }
 loadram = fopen(filename, "rb");
 if (loadram==NULL)
 {
  return;
 }
 bGblDisplayFullRefresh=FALSE;
 bGblLightpenEnable=FALSE;
 bGblGoggleGame=FALSE;
 ImagerColorNow=255;
 wGblImagerMode=0;
 wGblColor1Degrees=60;
 wGblColor2Degrees=60;
 wGblColor3Degrees=60;
 wGblImagerXOffset=0;
 wGblImagerXOffset1=0;
 wGblImagerXOffset2=0;

 rectengle_clipping_enable=TRUE;

 divider_x[0]=83;
 divider_y[0]=83;

 divider_x[0]=66;
 divider_y[0]=66;

 divider_x[0]=52;
 divider_y[0]=52;

 divider_x[0]=39;
 divider_y[0]=39;

 divider_x[0]=33;
 divider_y[0]=33;

 display_infox[0].iDivShiftA=7;
 display_infox[0].iDivShiftB=8;
 display_infoy[0].iDivShiftA=7;
 display_infoy[0].iDivShiftB=8;

 display_infox[1].iDivShiftA=7;
 display_infox[1].iDivShiftB=7;
 display_infoy[1].iDivShiftA=7;
 display_infoy[1].iDivShiftB=7;

 display_infox[2].iDivShiftA=6;
 display_infox[2].iDivShiftB=8;
 display_infoy[2].iDivShiftA=6;
 display_infoy[2].iDivShiftB=8;

 display_infox[3].iDivShiftA=6;
 display_infox[3].iDivShiftB=7;
 display_infoy[3].iDivShiftA=6;
 display_infoy[3].iDivShiftB=7;

 display_infox[4].iDivShiftA=6;
 display_infox[4].iDivShiftB=6;
 display_infoy[4].iDivShiftA=6;
 display_infoy[4].iDivShiftB=6;

 screen=fIOGetScreenRes();
 iXOffset=screen.iXpos/2;
 iYOffset=screen.iYpos/2;
 if (bGblDisplayExactScaling==FALSE)
 {
  iX1clip=(lwGblClipX>>iDivShiftAx) + (lwGblClipX>>iDivShiftBx);
  iY1clip=(lwGblClipY>>iDivShiftAy) + (lwGblClipY>>iDivShiftBy);
  iX2clip=(-lwGblClipX>>iDivShiftAx) + (-lwGblClipX>>iDivShiftBx);
  iY2clip=(-lwGblClipY>>iDivShiftAy) + (-lwGblClipY>>iDivShiftBy);

  mausxmin=iXOffset-((20000>>iDivShiftAx)+(20000>>iDivShiftBx));
  mausxmax=iXOffset+((20000>>iDivShiftAx)+(20000>>iDivShiftBx));
  mausymin=iYOffset-((20000>>iDivShiftAy)+(20000>>iDivShiftBy));
  mausymax=iYOffset+((20000>>iDivShiftAy)+(20000>>iDivShiftBy));
 }
 else
 {
  iX1clip=lwGblClipX/dividerx;
  iY1clip=lwGblClipY/dividery;
  iX2clip=-lwGblClipX/dividerx;
  iY2clip=-lwGblClipY/dividery;

  mausxmin=iXOffset-20000/dividerx; // just in case a lightpen
  mausxmax=iXOffset+20000/dividerx; // will be enabled
  mausymin=iYOffset-20000/dividery;
  mausymax=iYOffset+20000/dividery;
 }
 iX1clip+=iXOffset;
 iY1clip+=iYOffset;
 iX2clip+=iXOffset;
 iY2clip+=iYOffset;

 /**************************************/
 /* Initialise Line store & hash table */
 /**************************************/
 for(loop=0;loop<LINESTORE_DEPTH-1;loop++)
 {
  line_store[loop].next_vector=loop+1;
  hash_store[loop].next_hash=loop+1;
 }
 line_store[loop].next_vector=-1;
 hash_store[loop].next_hash=-1;

 line_freelist=0;
 hash_freelist=0;

 line_rootnode=-1;

 for(loop=0;loop<HASH_TABLE_SIZE;loop++)
 {
  hash_rootnodes[loop]=-1;
 }

 if (bGblDisplayAutoRefresh)
  wGblDisplayLinePeriod=1;

// fIOSetClipping(-1,-1,-1,-1);
 fDisplayClear();

 /********************************************/
 /* Initialise overlay memory to default ink */
 /********************************************/
// screen=fIOGetScreenRes();
// if ((overlay_memory)==NULL)
//  fFatal("fIOGetOverlayBuffer() returned NULL");
// screenmem=(ULONG)screen.iXpos*(ULONG)screen.iYpos;
// memset(overlay_memory,wGblDisplayBaseColour<<4,screenmem);

 /***********************************************************/
 /* Now screen is reset then reload the Overlay if required */
 /***********************************************************/
 bGblDisplayFullRefresh=FALSE;
 bGblLightpenEnable=FALSE;
 rectengle_clipping_enable=TRUE;

 fDisplayLoadVOL();

 if (wGblImagerMode!=0)
 {
  /* initialise degrees of color wheel in system ticks */
  /* in lwGblWheelTicks is 1/2 turn of the wheel in ticks */
  /* wGblColor1Degrees+wGblColor2Degrees+wGblColor3Degrees should be 180 degrees */
  ULONG ltmp_all=wGblColor1Degrees+wGblColor2Degrees+wGblColor3Degrees;
  wGblColor1Degrees = (lwGblWheelTicks*wGblColor1Degrees)/ltmp_all;
  wGblColor2Degrees = (lwGblWheelTicks*wGblColor2Degrees)/ltmp_all;
  wGblColor3Degrees = (lwGblWheelTicks*wGblColor3Degrees)/ltmp_all;
  if (lwGblWheelTicks!=(wGblColor1Degrees+wGblColor2Degrees+wGblColor3Degrees))
  {
   wGblColor1Degrees+=lwGblWheelTicks-(wGblColor1Degrees+wGblColor2Degrees+wGblColor3Degrees);
  }
  if (wGblImagerMode==6)
  {
   if (bGblDisplayExactScaling==FALSE)
   {
    wGblImagerXOffset1=-(iXOffset+(-30000>>iDivShiftAx) + (-30000>>iDivShiftBx));
    wGblImagerXOffset2=screen.iXpos-((30000>>iDivShiftAx) + (30000>>iDivShiftBx))-iXOffset ;
    wGblImagerXOffset1=(wGblImagerXOffset1<<iDivShiftAx)+(wGblImagerXOffset1<<iDivShiftBx);
    wGblImagerXOffset2=(wGblImagerXOffset2<<iDivShiftAx)+(wGblImagerXOffset2<<iDivShiftBx);
   }
   else
   {
    wGblImagerXOffset1=-(iXOffset+(-30000/dividerx));
    wGblImagerXOffset2=screen.iXpos-(30000/dividerx)-iXOffset;
    wGblImagerXOffset1*=dividerx;
    wGblImagerXOffset2*=dividerx;
   }
  }
 }



 /*************************************************/
 /* Before exiting set the vector clipping region */
 /*************************************************/
#ifndef DEBUG_MON
#ifndef DEBUG_MSG
 fIOSetClipping(iX1clip,iY1clip,iX2clip,iY2clip);
#endif
#endif
/***/

 for (i=12*4096+8*256;i<13*4096;i++)
  fread((char *)(atSystemImage+i),1,1,loadram);

 if (bGblExtendedRam==TRUE)          /* Animaction Ram 2K */
 {
  for (i=2*4096;i<2*4096+8*256;i++)
   fread((char *)(atSystemImage+i),1,1,loadram);
 }
 /***/
 fread(&lXSampleHold,1,sizeof(lXSampleHold),loadram);
 fread(&lYSampleHold,1,sizeof(lYSampleHold),loadram);
 fread(&tZSampleHold,1,sizeof(tZSampleHold),loadram);

 fread(&lSIGDEF_XOUTPUT,1,sizeof(lSIGDEF_XOUTPUT),loadram);
 fread(&lSIGDEF_YOUTPUT,1,sizeof(lSIGDEF_YOUTPUT),loadram);
 fread(&tSIGDEF_ZOUTPUT,1,sizeof(tSIGDEF_ZOUTPUT),loadram);
 /***/

 /***/
 fread(&t6522ORB_0,1,sizeof(t6522ORB_0),loadram);
 fread(&t6522IRB_0,1,sizeof(t6522IRB_0),loadram);

 fread(&t6522IRB_0L, 1, sizeof( t6522IRB_0L), loadram);

 fread(&t6522ORB_1,1,sizeof(t6522ORB_1),loadram);
 fread(&t6522IRB_1,1,sizeof(t6522IRB_1),loadram);
 fread(&t6522IRB_1L,1,sizeof(t6522IRB_1L),loadram);

 fread(&t6522ORB_2,1,sizeof(t6522ORB_2),loadram);
 fread(&t6522IRB_2,1,sizeof(t6522IRB_2),loadram);
 fread(&t6522IRB_2L,1,sizeof(t6522IRB_2L),loadram);

 fread(&t6522ORB_3,1,sizeof(t6522ORB_3),loadram);
 fread(&t6522IRB_3,1,sizeof(t6522IRB_3),loadram);
 fread(&t6522IRB_3L,1,sizeof(t6522IRB_3L),loadram);

 fread(&t6522ORB_4,1,sizeof(t6522ORB_4),loadram);
 fread(&t6522IRB_4,1,sizeof(t6522IRB_4),loadram);
 fread(&t6522IRB_4L,1,sizeof(t6522IRB_4L),loadram);

 fread(&t6522ORB_5,1,sizeof(t6522ORB_5),loadram);
 fread(&t6522IRB_5,1,sizeof(t6522IRB_5),loadram);
 fread(&t6522IRB_5L,1,sizeof(t6522IRB_5L),loadram);

 fread(&t6522ORB_6,1,sizeof(t6522ORB_6),loadram);
 fread(&t6522IRB_6,1,sizeof(t6522IRB_6),loadram);
 fread(&t6522IRB_6L,1,sizeof(t6522IRB_6L),loadram);

 fread(&t6522ORB_7REG,1,sizeof(t6522ORB_7REG),loadram);
 fread(&t6522ORB_7,1,sizeof(t6522ORB_7),loadram);
 fread(&t6522IRB_7,1,sizeof(t6522IRB_7),loadram);
 fread(&t6522IRB_7L,1,sizeof(t6522IRB_7L),loadram);

 fread(&t6522ORA,1,sizeof(t6522ORA),loadram);
 fread(&t6522IRA,1,sizeof(t6522IRA),loadram);
 fread(&t6522IRAL,1,sizeof(t6522IRAL),loadram);

 fread(&t6522DDRB,1,sizeof(t6522DDRB),loadram);
 fread(&t6522DDRA,1,sizeof(t6522DDRA),loadram);

 fread(&t6522SHFTREG,1,sizeof(t6522SHFTREG),loadram);
 fread(&t6522SHFTREGACT,1,sizeof(t6522SHFTREGACT),loadram);

 fread(&u6522Timer1,1,sizeof(TIMER),loadram);
 fread(&u6522Timer2,1,sizeof(TIMER),loadram);
 fread(&bTimer2OneShot,1,sizeof(BOOL),loadram);

 fread(&t6522AUX_T1CTRL,1,sizeof(t6522AUX_T1CTRL),loadram);
 fread(&t6522AUX_T2CTRL,1,sizeof(t6522AUX_T2CTRL),loadram);
 fread(&t6522AUX_SHFTCTRL,1,sizeof(t6522AUX_SHFTCTRL),loadram);
 fread(&t6522AUX_PA_LE,1,sizeof(t6522AUX_PA_LE),loadram);
 fread(&t6522AUX_PB_LE,1,sizeof(t6522AUX_PB_LE),loadram);

 fread(&t6522PCR_CB2CTRL,1,sizeof(t6522PCR_CB2CTRL),loadram);
 fread(&t6522PCR_CB1CTRL,1,sizeof(t6522PCR_CB1CTRL),loadram);
 fread(&t6522PCR_CA2CTRL,1,sizeof(t6522PCR_CA2CTRL),loadram);
 fread(&t6522PCR_CA1CTRL,1,sizeof(t6522PCR_CA1CTRL),loadram);

 fread(&t6522IER,1,sizeof(t6522IER),loadram);

 fread(&t6522IFLG,1,sizeof(t6522IFLG),loadram);

 fread(&t6522CA1,1,sizeof(t6522CA1),loadram);
 fread(&t6522CA2,1,sizeof(t6522CA2),loadram);
 fread(&t6522CB1,1,sizeof(t6522CB1),loadram);
 fread(&t6522CB2,1,sizeof(t6522CB2),loadram);

 fread(&tsCPURegisters,1,sizeof(tsCPURegisters),loadram);
 fread(&iToggle,1,sizeof(iToggle),loadram);
 fread(&tcount,1,sizeof(tcount),loadram);
 fread(&t1int,1,sizeof(t1int),loadram);
 /***/

 /***/
 fread(&iInstruction,1,sizeof(iInstruction),loadram);

 fread(&i6809status,1,sizeof(i6809status),loadram);
 fread(&iLastAddress,1,sizeof(iLastAddress),loadram);
 /***/

 /***/
// fread(&lwGblSystemTickTotal,1,sizeof(lwGblSystemTickTotal),loadram);
// fread(&lwGblDisplayTickCounter,1,sizeof(lwGblDisplayTickCounter),loadram);
// fread(&tick_timeout,1,sizeof(tick_timeout),loadram);
// fread(&last_timer_tick,1,sizeof(last_timer_tick),loadram);
 /***/

 /***/
 fread(&tOLD_BC1,1,sizeof(tOLD_BC1),loadram);
 fread(&tOLD_BDIR,1,sizeof(tOLD_BDIR),loadram);
 fread(&tAddr,1,sizeof(tAddr),loadram);
 /***/

 /***/
/*
 fread(&iLineSelected,1,sizeof(iLineSelected),loadram);
 fread(&iLineSelector,1,sizeof(iLineSelector),loadram);
 fread(&lwRefreshMarker,1,sizeof(lwRefreshMarker),loadram);

 for (i=0;i<LINESTORE_DEPTH;i++)
  fread(&(line_store[0])+i*(sizeof(SDLINE)),1,sizeof(SDLINE),loadram);

 fread(&line_freelist,1,sizeof(line_freelist),loadram);
 fread(&line_rootnode,1,sizeof(line_rootnode),loadram);

 fread(hash_rootnodes,HASH_TABLE_SIZE,sizeof(SWORD),loadram);
 fread(&hash_freelist,1,sizeof(SWORD),loadram);

 for (i=0;i<LINESTORE_DEPTH;i++)
  fread(&(hash_store[0])+i*(sizeof(SHASH)),1,sizeof(SHASH),loadram);
*/
 /***/

 fclose(loadram);
 tKeyPress=0;
}

/* End of SEQUENCE.C */



