/*****************************************************************************/
/*                                                                           */
/*   VECTREX HP3000 Software Emulator              Copyright K.Wilkins 1996  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Title      : Vectrex System RAM Emulation function module               */
/*                                                                           */
/*   File Name  : SYS_RAM.C                                                  */
/*                                                                           */
/*   Author     : Keith Wilkins                                              */
/*                                                                           */
/*   Version    : 1.00                                                       */
/*                                                                           */
/*   Desciption : This file contains the functions that emulate the Vectrex  */
/*                System RAM.                                                */
/*                                                                           */
/*   Functions  : fSystemRAMInit()      - Init & Load the System RAM         */
/*                fSystemRAMSave()      - Save the System RAM area to disk   */
/*                fSystemRAMRead()      - Read a RAM location                */
/*                fSystemRAMWrite()     - Write to a RAM location            */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Revision History:                                                       */
/*                                                                           */
/*   Version    Date    Who  Description of changes                          */
/*   -------    ----    ---  ----------------------                          */
/*                                                                           */
/*    0.01    14/03/93  K.W  Creation of empty file and descriptions         */
/*    1.00    05/08/96  K.W  Public release of DVE V1.0                      */
/*                                                                           */
/*****************************************************************************/

#define SYS_RAM_C

#include "vectrex.h"




/*****************************************************************************/
/*                                                                           */
/*   Function    : fSystemRAMLoad()                                          */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : This function will open and load a binary file into the   */
/*                 1Kb System RAM space. Before loading all addresses        */
/*                 will be written to SYSTEM_RAM_FILL_VALUE. The file is     */
/*                 opened before the RAM space is cleared as if the load     */
/*                 fails we havent destroyed the current RAM Image.          */
/*                 If a NULL file is passed we fill ram only.                */
/*                                                                           */
/*   Accepts     : char * Pointer to a file name, any length.                */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE/FALSE depending on load success.             */
/*                                                                           */
/*****************************************************************************/

BOOL fSystemRAMLoad(UBYTE *pcRAMFileName)
{
    static BOOL tFirstTime=TRUE;

    BOOL         bRetVal=TRUE;
    FILE         *pfSystemRAM=NULL;
    ULONG        tLoop;
    UBYTE        tData;

    /* First Time through we must decare the RAM to the CPU */

    if(tFirstTime==TRUE)
    {
        tFirstTime=FALSE;
        if(f6809DeclareAddress(SYSTEM_RAM_BASE,SYSTEM_RAM_SIZE,fSystemRAMRead,fSystemRAMWrite)==FALSE)
        {
            fDebug(VERROR,("fSystemRAMLoad() RAM Space delaration failed"));
            return(FALSE);
        }
    }

    /* Now clear out the RAM area */

    for(tLoop=SYSTEM_RAM_BASE;tLoop<=SYSTEM_RAM_TOP;tLoop++)
    {
        f6809AddressWrite((UWORD)tLoop,SYSTEM_RAM_FILL_VALUE);
    }
    fDebug(MED,("fSystemRAMLoad() RAM Space Cleared"));

    /* Try and open the RAM image file, if NULL dont open */

    if(pcRAMFileName!=NULL && strcmp(pcRAMFileName,"")!=0)
    {
        if((pfSystemRAM=fopen(pcRAMFileName,"rb"))==NULL)
        {
            fDebug(VERROR,("fSystemRAMLoad() Couldnt open file \"%s\"",pcRAMFileName));
            bRetVal=FALSE;
        }
        else
        {
            fDebug(MED,("fSystemRAMLoad() \"%s\" opened O.K",pcRAMFileName));

            /* Now load until we reach RAM_TOP or the file ends */

            tLoop=SYSTEM_RAM_BASE;

            while(tLoop<=SYSTEM_RAM_TOP && fread(&tData,sizeof(UBYTE),1,pfSystemRAM)==1)
            {
                f6809AddressWrite((UWORD)tLoop++,tData);
            }
            fDebug(MED,("fSystemRAMLoad() Loaded %d Bytes",tLoop-SYSTEM_RAM_BASE));
        
            fclose(pfSystemRAM);
        }
    }

    /* All finished */

    return(bRetVal);

}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fSystemRAMSave()                                          */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Open a file on disk, write the contents of the system RAM */
/*                 to disk and close the file                                */
/*                                                                           */
/*   Accepts     : char * Pointer to a file name, any length.                */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE/FALSE depending on load success.             */
/*                                                                           */
/*****************************************************************************/

BOOL fSystemRAMSave(UBYTE *pcRAMFileName)
{
    FILE  *pfSystemRAM=NULL;
    UWORD tLoop;
    UBYTE tData;

    /* Try and open the RAM image file */

    if((pfSystemRAM=fopen(pcRAMFileName,"wb"))==NULL)
    {
        fDebug(VERROR,("fSystemRAMSave() Couldnt open file %s",pcRAMFileName));
        return(FALSE);
    }
    else
    {
        fDebug(MED,("fSystemRAMSave() %s opened O.K",pcRAMFileName));
    }

    /* Now Save SYSTEM_RAM_SIZE bytes */

    tLoop=SYSTEM_RAM_BASE;

    while(tLoop<SYSTEM_RAM_BASE+SYSTEM_RAM_SIZE)
    {
        tData=fSystemRAMRead(tLoop++);
        if(fwrite(&tData,sizeof(UBYTE),1,pfSystemRAM)==0)
        {
            fclose(pfSystemRAM);
            fDebug(VERROR,("fSystemRAMSave() File write error, aborting save"));
            return(FALSE);
        }
    }
    fDebug(MED,("fSystemRAMSave() Saved %d Bytes",tLoop-SYSTEM_RAM_BASE));
            
    fclose(pfSystemRAM);

    /* All finished */

    return(TRUE);

}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fSystemRAMRead()                                          */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Read a value fRAM the RAM image, apply the address mask   */
/*                 to the address before reading to ensure were not out of   */
/*                 bounds.                                                   */
/*                                                                           */
/*   Accepts     : WORD Address pointer (16bit)                              */
/*                                                                           */
/*   Returns     : BYTE Data value                                           */
/*                                                                           */
/*****************************************************************************/

UBYTE fSystemRAMRead(UWORD tAddress)
{
    return(atSystemImage[tAddress]);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fSystemRAMWrite()                                         */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Write a byte to the image array, the address is masked    */
/*                 before the write to ensure we're within the bounds of the */
/*                 array. The write is only performed if the RAM write       */
/*                 enable flag is TRUE.                                      */
/*                                                                           */
/*   Accepts     : WORD Address and a BYTE of Data                           */
/*                                                                           */
/*   Returns     : BYTE of Data                                              */
/*                                                                           */
/*****************************************************************************/

UBYTE fSystemRAMWrite(UWORD tAddress,UBYTE tData)
{
    atSystemImage[tAddress]=tData;
    return(tData);
}



/* End of SYS_RAM.C */




























