/*****************************************************************************/
/*                                                                           */
/*   VECTREX HP3000 Software Emulator              Copyright K.Wilkins 1996  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Title      : Vectrex emulator utility functions.                        */
/*                                                                           */
/*   File Name  : UTIL.C                                                     */
/*                                                                           */
/*   Author     : Keith Wilkins                                              */
/*                                                                           */
/*   Version    : 1.00                                                       */
/*                                                                           */
/*   Desciption : This file contains utility functions (mainly text) for     */
/*                the emulator.                                              */
/*                                                                           */
/*   Functions  : fIsWhitespace()    - Is the char passed a whitespace char? */
/*                fCharToInt()       - Convert ASCII char to an integer      */
/*              : fStringToWord()    - Convert ASCII string to a WORD        */
/*                fStringToLongWord()- Convert ASCII string to a LONGWORD    */
/*                fHexStringToByte() - Convert a 2 digit ascii str to a BYTE */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Revision History:                                                       */
/*                                                                           */
/*   Version    Date    Who  Description of changes                          */
/*   -------    ----    ---  ----------------------                          */
/*                                                                           */
/*    0.01    11/04/94  K.W  Creation                                        */
/*    1.00    05/08/96  K.W  Public release of DVE V1.0                      */
/*                                                                           */
/*****************************************************************************/

#define UTIL_C

#include "vectrex.h"




/*****************************************************************************/
/*                                                                           */
/*   Function    : fHexStringToByte()                                        */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Convert the passed hex format ascii string into a BYTE    */
/*                 accepts the following numeric formats.                    */
/*                                                                           */
/*                 53      - Hexadecimal if 1st char is 0-F                  */
/*                 A2      - Hexadecimal if 1st char is 0-F                  */
/*                                                                           */
/*   Accepts     : Pointer to source string and pointer to dest BYTE         */
/*                                                                           */
/*   Returns     : BOOL TRUE/FALSE on success/fail                           */
/*                                                                           */
/*****************************************************************************/

BOOL fHexStringToByte(UBYTE *pcString,UBYTE *pbResult)
{
    UBYTE       *pcTextPtr,*pcStartPos;
    SWORD       iExponent=0,iMultiplier=0;
    SWORD       iTmp,iLoop;
    ULONG       lwValue=0;
        
    /* Step 1 make the string entirely uppercase */

    pcTextPtr=pcString;

    while(*pcTextPtr!=0)
    {
        *pcTextPtr=toupper(*pcTextPtr);
        pcTextPtr++;
    }

    /* Step 2, skip any whitespace on the front of the string */
    
    while(fIsWhiteSpace(*pcString)==TRUE)
    {
        if(*pcString==0) return(FALSE);
        pcString++;
    }

    pcStartPos=pcString;

    while(*(pcString+1)!=0) pcString++;

    do
    {
        iTmp=fCharToInt(*pcString);

        /* Test for validity */

        if(iTmp<0 || iTmp>=16) return(FALSE);

        /* Calculate the Multiplier */

        iMultiplier=1;
        for(iLoop=0;iLoop<iExponent;iLoop++) iMultiplier*=16;
     
        lwValue+=(ULONG)iTmp*iMultiplier;

        /* Increment the exponent */

        iExponent++;

        /* Test the Value to see if we exceed the WORD limit */
        
        if(lwValue>256)
        {
            return(FALSE);
        }
    }
    while(pcString--!=pcStartPos);
        
    /* Step 5 Save the value */

    *pbResult=(UBYTE)lwValue;
        
    return(TRUE);        
}



/*****************************************************************************/
/*                                                                           */
/*   Function    : fStringToWord()                                           */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Convert the passed ascii string to hex WORD. Will         */
/*                 accept the following numeric formats.                     */
/*                                                                           */
/*                 12345   - Decimal                                         */
/*                 A2345   - Hexadecimal if 1st char is A-F                  */
/*                 0X1234  - Hexadecimal                                     */
/*                 $1234   - Hexadecimal                                     */
/*                 01234   - Octal                                           */
/*                 0B01010 - Binary                                          */
/*                                                                           */
/*   Accepts     : Pointer to source string and pointer to dest WORD         */
/*                                                                           */
/*   Returns     : BOOL TRUE/FALSE on success/fail                           */
/*                                                                           */
/*****************************************************************************/

BOOL fStringToWord(UBYTE *pcString,UWORD *pwResult)
{
    UBYTE *pcTextPtr,*pcStartPos;
    SWORD iRadix=0,iExponent=0,iMultiplier=0;
    SWORD iTmp,iLoop;
    ULONG lwValue=0;

    /* Step 1 make the string entirely uppercase */

    pcTextPtr=pcString;

    while(*pcTextPtr!=0)
    {
        *pcTextPtr=toupper(*pcTextPtr);
        pcTextPtr++;
    }

    /* Step 2, skip any whitespace on the front of the string */
    
    while(fIsWhiteSpace(*pcString)==TRUE)
    {
        if(*pcString==0) return(FALSE);
        pcString++;
    }

    /* Step 3 identify the 1st valid char and switch on it */

    switch(*pcString)
    {
        case'$':
            iRadix=16;
            pcString++;
            break;

        case'0':
            if(*(pcString+1)=='X')
            {
                iRadix=16;
                pcString+=2;
            }
            else if(*(pcString+1)=='B')
            {
                iRadix=2;
                pcString+=2;
            }
            else if(*(pcString+1)>='0' && *(pcString+1)<='7')
            {
                iRadix=8;
            }
            else if(*(pcString+1)>='8' && *(pcString+1)<='9')
            {
                iRadix=10;
            }
            else if(*(pcString+1)=='\0')
            {
                /* Special case of the lone zero */
                *pwResult=0;
                return(TRUE);
            }
            else
            {
                return(FALSE);
            }
            break;

        case'1': case'2': case'3': case'4': case'5':
        case'6': case'7': case'8': case'9':
            iRadix=10;
            break;

        case'A': case'B': case'C': case'D': case'E': case 'F':
            iRadix=16;
            break;

        default:
            return(FALSE);
    }

    /* Step 4 decode the string backwards */

    pcStartPos=pcString;

    while(*(pcString+1)!=0) pcString++;

    do
    {
        iTmp=fCharToInt(*pcString);

        /* Test for validity */

        if(iTmp<0 || iTmp>=iRadix) return(FALSE);

        /* Calculate the Multiplier */

        iMultiplier=1;
        for(iLoop=0;iLoop<iExponent;iLoop++) iMultiplier*=iRadix;
     
        lwValue+=(ULONG)iTmp*iMultiplier;

        /* Increment the exponent */

        iExponent++;

        /* Test the Value to see if we exceed the WORD limit */
        
        if(lwValue>65535)
        {
            return(FALSE);
        }
    }
    while(pcString--!=pcStartPos);
        
    /* Step 5 Save the value */

    *pwResult=(UWORD)lwValue;
        
    return(TRUE);        
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fStringToLongWord()                                       */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Convert the passed ascii string to a LONGWORD. Will       */
/*                 accept the following numeric formats.                     */
/*                                                                           */
/*                 12345   - Decimal                                         */
/*                 A2345   - Hexadecimal if 1st char is A-F                  */
/*                 0X1234  - Hexadecimal                                     */
/*                 $1234   - Hexadecimal                                     */
/*                 01234   - Octal                                           */
/*                 0B01010 - Binary                                          */
/*                                                                           */
/*   Accepts     : Pointer to source string and pointer to dest LONGWORD     */
/*                                                                           */
/*   Returns     : BOOL TRUE/FALSE on success/fail                           */
/*                                                                           */
/*****************************************************************************/

BOOL fStringToLongWord(UBYTE *pcString,ULONG *plwResult)
{
    UBYTE *pcTextPtr,*pcStartPos;
    ULONG lwRadix=0,lwExponent=0,lwMultiplier=0;
    ULONG lwTmp,lwLoop;
    ULONG lwValue=0;

    /* Step 1 make the string entirely uppercase */

    pcTextPtr=pcString;

    while(*pcTextPtr!=0)
    {
        *pcTextPtr=toupper(*pcTextPtr);
        pcTextPtr++;
    }

    /* Step 2, skip any whitespace on the front of the string */
    
    while(fIsWhiteSpace(*pcString)==TRUE)
    {
        if(*pcString==0) return(FALSE);
        pcString++;
    }

    /* Step 3 identify the 1st valid char and switch on it */

    switch(*pcString)
    {
        case'$':
            lwRadix=16;
            pcString++;
            break;

        case'0':
            if(*(pcString+1)=='X')
            {
                lwRadix=16;
                pcString+=2;
            }
            else if(*(pcString+1)=='B')
            {
                lwRadix=2;
                pcString+=2;
            }
            else if(*(pcString+1)>='0' && *(pcString+1)<='7')
            {
                lwRadix=8;
            }
            else if(*(pcString+1)>='8' && *(pcString+1)<='9')
            {
                lwRadix=10;
            }
            else if(*(pcString+1)=='\0')
            {
                /* Special case of the lone zero */
                *plwResult=0;
                return(TRUE);
            }
            else
            {
                return(FALSE);
            }
            break;

        case'1': case'2': case'3': case'4': case'5':
        case'6': case'7': case'8': case'9':
            lwRadix=10;
            break;

        case'A': case'B': case'C': case'D': case'E': case 'F':
            lwRadix=16;
            break;

        default:
            return(FALSE);
    }

    /* Step 4 decode the string backwards */

    pcStartPos=pcString;

    while(*(pcString+1)!=0) pcString++;

    do
    {
        lwTmp=(ULONG)fCharToInt(*pcString);

        /* Test for validity */

//        if(lwTmp<0 || lwTmp>=lwRadix) return(FALSE);
        if(lwTmp>=lwRadix) return(FALSE);


        /* Calculate the Multiplier */

        lwMultiplier=1;
        for(lwLoop=0;lwLoop<lwExponent;lwLoop++) lwMultiplier*=lwRadix;
     
        lwValue+=(ULONG)lwTmp*lwMultiplier;

        /* Increment the exponent */

        lwExponent++;

    }
    while(pcString--!=pcStartPos);
        
    /* Step 5 Save the value */

    *plwResult=lwValue;
        
    return(TRUE);        
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fIsWhitespace()                                           */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Tests if the passed char is a whitespace char.            */
/*                                                                           */
/*   Accepts     : char to be tested                                         */
/*                                                                           */
/*   Returns     : TRUE/FALSE depeding if whitespece or not                  */
/*                                                                           */
/*****************************************************************************/

UBYTE fIsWhiteSpace(UBYTE tData)
{
    if(tData==' ' || tData==',' || tData=='\t') return(TRUE); else return(FALSE);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fCharToInt()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Convert a char to an int in the range 0-15 as hex is      */
/*                 also valid as numerals.                                   */
/*                                                                           */
/*   Accepts     : char                                                      */
/*                                                                           */
/*   Returns     : int                                                       */
/*                                                                           */
/*****************************************************************************/

SWORD fCharToInt(UBYTE cWork)
{
    cWork=toupper(cWork);
    if(cWork>='0' && cWork<='9') return((int)cWork-'0');
    else
    if(cWork>='A' && cWork<='F') return((int)cWork-'A'+10);
    else
    return(-1);
}

/************************************************************/
/* Allocate a region of DOS memory.                         */
/************************************************************/

void far* fDOSmalloc(REALPTR * rp,ULONG bytes_to_allocate)
{
 void far *ptr=NULL;
 union REGS regs;

 bytes_to_allocate = ((bytes_to_allocate + 15) & 0xfffffff0);
 memset(&regs,0,sizeof(regs));
 regs.w.ax = 0x100;
 regs.w.bx = (UWORD)(bytes_to_allocate >> 4);            //Allocate dos
 int386(0x31,&regs,&regs);
 if(regs.x.cflag == 0)
 {       //Everything OK.
   rp->Segment = regs.w.ax;
   rp->Selector = regs.w.dx;
   ptr = MK_FP(regs.w.dx,0);
 }
 return(ptr);
}

/********************************************/
/* Free an area of DOS memory.              */
/********************************************/

void fDOSfree(REALPTR * rp)
{
  union   REGS            regs;

  regs.w.ax = 0x101;
  regs.w.dx = rp->Selector;
  int386(0x31,&regs,&regs);
}



/* End of UTIL.C */

