#ifdef DEBUG_MSG

/*****************************************************************************/
/*                                                                           */
/*   VECTREX HP3000 Software Emulator              Copyright K.Wilkins 1996  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Title      : Vectrex emulator graphics window functions                 */
/*                                                                           */
/*   File Name  : WINDOW.C                                                   */
/*                                                                           */
/*   Author     : Keith Wilkins                                              */
/*                                                                           */
/*   Version    : 1.00                                                       */
/*                                                                           */
/*   Desciption : This file contains the graphics window module of the       */
/*                Vectrex software emulator. These functions only use the    */
/*                services provided by IO.C and rely on no other modules     */
/*                except the DEBUG module for ERROR printing.                */
/*                                                                           */
/*   Functions  : fWinOpen()          - Open a graphics window               */
/*                fWinClose()         - Close a graphics window              */
/*                fWinTitle()         - Add a Title Bar to the window        */
/*                fWinHide()          - Hide a graphics window from view     */
/*                fWinRestore()       - Restore a graphics window            */
/*                fWinActivate()      - Push a window to the top of the disp */
/*                fWinMove()          - Move a window                        */
/*                fWinSetCursorPos()  - Set the cursor position              */
/*                fWinGetCursorPos()  - Return the cursor position           */
/*                fWinPrintf()        - Print to a window                    */
/*                fWinOutText()       - Print to a window (unformatted)      */
/*                fWinClear()         - Clean a window                       */
/*                fWinInput()         - Get a text string from a window      */
/*                fWinPlot()          - Plot a pixel in a window             */
/*                fWinLine()          - Draw a line in a window              */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Revision History:                                                       */
/*                                                                           */
/*   Version    Date    Who  Description of changes                          */
/*   -------    ----    ---  ----------------------                          */
/*                                                                           */
/*    0.01    05/04/93  K.W  Creation                                        */
/*    1.00    05/08/96  K.W  Public release of DVE V1.0                      */
/*                                                                           */
/*****************************************************************************/

#define WINDOW_C

#include "vectrex.h"

#define ALERTBORDWID	4
#define ALERTBACKCOL    (BLUE_BASE+BRIGHT2)
#define ALERTBORDCOL1   (BLUE_BASE+BRIGHT8)
#define ALERTBORDCOL2   (BLUE_BASE+BRIGHT11)


#define WIN_MAGIC_NUMBER    12345678l
#define WIN_PRINT_BUF		1024


/* Define Local variables */

static GWINDOWPTR   tWindowList=NULL;


/* Define local support functions */

BOOL fWinOverlapping(GWINDOWPTR tWin1,GWINDOWPTR tWin2);

BOOL fWinScreenLoad(GWINDOWPTR tWorkWin);
BOOL fWinScreenSave(GWINDOWPTR tWorkWin);
BOOL fWinWindowLoad(GWINDOWPTR tWorkWin);
BOOL fWinWindowSave(GWINDOWPTR tWorkWin);

BOOL fWinListInsert(GWINDOWPTR tNewNode);
BOOL fWinListAppend(GWINDOWPTR tNewNode);
BOOL fWinListDelete(GWINDOWPTR tDeadNode);

/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinOpen()                                                */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Open a graphics window with given position size and width */
/*                 and colour and return pointer to said window.             */
/*                                                                           */
/*   Accepts     : see below                                                 */
/*                                                                           */
/*   Returns     : pointer to GWINDOW structure or NULL if error             */
/*                                                                           */
/*****************************************************************************/

GWINDOWPTR fWinOpen(SWORD iXpos,SWORD iYpos,SWORD iXsize,SWORD iYsize,SWORD iForegroundColour,SWORD iBackgroundColour,SWORD iBorderWidth,SWORD iBorderColour)
{
    GWINDOWPTR  tNewWin=NULL;
    PIXEL       tScreenRes;

    /* Do a bounds check on the window dimensions */

    tScreenRes=fIOGetScreenRes();

    if(iXsize>=WIN_MIN_SIZE && iYsize>=WIN_MIN_SIZE && iXpos>=0 && iYpos>=0 && iBorderWidth>=0 &&
    (iXpos+(2*iBorderWidth)+iXsize<tScreenRes.iXpos) && (iYpos+(2*iBorderWidth)+iYsize<tScreenRes.iYpos))
    {
        /* Lets Malloc a structure for the windowptr */

        if((tNewWin=(GWINDOWPTR)malloc(sizeof(GWINDOW)))==NULL)
        {
            fFatal("WINDOW ERROR: fWinOpen() malloc failed, out of memory.");
        }
        else
        {
            /* Initialise out new structure */

            tNewWin->tMagicNumber=WIN_MAGIC_NUMBER;

            tNewWin->iXAnchorPoint=iXpos;
            tNewWin->iYAnchorPoint=iYpos;

            tNewWin->iXSizeInternal=iXsize;
            tNewWin->iYSizeInternal=iYsize;

            tNewWin->iBorderWidth=iBorderWidth;
            tNewWin->iTitleBarHeight=0;

            tNewWin->iBorderColour=iBorderColour;
            tNewWin->iBackgroundColour=iBackgroundColour;
            tNewWin->iTitleBarBackgroundColour=0;
            strcpy(tNewWin->acTitle,"Untitled");

            fWinSetCursorPos(tNewWin,0,0,iForegroundColour);

            tNewWin->bHidden=FALSE;

            tNewWin->tScreenSave=NULL;
            tNewWin->tWindowSave=NULL;

            tNewWin->ptNextWindow=NULL;
            tNewWin->ptPrevWindow=NULL;

            /* Now lets save the FULL extent of the Window */

            if(fWinScreenSave(tNewWin)==FALSE)
            {
                free(tNewWin);
                tNewWin=NULL;
            }
            else
            {
                /* Scrap the clippling window */

//                fIOSetClipping(-1,-1,-1,-1);

                /* Now PAINT our new window and clear it */

                fIODrawBox(iXpos,iYpos,iXpos+(2*iBorderWidth)+iXsize,iYpos+(2*iBorderWidth)+iYsize,iBorderColour,BOX_FILLED);

                fIODrawBox(iXpos+iBorderWidth,iYpos+iBorderWidth,iXpos+iBorderWidth+iXsize,iYpos+iBorderWidth+iYsize,iBackgroundColour,
                                BOX_FILLED);

                /* Save out new window in the window list */
                fIOSetClipBox(iXpos,iYpos,iXpos+(2*iBorderWidth)+iXsize,iYpos+(2*iBorderWidth)+iYsize);

                fWinListAppend(tNewWin);
            }
        }
    }
    return(tNewWin);
}





/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinClose()                                               */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Close, free and remove a window structure.                */
/*                                                                           */
/*   Accepts     : Pointer to window structure                               */
/*                                                                           */
/*   Returns     : BOOLEAN, TRUE=Close OK, FALSE=Error                       */
/*                                                                           */
/*****************************************************************************/

BOOL fWinClose(GWINDOWPTR tWorkWin)
{
    BOOL bRetVal=TRUE;


    /* Test the window for validity */
    
    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinClose() Bad Window Handle");
        bRetVal=FALSE;
    }
    else
    {
        /* We dont care if the activate fails, well assume we are at the top */

        bRetVal&=fWinActivate(tWorkWin);

        /* Step 1, restore the saved screen */

        bRetVal&=fWinScreenLoad(tWorkWin);


         {
            int iXpos=tWorkWin->iXAnchorPoint;
            int iYpos=tWorkWin->iYAnchorPoint;

            int iXsize=tWorkWin->iXSizeInternal;
            int iYsize=tWorkWin->iYSizeInternal+tWorkWin->iTitleBarHeight;

            int iBorderWidth=tWorkWin->iBorderWidth;
            fIOUnSetClipBox(iXpos,iYpos,iXpos+(2*iBorderWidth)+iXsize,iYpos+(2*iBorderWidth)+iYsize);
        }

        /* Step 2, Delete us from the Window List */

        fWinListDelete(tWorkWin);

        /* Step 3, Free the memory allocated to the window */

        free(tWorkWin);
    }
    return(bRetVal);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinTitle()                                               */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Add a title-bar to a window of a given height. Requires   */
/*                 window to be, saved, screen restored, save larger screen  */
/*                 area, draw title, draw window.                            */
/*                                                                           */
/*   Accepts     :                                                           */
/*                                                                           */
/*   Returns     :                                                           */
/*                                                                           */
/*****************************************************************************/

BOOL fWinTitle(GWINDOWPTR tWorkWin,char *pcTitle,SWORD iFontSize,SWORD iTitleHeight,SWORD iTitleColour,SWORD iBackgroundColour)
{
    PIXEL   tScreenRes,tCursorSave;
    BOOL bRetVal=TRUE;
    SWORD     iXtmp,iYtmp;

    /* Do a sanity check on the parameters */

    tScreenRes=fIOGetScreenRes();

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinTitle() Bad Window Handle");
        bRetVal=FALSE;
    }
    else if(strlen(pcTitle)>WIN_MAX_TITLE_LENGTH)
    {
        bRetVal=FALSE;
    }
    else if(iTitleHeight<BASE_FONT_HEIGHT*iFontSize)
    {
        bRetVal=FALSE;
    }
    else if(tWorkWin->iYAnchorPoint+(2*tWorkWin->iBorderWidth)+tWorkWin->iYSizeInternal > tScreenRes.iYpos)
    {
        bRetVal=FALSE;
    }
    else
    {
        bRetVal&=fWinActivate(tWorkWin);
        
        /* Save the Window internals and restore the screen */

        if((bRetVal&=fWinWindowSave(tWorkWin))!=FALSE && (bRetVal&=fWinScreenLoad(tWorkWin))!=FALSE)
        {
           /* Set the window Params */

           tCursorSave=fWinGetCursorPos(tWorkWin);

           strcpy(tWorkWin->acTitle,pcTitle);
           tWorkWin->iTitleBarBackgroundColour=iBackgroundColour;
           tWorkWin->iTitleBarTextColour=iTitleColour;
           tWorkWin->iTitleBarHeight=iTitleHeight;

           fWinSetCursorPos(tWorkWin,tCursorSave.iXpos,tCursorSave.iYpos,tCursorSave.iColour);

           /* Save the area under the new window */

            if((bRetVal&=fWinScreenSave(tWorkWin))!=FALSE)
            {

            /* Draw the Title Box */

//                fIOSetClipping(tWorkWin->iXAnchorPoint,
//                               tWorkWin->iYAnchorPoint,
//                               tWorkWin->iXAnchorPoint+(2*tWorkWin->iBorderWidth)+tWorkWin->iXSizeInternal,
//                               tWorkWin->iYAnchorPoint+tWorkWin->iTitleBarHeight);

                fIODrawBox(tWorkWin->iXAnchorPoint,
                           tWorkWin->iYAnchorPoint,
                           tWorkWin->iXAnchorPoint+(2*tWorkWin->iBorderWidth)+tWorkWin->iXSizeInternal,
                           tWorkWin->iYAnchorPoint+tWorkWin->iTitleBarHeight,
                           tWorkWin->iTitleBarBackgroundColour,BOX_FILLED);

                /* Display the title using the clipping window to stop overflow */
                /* Work out the text centre position                            */

                iXtmp=tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth+(tWorkWin->iXSizeInternal/2)-
                        ((strlen(pcTitle)*(BASE_FONT_WIDTH*iFontSize)/2));

                iYtmp=tWorkWin->iYAnchorPoint+(tWorkWin->iTitleBarHeight/2)-((BASE_FONT_HEIGHT*iFontSize)/2);

                fIOOutText(iXtmp,iYtmp,tWorkWin->iTitleBarTextColour,tWorkWin->iTitleBarBackgroundColour,iFontSize,tWorkWin->acTitle);

                /* Restore normal clipping window */
        
//                fIOSetClipping(-1,-1,-1,-1);

                /* Now to restore the old window contents, bodge Y pos for border */

                tWorkWin->iYAnchorPoint+=tWorkWin->iTitleBarHeight;
                bRetVal&=fWinWindowLoad(tWorkWin);
                tWorkWin->iYAnchorPoint-=tWorkWin->iTitleBarHeight;

         {
            int iXpos=tWorkWin->iXAnchorPoint;
            int iYpos=tWorkWin->iYAnchorPoint;

            int iXsize=tWorkWin->iXSizeInternal;
            int iYsize=tWorkWin->iYSizeInternal+tWorkWin->iTitleBarHeight;

            int iBorderWidth=tWorkWin->iBorderWidth;
            fIOSetClipBox(iXpos,iYpos,iXpos+(2*iBorderWidth)+iXsize,iYpos+(2*iBorderWidth)+iYsize);
        }


            }
        }
    }
    return(bRetVal);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinHide()                                                */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Bring the window to the top, restore the screen and free  */
/*                 the screen space and save the contents of the window but  */
/*                 dont free the pointer and mark as hidden. Same as a close */
/*                 but we dont free an we dont touch the window list.        */
/*                                                                           */
/*   Accepts     : pointer to the window to be hiddden                       */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE=Hide OK, FALSE=Hide failed                   */
/*                                                                           */
/*****************************************************************************/

BOOL fWinHide(GWINDOWPTR tWorkWin)
{
    BOOL bRetVal=TRUE;

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinHide() Bad Window Handle");
        bRetVal=FALSE;
    }
    else
    {
        fWinActivate(tWorkWin);

        /* Save the window and set the Hidden Flag */

        bRetVal&=fWinWindowSave(tWorkWin);

        tWorkWin->bHidden=TRUE;

        /* Restore the screen */
        bRetVal&=fWinScreenLoad(tWorkWin);

         {
            int iXpos=tWorkWin->iXAnchorPoint;
            int iYpos=tWorkWin->iYAnchorPoint;

            int iXsize=tWorkWin->iXSizeInternal;
            int iYsize=tWorkWin->iYSizeInternal+tWorkWin->iTitleBarHeight;

            int iBorderWidth=tWorkWin->iBorderWidth;
            fIOUnSetClipBox(iXpos,iYpos,iXpos+(2*iBorderWidth)+iXsize,iYpos+(2*iBorderWidth)+iYsize);
        }

    }
    return(bRetVal);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinRestore()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Unhide a previously hidden window. We dont need to do     */
/*                 and activate as our window isnt visible. We will stuff    */
/*                 it on the end of list.                                    */
/*                                                                           */
/*   Accepts     : Pointer to the hidden window                              */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE=Window restored OK, FALSE=Error              */
/*                                                                           */
/*****************************************************************************/

BOOL fWinRestore(GWINDOWPTR tWorkWin)
{
    BOOL bRetVal=TRUE;

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinRestore() Bad Window Handle");
        bRetVal=FALSE;
    }
    else if(tWorkWin->bHidden!=TRUE)
    {
        bRetVal=FALSE;
    }
    else
    {
        /* Step 1, Stuff our window onto the end of the window list */

        fWinListDelete(tWorkWin);
        fWinListAppend(tWorkWin);

        /* Step 2, Kill the hidden status */

        tWorkWin->bHidden=FALSE;

        /* Step 3 Save the screen and restore the window */

        bRetVal&=fWinScreenSave(tWorkWin);
        bRetVal&=fWinWindowLoad(tWorkWin);
         {
            int iXpos=tWorkWin->iXAnchorPoint;
            int iYpos=tWorkWin->iYAnchorPoint;

            int iXsize=tWorkWin->iXSizeInternal;
            int iYsize=tWorkWin->iYSizeInternal+tWorkWin->iTitleBarHeight;

            int iBorderWidth=tWorkWin->iBorderWidth;
            fIOSetClipBox(iXpos,iYpos,iXpos+(2*iBorderWidth)+iXsize,iYpos+(2*iBorderWidth)+iYsize);
        }

    }
    return(bRetVal);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinActive()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Bring the chosen window to the top of the window stack by */
/*                 traversing the list and saving each window and restoring  */
/*                 the screen. We then push our window onto the end of the   */
/*                 list and traverse the window stack again.                 */
/*                                                                           */
/*   Accepts     : Pointer to the required top window                        */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE=Activated OK, FALSE=Error occured            */
/*                                                                           */
/*****************************************************************************/

BOOL fWinActivate(GWINDOWPTR tWorkWin)
{
    GWINDOW     tDummy;
    GWINDOWPTR  tTmp,tPtr;
    BOOL     bRetVal=TRUE,bFlag;

    /* The intelligent approach is to find where our target window is in*/
    /* the window stack and only do the things above it                 */

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinActivate() Bad Window Handle");
        return(FALSE);
    }

    /* Step 1 By appending a dummy node taking its prev pointer then    */
    /* delete the dummy node we an get the end node                     */

    tDummy.ptNextWindow=NULL;
    tDummy.ptPrevWindow=NULL;
    fWinListAppend(&tDummy);
    tPtr=tDummy.ptPrevWindow;
    fWinListDelete(&tDummy);

    /* Check if our window is alrady the top node */

    if(tPtr==tWorkWin) return(TRUE);

    /* Scan UP the list, if nothing is overlapping then we will just pop */
    /* our window to the top with nothing being re-drawn, else we will   */
    /* do a full redraw                                                  */

    bFlag=FALSE;
    tTmp=tPtr;
    do
    {
        /* Test for overlap */

        if(fWinOverlapping(tTmp,tWorkWin)==TRUE)
        {
            bFlag=TRUE;
            break;
        }

        /* Skip to the next window */

        tTmp=tTmp->ptPrevWindow;

    }while(tTmp!=tWorkWin);

    /* Test the result and do append or redraw */

    if(bFlag==FALSE)
    {
        fWinListDelete(tWorkWin);
        fWinListAppend(tWorkWin);
    }
    else
    {
        do
        {
            if(tPtr->bHidden==FALSE)
            {
                /* We only want bRetVal to be valid for the WINDOW that is */
                /* being activated, Save/Load will print errors for probs  */
                /* with other windows but we wont return and error for     */
                /* any window other that our own.                          */

                bRetVal=fWinWindowSave(tPtr);
                bRetVal&=fWinScreenLoad(tPtr);
            }

            if(tPtr==tWorkWin) break;

            tPtr=tPtr->ptPrevWindow;
        }
        while(tPtr!=NULL);

        /* Step 3 Append our window to the end of the list, make tPtr point */
        /* To the NEXT window, the MUST be a next window because we tested  */
        /* the case of the top window begin activated already.              */

        tPtr=tPtr->ptNextWindow;

        fWinListDelete(tWorkWin);
        fWinListAppend(tWorkWin);

        /* Step 4 Traverse the list again from the last position and redraw */

        while(tPtr!=NULL)
        {
            if(tPtr->bHidden==FALSE)
            {
                /* Only set the Error flags for OUR work window */

                if(tPtr==tWorkWin)
                {
                    bRetVal&=fWinScreenSave(tPtr);
                    bRetVal&=fWinWindowLoad(tPtr);
                }
                else
                {
                    fWinScreenSave(tPtr);
                    fWinWindowLoad(tPtr);
                }
            }
            tPtr=tPtr->ptNextWindow;
        }
    }

    return(bRetVal);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinMove()                                                */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Just as the line says move a window, a consistancy check  */
/*                 is done on the new position and the window is moved to    */
/*                 the top of the stack.                                     */
/*                                                                           */
/*   Accepts     : Pointer to window and new X,Y pos                         */
/*                                                                           */
/*   Returns     : BOOLEAN Sucess or fail                                    */
/*                                                                           */
/*****************************************************************************/

BOOL fWinMove(GWINDOWPTR tWorkWin,SWORD iXpos,SWORD iYpos)
{
    BOOL     bRetVal=TRUE;
    PIXEL       tScreenRes;

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinMove() Bad Window Handle");
        return(FALSE);
    }

    /* Do a bounds check on the window dimensions */

    tScreenRes=fIOGetScreenRes();

    if(iXpos>=0 && iYpos>=0 &&
        iXpos+(2*tWorkWin->iBorderWidth)+tWorkWin->iXSizeInternal<tScreenRes.iXpos &&
        iYpos+(2*tWorkWin->iBorderWidth)+tWorkWin->iYSizeInternal+tWorkWin->iTitleBarHeight<tScreenRes.iYpos)
    {
        bRetVal&=fWinActivate(tWorkWin);

        bRetVal&=fWinWindowSave(tWorkWin);
        bRetVal&=fWinScreenLoad(tWorkWin);

        tWorkWin->iXAnchorPoint=iXpos;
        tWorkWin->iYAnchorPoint=iYpos;

        bRetVal&=fWinScreenSave(tWorkWin);
        bRetVal&=fWinWindowLoad(tWorkWin);
    }
    else
    {
        bRetVal=FALSE;
    }
    return(bRetVal);
}  




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinSetCursorPos()                                        */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Set the Window cursor position, but dont activate the     */
/*                 window.                                                   */
/*                                                                           */
/*   Accepts     : Window ptr and position & colour of cursor                */
/*                                                                           */
/*   Returns     : BOOLEAN                                                   */
/*                                                                           */
/*****************************************************************************/

BOOL fWinSetCursorPos(GWINDOWPTR tWorkWin,SWORD iXpos,SWORD iYpos,SWORD iColour)
{
    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinSetCursorPos() Bad Window Handle");
        return(FALSE);
    }

    /* The cursor is a relative cursor, we wont bound check it as its clipped */
    /* BUT WE MUST convert it to an absolute cursor position */

    tWorkWin->tCursorPos.iXpos=iXpos+tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth;
    tWorkWin->tCursorPos.iYpos=iYpos+tWorkWin->iYAnchorPoint+tWorkWin->iTitleBarHeight+tWorkWin->iBorderWidth;

    tWorkWin->tCursorPos.iColour=iColour;

    return(TRUE);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinGetCursorPos()                                        */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Return the cursor position in a PIXEL struct              */
/*                                                                           */
/*   Accepts     : Pointer to window                                         */
/*                                                                           */
/*   Returns     : Pixel struct 0,0,0 if error                               */
/*                                                                           */
/*****************************************************************************/

PIXEL fWinGetCursorPos(GWINDOWPTR tWorkWin)
{
    PIXEL   tRelative={0,0,0};

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinGetCursorPos() Bad Window Handle");
    }
    else
    {
        /* We must convert from an absolute cursor to a relative one */

        tRelative.iXpos=tWorkWin->tCursorPos.iXpos-tWorkWin->iXAnchorPoint-tWorkWin->iBorderWidth;
        tRelative.iYpos=tWorkWin->tCursorPos.iYpos-tWorkWin->iYAnchorPoint-tWorkWin->iTitleBarHeight-tWorkWin->iBorderWidth;
        tRelative.iColour=tWorkWin->tCursorPos.iColour;
    }
    return(tRelative);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinPrintf()                                              */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Print a formatted string to a window.                     */
/*                                                                           */
/*   Accepts     : Window ptr and formatted text string                      */
/*                                                                           */
/*   Returns     : BOOLEAN Sucess/Fail TRUE/FALSE                            */
/*                                                                           */
/*****************************************************************************/

BOOL fWinPrintf(GWINDOWPTR tWorkWin,SWORD iFontSize,char *pcFormat,...)
{
    BOOL     bRetVal=TRUE;
    char        acTextBuffer[WIN_PRINT_BUF];
    va_list     tGlop;

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinPrintf() Bad Window Handle");
        bRetVal=FALSE;
    }
    else
    {
        /* Expand the varable argument list into the text buffer */

        va_start(tGlop,pcFormat);                   						/* Initialise the variable array */
        
        if(vsprintf(acTextBuffer,pcFormat,tGlop)==-1)      	/* Expand the args to a string */
		{
	            fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_LARGE,"fWinPrintf: Text Buffer Overflow");
	    }
        
        va_end(tGlop);

        /* Use out text to print it */

        bRetVal=fWinOutText(tWorkWin,iFontSize,acTextBuffer);
    }
    return(bRetVal);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinOutChar()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Output a char to a window, not control of formatting is   */
/*                 done for \t \n \b etc.                                    */
/*                                                                           */
/*   Accepts     : Pointer to window, text size and cData char               */
/*                                                                           */
/*   Returns     : BOOLEAN sucess or failure                                 */
/*                                                                           */
/*****************************************************************************/

BOOL fWinOutChar(GWINDOWPTR tWorkWin,SWORD iFontSize,char cData)
{
    BOOL     bRetVal=TRUE;

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinOutChar() Bad Window Handle");
        bRetVal=FALSE;
    }
    else
    {
        /* Push window to top of the window stack */

        bRetVal&=fWinActivate(tWorkWin);

        /* Set the Clipping Window */

//        fIOSetClipping(tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth,
//                       tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight,
//                       tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iXSizeInternal,
//                       tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight+tWorkWin->iYSizeInternal);

        tWorkWin->tCursorPos=fIOOutChar(tWorkWin->tCursorPos.iXpos,
                                        tWorkWin->tCursorPos.iYpos,
                                        tWorkWin->tCursorPos.iColour,
                                        tWorkWin->iBackgroundColour,
                                        iFontSize,
                                        cData);

        /* Reset the Clipping Window */

//        fIOSetClipping(-1,-1,-1,-1);
    }
    return(bRetVal);
}





/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinOutText()                                             */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Output a text string to a window. Escape sequences are    */
/*                 interrpreted for \n \b \t. Text is printed in the given   */
/*                 text size.                                                */
/*                                                                           */
/*   Accepts     : Pointer to window, text size and text string              */
/*                                                                           */
/*   Returns     : BOOLEAN sucess or failure                                 */
/*                                                                           */
/*****************************************************************************/

BOOL fWinOutText(GWINDOWPTR tWorkWin,SWORD iFontSize,char *pcText)
{
    BOOL     bRetVal=TRUE;

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinOutText() Bad Window Handle");
        bRetVal=FALSE;
    }
    else
    {
        /* Push window to top of the window stack */

        bRetVal&=fWinActivate(tWorkWin);

        /* Set the Clipping Window */

//        fIOSetClipping(tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth,
//                       tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight,
//                       tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iXSizeInternal,
//                       tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight+tWorkWin->iYSizeInternal);

        /* Decode the text buffer, one char at a time decoding escape sequences */

        while(*pcText!='\0')
        {
            /* To stop screwups and scrolling to far we cant do a CR/Scroll immediately    */
            /* because we dont know the hieight of the text to be printed so we must check */
            /* the cursor position is valid on every write                                 */

            if(tWorkWin->tCursorPos.iYpos+(BASE_FONT_HEIGHT*iFontSize) >
                tWorkWin->iYAnchorPoint+tWorkWin->iTitleBarHeight+tWorkWin->iBorderWidth+tWorkWin->iYSizeInternal)
            {

                /* We will scroll from the inside window corner + one line to the bottom */
                /* with a destination of the window top                                  */

                fIOAreaScroll(tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth,
                              tWorkWin->iYAnchorPoint+tWorkWin->iTitleBarHeight+tWorkWin->iBorderWidth+(BASE_FONT_HEIGHT*iFontSize),
                              tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iXSizeInternal,
                              tWorkWin->iYAnchorPoint+tWorkWin->iTitleBarHeight+tWorkWin->iBorderWidth+tWorkWin->iYSizeInternal,
                              tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth,
                              tWorkWin->iYAnchorPoint+tWorkWin->iTitleBarHeight+tWorkWin->iBorderWidth,
                              tWorkWin->iBackgroundColour);

                /* Correct the cursor position */

                tWorkWin->tCursorPos.iYpos-=BASE_FONT_HEIGHT*iFontSize;
                
            }

            switch(*pcText)
            {
                case '\n':

                    tWorkWin->tCursorPos.iXpos=tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth;
                    tWorkWin->tCursorPos.iYpos+=BASE_FONT_HEIGHT*iFontSize;
                    break;

                case '\b':

                    tWorkWin->tCursorPos.iXpos-=BASE_FONT_WIDTH*iFontSize;

                    if(tWorkWin->tCursorPos.iXpos < tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth)
                    {
                        tWorkWin->tCursorPos.iXpos= tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth;
                    }
                    break;

                case '\t':
                    tWorkWin->tCursorPos=fIOOutText(tWorkWin->tCursorPos.iXpos,
                                                    tWorkWin->tCursorPos.iYpos,
                                                    tWorkWin->tCursorPos.iColour,
                                                    tWorkWin->iBackgroundColour,
                                                    iFontSize,
                                                    "    ");
                    break;

                default:
                    tWorkWin->tCursorPos=fIOOutChar(tWorkWin->tCursorPos.iXpos,
                                                    tWorkWin->tCursorPos.iYpos,
                                                    tWorkWin->tCursorPos.iColour,
                                                    tWorkWin->iBackgroundColour,
                                                    iFontSize,
                                                    *pcText);
                    break;
            }
            pcText++;
        }

        /* Reset the Clipping Window */

//        fIOSetClipping(-1,-1,-1,-1);
    }
    return(bRetVal);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinClear()                                               */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Clear a window an reset its cursor position.              */
/*                                                                           */
/*   Accepts     : Pointer to a window                                       */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE/FALSE                                        */
/*                                                                           */
/*****************************************************************************/

BOOL fWinClear(GWINDOWPTR tWorkWin)
{
    BOOL bRetVal;

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinClear() Bad Window Handle");
        bRetVal=FALSE;
    }
    else
    {
        bRetVal=fWinActivate(tWorkWin);

        fIOAreaClear(tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth,
                     tWorkWin->iYAnchorPoint+tWorkWin->iTitleBarHeight+tWorkWin->iBorderWidth,
                     tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iXSizeInternal,
                     tWorkWin->iYAnchorPoint+tWorkWin->iTitleBarHeight+tWorkWin->iBorderWidth+tWorkWin->iYSizeInternal,
                     tWorkWin->iBackgroundColour);

        fWinSetCursorPos(tWorkWin,0,0,tWorkWin->tCursorPos.iColour);
    }
    return(bRetVal);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinInput()                                               */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Read an input string from the keyboard with editing and   */
/*                 echo to the window.                                       */
/*                                                                           */
/*   Accepts     : Pointer to window, fontsize, input buffer max string len  */
/*                                                                           */
/*   Returns     : int Number of characters entered or 0 for error/no data   */
/*                                                                           */
/*****************************************************************************/

SWORD fWinInput(GWINDOWPTR tWorkWin,SWORD iFontSize,char *pcBuffer,SWORD iMaxLength)
{
    SWORD     iRetVal;

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinInput() Bad Window Handle");
        iRetVal=0;
    }
    else
    {
        if(fWinActivate(tWorkWin)==FALSE) return(0);

        /* Set the clipping window */

//        fIOSetClipping(tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth,
//                       tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight,
//                       tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iXSizeInternal,
//                       tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight+tWorkWin->iYSizeInternal);


        tWorkWin->tCursorPos=fIOGetLine(tWorkWin->tCursorPos.iXpos,
                                     tWorkWin->tCursorPos.iYpos,
                                     tWorkWin->tCursorPos.iColour,
                                     tWorkWin->iBackgroundColour,
                                     iFontSize,
                                     pcBuffer,
                                     (UBYTE)iMaxLength);

        /* Clear the clipping window */

//        fIOSetClipping(-1,-1,-1,-1);


    }
    return(iRetVal);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinPlot()                                                */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Plot a pixel inside and relative to a window, clipped     */
/*                                                                           */
/*   Accepts     : Wondow pointer X,Y position and colour                    */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE/FALSE                                        */
/*                                                                           */
/*****************************************************************************/

BOOL fWinPlot(GWINDOWPTR tWorkWin,SWORD iXpos,SWORD iYpos,SWORD iColour)
{
    BOOL bRetVal;

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinPlot() Bad Window Handle");
        bRetVal=FALSE;
    }
    else
    {
        bRetVal&=fWinActivate(tWorkWin);
    
        /* Set the clipping window */

//        fIOSetClipping(tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth,
//                       tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight,
//                       tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iXSizeInternal,
//                       tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight+tWorkWin->iYSizeInternal);

        /* Plot the Pixel, relative to the corner of the window */

        fIOPlotPixel(tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth+iXpos,
                     tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight+iYpos,
                     iColour);

        /* Clear the clipping window */

//        fIOSetClipping(-1,-1,-1,-1);

    }
    return(bRetVal);
}




/*****************************************************************************/
/*                                                                           */
/*   Function    : fWinLine()                                                */
/*                                                                           */
/*   Author      : K.W                                                       */
/*                                                                           */
/*   Description : Draw a line relative inside and clipped to a window       */
/*                                                                           */
/*   Accepts     : Window and Line X,Y start, X,Y end and colour             */
/*                                                                           */
/*   Returns     : BOOLEAN TRUE/FALSE                                        */
/*                                                                           */
/*****************************************************************************/

BOOL fWinLine(GWINDOWPTR tWorkWin,SWORD iX1,SWORD iY1,SWORD iX2,SWORD iY2,SWORD iColour)
{
    BOOL bRetVal;

    if(tWorkWin->tMagicNumber!=WIN_MAGIC_NUMBER)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinLine() Bad Window Handle");
        bRetVal=FALSE;
    }
    else
    {
        bRetVal&=fWinActivate(tWorkWin);
        
        /* Set the clipping window */

//        fIOSetClipping(tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth,
//                       tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight,
//                       tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iXSizeInternal,
//                       tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight+tWorkWin->iYSizeInternal);

        /* Draw the Line */

        fIODrawLine(tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth+iX1,
                    tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight+iY1,
                    tWorkWin->iXAnchorPoint+tWorkWin->iBorderWidth+iX2,
                    tWorkWin->iYAnchorPoint+tWorkWin->iBorderWidth+tWorkWin->iTitleBarHeight+iY2,
                    iColour);

        /* Clear the clipping window */

//        fIOSetClipping(-1,-1,-1,-1);

    }
    return(bRetVal);
}





/********************************************************************/
/*                                                                  */
/*                                                                  */
/*                 Local Support Functions (Private)                */
/*                                                                  */
/*                                                                  */
/********************************************************************/

BOOL fWinOverlapping(GWINDOWPTR tWin1,GWINDOWPTR tWin2)
{
    SWORD iWin1X1,iWin1X2,iWin1Y1,iWin1Y2;
    SWORD iWin2X1,iWin2X2,iWin2Y1,iWin2Y2;

    /* Work out the window extents */

    iWin1X1=tWin1->iXAnchorPoint;
    iWin1Y1=tWin1->iYAnchorPoint;
    iWin1X2=tWin1->iXAnchorPoint+(2*tWin1->iBorderWidth)+tWin1->iXSizeInternal;
    iWin1Y2=tWin1->iYAnchorPoint+tWin1->iTitleBarHeight+(2*tWin1->iBorderWidth)+tWin1->iYSizeInternal;

    iWin2X1=tWin2->iXAnchorPoint;
    iWin2Y1=tWin2->iYAnchorPoint;
    iWin2X2=tWin2->iXAnchorPoint+(2*tWin2->iBorderWidth)+tWin2->iXSizeInternal;
    iWin2Y2=tWin2->iYAnchorPoint+tWin2->iTitleBarHeight+(2*tWin2->iBorderWidth)+tWin2->iYSizeInternal;

    /* Now test for overlap, if any Win1 corner is inside Win2 */

    if(iWin1X1>=iWin2X1 && iWin1X1<=iWin2X2 && iWin1Y1>=iWin2Y1 && iWin1Y1<=iWin2Y2) return(TRUE);
    if(iWin1X2>=iWin2X1 && iWin1X2<=iWin2X2 && iWin1Y2>=iWin2Y1 && iWin1Y2<=iWin2Y2) return(TRUE);
    if(iWin1X1>=iWin2X1 && iWin1X1<=iWin2X2 && iWin1Y2>=iWin2Y1 && iWin1Y2<=iWin2Y2) return(TRUE);
    if(iWin1X2>=iWin2X1 && iWin1X2<=iWin2X2 && iWin1Y1>=iWin2Y1 && iWin1Y1<=iWin2Y2) return(TRUE);

    /* Test the case were no corners are inside but one window crosses the other */
    /* with no corners being inside the other                                    */

    if(iWin1X1>iWin2X1 && iWin1X1<iWin2X2 && iWin1X2>iWin2X1 && iWin1X2<iWin2X2 && 
      (iWin1Y1<=iWin2Y1 || iWin1Y1>=iWin2Y2) && (iWin1Y2<=iWin2Y1 || iWin1Y2>=iWin2Y2)) return(TRUE);

    if((iWin1X1<=iWin2X1 || iWin1X1>=iWin2X2) && (iWin1X2<=iWin2X1 || iWin1X2>=iWin2X2) && 
       iWin1Y1>iWin2Y1 && iWin1Y1<iWin2Y2 && iWin1Y2>iWin2Y1 && iWin1Y2<iWin2Y2) return(TRUE);

    if(iWin2X1>iWin1X1 && iWin2X1<iWin1X2 && iWin2X2>iWin1X1 && iWin2X2<iWin1X2 && 
      (iWin2Y1<=iWin1Y1 || iWin2Y1>=iWin1Y2) && (iWin2Y2<=iWin1Y1 || iWin2Y2>=iWin1Y2)) return(TRUE);

    if((iWin2X1<=iWin1X1 || iWin2X1>=iWin1X2) && (iWin2X2<=iWin1X1 || iWin2X2>=iWin1X2) && 
       iWin2Y1>iWin1Y1 && iWin2Y1<iWin1Y2 && iWin2Y2>iWin1Y1 && iWin2Y2<iWin1Y2) return(TRUE);

    return(FALSE);
}

/* Base window load and save functions */

BOOL fWinScreenLoad(GWINDOWPTR tWorkWin)
{
    BOOL bRetVal=TRUE;

    if(tWorkWin->tScreenSave==NULL)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinScreenLoad() NULL Screen save handle");
        bRetVal=FALSE;
    }
    else
    {
        if(fIOAreaLoad(tWorkWin->iXAnchorPoint,tWorkWin->iYAnchorPoint,tWorkWin->tScreenSave)==FALSE)
        {
            bRetVal=FALSE;
        }
        else
        {
            fIOBitBlockRelease(tWorkWin->tScreenSave);
            tWorkWin->tScreenSave=NULL;
        }
    }
    return(bRetVal);
}



BOOL fWinScreenSave(GWINDOWPTR tWorkWin)
{
    BOOL bRetVal=TRUE;

    if(tWorkWin->tScreenSave!=NULL)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinScreenSave() Screen save handle is not NULL");
        bRetVal=FALSE;
    }
    else
    {
       tWorkWin->tScreenSave=fIOAreaSave(tWorkWin->iXAnchorPoint,
                                        tWorkWin->iYAnchorPoint,
                                        tWorkWin->iXAnchorPoint+tWorkWin->iXSizeInternal+(2*tWorkWin->iBorderWidth),
                                        tWorkWin->iYAnchorPoint+tWorkWin->iYSizeInternal+(2*tWorkWin->iBorderWidth)+tWorkWin->iTitleBarHeight);

       if(tWorkWin->tScreenSave==NULL)
       {
           bRetVal=FALSE;
       }
    }
    return(bRetVal);
}



BOOL fWinWindowLoad(GWINDOWPTR tWorkWin)
{
    BOOL bRetVal=TRUE;

    if(tWorkWin->tWindowSave==NULL)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinWindowLoad() NULL Screen save handle");
        bRetVal=FALSE;
    }
    else
    {
        if(fIOAreaLoad(tWorkWin->iXAnchorPoint,tWorkWin->iYAnchorPoint,tWorkWin->tWindowSave)==FALSE)
        {
            bRetVal=FALSE;
        }
        else
        {
            fIOBitBlockRelease(tWorkWin->tWindowSave);
            tWorkWin->tWindowSave=NULL;
        }
    }
    return(bRetVal);
}



BOOL fWinWindowSave(GWINDOWPTR tWorkWin)
{
    BOOL bRetVal=TRUE;

    if(tWorkWin->tWindowSave!=NULL)
    {
        fIOOutText(0,0,BRIGHT15,BRIGHT0,FONT_NORMAL,"WINDOW ERROR: fWinWindowSave() Window save handle is not NULL");
        bRetVal=FALSE;
    }
    else
    {
       tWorkWin->tWindowSave=fIOAreaSave(tWorkWin->iXAnchorPoint,
                                        tWorkWin->iYAnchorPoint,
                                        tWorkWin->iXAnchorPoint+tWorkWin->iXSizeInternal+(2*tWorkWin->iBorderWidth),
                                        tWorkWin->iYAnchorPoint+tWorkWin->iYSizeInternal+(2*tWorkWin->iBorderWidth)+tWorkWin->iTitleBarHeight);

       if(tWorkWin->tWindowSave==NULL)
       {
           bRetVal=FALSE;
       }
    }
    return(bRetVal);
}




/* Window Node list manipulation */

BOOL fWinListInsert(GWINDOWPTR tNewNode)
{
    GWINDOWPTR  tPtr;

    /* Save the old pointer */

    tPtr=tWindowList;

    if(tPtr!=NULL) tPtr->ptPrevWindow=tNewNode;

    tWindowList=tNewNode;

    /* Set the pointers in our new top node */

    tWindowList->ptNextWindow=tPtr;
    tWindowList->ptPrevWindow=NULL;

    return(TRUE);
}



BOOL fWinListAppend(GWINDOWPTR tNewNode)
{
    GWINDOWPTR  tPtr=tWindowList;

    /* Check to the special case of an empty list */

    if(tPtr==NULL)
    {
        tPtr=tNewNode;
        tPtr->ptNextWindow=NULL;
        tPtr->ptPrevWindow=NULL;
        tWindowList=tPtr;
        return(TRUE);
    }
    else
    {
        /* Scan to the list end */

        while(tPtr!=NULL)
        {
            if(tPtr->ptNextWindow==NULL)
            {
                tPtr->ptNextWindow=tNewNode;
                tNewNode->ptPrevWindow=tPtr;
                tNewNode->ptNextWindow=NULL;
                return(TRUE);
            }
            else
            {
                tPtr=tPtr->ptNextWindow;
            }
        }
    }
    return(FALSE);
}



BOOL fWinListDelete(GWINDOWPTR tDeadNode)
{
    GWINDOWPTR  tPtr=tWindowList;

    /* Check the explicit case of an empty list */

    if(tPtr==NULL) return(FALSE);

    /* And the explicit case of the ROOT node */

    if(tDeadNode==tPtr)
    {
    	if(tPtr->ptNextWindow==NULL)
    	{
    		tPtr->ptNextWindow=NULL;
    		tPtr->ptPrevWindow=NULL;
    		tWindowList=NULL;
    	}
    	else
    	{
    		/* Make the under node the root node */
    		
    		tWindowList=tWindowList->ptNextWindow;
    		tWindowList->ptPrevWindow=NULL;
    		
    		/* Clean the old root node of links */
    		
    		tPtr->ptNextWindow=NULL;
    		tPtr->ptPrevWindow=NULL;
    		
        }
        return(TRUE);
    }

    /* Now do a list scan */

    while(tPtr!=NULL)
    {
        if(tPtr==tDeadNode)
        {
            (tPtr->ptPrevWindow)->ptNextWindow=tPtr->ptNextWindow;

            if(tPtr->ptNextWindow!=NULL) (tPtr->ptNextWindow)->ptPrevWindow=tPtr->ptPrevWindow;

            tPtr->ptNextWindow=NULL;
            tPtr->ptPrevWindow=NULL;

            return(TRUE);
        }
        else
        {
            tPtr=tPtr->ptNextWindow;
        }
    }
    fDebug(VERROR,("fWinListDelete() Delete node not found in window list"));
    return(FALSE);
}


/**********************************************/
/*                                            */
/*                                            */
/*          Utility routines                  */
/*                                            */
/*                                            */
/**********************************************/

GWINDOWPTR fQuickWin(SWORD iColour,UBYTE *pcFormat,...)
{
    UBYTE       acTextBuffer[WIN_PRINT_BUF];
    va_list     tGlop;
    GWINDOWPTR	pTmpWin=NULL;
    PIXEL		sScreen;
    SWORD		iWindowXSize,iWindowYSize,iTmp;
    UBYTE		*pcPtr;


    /* Expand the varable argument list into the text buffer */

    va_start(tGlop,pcFormat);                   						/* Initialise the variable array */
        
    if(vsprintf(acTextBuffer,pcFormat,tGlop)==-1)      	/* Expand the args to a string */
	{
	    fFatal("fQuickWin() - Text Buffer Overflow");
	}
        
    va_end(tGlop);
    
    /* Scan the string and work out the window size in chars */

	sScreen=fIOGetScreenRes();    
	pcPtr=acTextBuffer;

	iTmp=1;
	while(*pcPtr!='\0')
	{
		if(*pcPtr=='\n') iTmp++;
		pcPtr++;
	}
	iWindowYSize=iTmp*8;
	
	iTmp=0;
	iWindowXSize=0;
	pcPtr=acTextBuffer;
	while(*pcPtr!='\0')
	{
		if(*pcPtr!='\n')
		{
			iTmp++;
		}
		else
		{
			if(iTmp>iWindowXSize) iWindowXSize=iTmp;
			iTmp=0;
		}
		pcPtr++;
	}
	if(iTmp>iWindowXSize) iWindowXSize=iTmp;
	iWindowXSize=iWindowXSize*6;
	fDebug(MID,("fQuickWin() - Xsixe=%d Ysize=%d",iWindowXSize,iWindowYSize));
		    
	/* Open the window to print in */

	if((pTmpWin=fWinOpen((sScreen.iXpos/2)-(iWindowXSize/2)-ALERTBORDWID,
						 (sScreen.iYpos/2)-(iWindowYSize/2)-ALERTBORDWID,
	  					 iWindowXSize,iWindowYSize,
					  	 iColour,ALERTBACKCOL,
					  	 ALERTBORDWID,ALERTBORDCOL1))==NULL)
	{
		fFatal("fQuickWin() - Could not open window");
	}

   /* Use out text to print it, fudge the cursor pos */

   if(fWinOutText(pTmpWin,FONT_SMALL,acTextBuffer)==FALSE)
   {
       fFatal("fQuickWin() - Error printing text");
       return(NULL);
   }

   return(pTmpWin);
}


GWINDOWPTR fQuickWinPos(SWORD iXpos,SWORD iYpos,SWORD iColour,UBYTE *pcFormat,...)
{
    char        acTextBuffer[WIN_PRINT_BUF];
    va_list     tGlop;
    GWINDOWPTR	pTmpWin=NULL;

    /* Expand the varable argument list into the text buffer */

    va_start(tGlop,pcFormat);                   						/* Initialise the variable array */
        
//    if(_vsnprintf(acTextBuffer,WIN_PRINT_BUF,pcFormat,tGlop)==-1)      	/* Expand the args to a string */
    if(vsprintf(acTextBuffer,pcFormat,tGlop)==-1)      	/* Expand the args to a string */
	{
	    fFatal("fQuickWin() - Text Buffer Overflow");
	    return(NULL);
	}
        
    va_end(tGlop);

	/* Cheat, do a quick win and then move it */
	
	if((pTmpWin=fQuickWin(iColour,acTextBuffer))==NULL) return(NULL);
	if(fWinMove(pTmpWin,iXpos,iYpos)==FALSE)
	{
		fWinClose(pTmpWin);
		return(NULL);
	}
	
	return(pTmpWin);
}
    



SWORD fAlertBox(SWORD iColour,UBYTE *pcFormat,...)
{
   UBYTE       acTextBuffer[WIN_PRINT_BUF];
   va_list     tGlop;
   GWINDOWPTR  pTmpWin=NULL;
   UWORD        iGetch;


   /* Expand the varable argument list into the text buffer */

   va_start(tGlop,pcFormat);                   						/* Initialise the variable array */
        
   if(vsprintf(acTextBuffer,pcFormat,tGlop)==-1)      	/* Expand the args to a string */
   {
       fFatal("fAlertBox() - Text Buffer Overflow");
   }
        
   va_end(tGlop);
    
	/* Open the window to print in */
    
	if((pTmpWin=fQuickWin(iColour,acTextBuffer))==NULL)
	{
		fFatal("fAlertBox() - Could not open window");
	}


	while(!fKBDhit());
	
	iGetch=fKBDgetch();
    
   fWinClose(pTmpWin);
    
   return(iGetch);
}


#endif

/* End of WINDOW.C */
